/* 
 * xerxes.c
 *   
 *   Copyright 1996 X-ray Instrumentation Associates
 *   All rights reserved
 *
 *   Created        25-Sep-1996  by Ed Oltman
 *   Extensive mods 19-Dec-1996  EO
 *   Modified:      03-Feb-1997  EO: Added "static" qualifier to variable defs
 *        Initialize some variables that may be undefined, add getenv to fopen
 *        for unix compatibility, replace '#include <dxp_area/...>' with
 *        '#include <...>' (unix compatability) -->compiler now requires
 *        path for include files; fixed bugs in dxp_initialize_ASC,
 *        dxp_assign_channel,dxp_det_to_elec(); replaced constants used in 
 *        dxp_initialize_ASC with parameters defined in dxp.h
 *    Modified      05-Feb-1997 EO: introduce FNAME replacement for VAX
 *        compatability (getenv crashes if logical not defined)
 *    Modified      17-Feb-1997 EO: fix bug in dxp_det_to_elec
 *    Modified      17-Jul-1997 EO: fix bug in dxp_initialize_ASC
 *    Modified      04-Oct-1997 EO: Make function declarations compatable
 *        compatable w/traditional C; Replace LITTLE_ENDIAN parameter with call
 *        to dxp_little_endian, fix bug in dxp_write_spectrua,Changed file 
 *        system:new file dxp_filename: Now get all file names from single list 
 *        of file names pointed to by "dxpfile_cfg"
 *    09-May-2000 JEW: Complete rework of library to allow use of both a
 *        DXP4C and DXP4C2X simultaneously.
 *
 *
 *   This file contains mid-level DXP control routines for handling a multi-
 *   channel array.  Most of these routines can be used as is or can serve as 
 *   examples for how the primitive routines are used. 
 *
 *      dxp_initialize:  Global initialization of the system.  Makes all 
 *        needed calls to routines to start the system running.
 *
 *      dxp_assign_channel:  Assigns an I/O channel to each DXP module listed
 *        in the MODULE definition file listed in file pointed to by XIA_CONFIG.
 *        This file associates a detector (e.g. logical channel number) to each
 *        module and channels within a module (e.g. electronic channel number). 
 *		  This routine then performs all associations and loading of DSP and Fippi
 *		  code for each module.
 *
 *      dxp_user_setup: Configures each DXP channel in the system according to 
 *        the information loaded previously with dxp_assign_channel.
 *        The configuration consists three parts: downloading the FIPPI firmware,
 *        downloading the DSP firmware, and downloading a set of parameters
 *        which control the DSP and FiPPI.  Calibration runs are also taken.
 *
 *		dxp_reset_channel: Reset a single channel by reloading the FIPPI, DSP
 *		  and default parameters.  Finish by performing all calibrations.
 *
 *		dxp_reset_dspconfig: Reload the configuration for the DSP in a single
 *		  single channel.
 *
 *		dxp_reset_fipconfig: Reload the configuration for the FIPPI in a single
 *		  single channel.
 *
 *static dxp_read_modules: Reads in the module information from the configuration
 *        file.  This just reads the modules section of the configuration, it is 
 *        used for both restoring configurations and for initial startup.
 *
 *      dxp_free_board:  Frees the memory allocated for a Board Structure.
 *
 *      dxp_add_dsp: add a DSP structure to the system.  Adds the information to
 *        a global linked list and then points the board information to this
 *        configuration. 
 *
 *      dxp_add_defaults: add a DSP defaults structure to the system.  Adds the information to
 *        a global linked list and then points the board information to this
 *        configuration. 
 *
 *      dxp_add_fippi: add a FIPPI structure to the system.  Adds the information to
 *        a global linked list and then points the board information to this
 *        configuration. 
 *
 *      dxp_add_iface: add a new Interface structure to the system.  Adds the information to
 *        a global linked list and then points the board information to this
 *        configuration. 
 *
 *static dxp_pick_filename:  Pick a filename from a string from point n to m
 *        assuming spaces in the filename.  This routine removes spaces from the 
 *        beginning and end of the string, returning the filname itself.  It 
 *        does not check if the file is valid.
 *
 *static dxp_strnstrm:  searches a for string matching a token starting from
 *        location n and returning the starting location m.
 *
 *		dxp_dspdefaults: routine to replace the parameter values for a channel in 
 *        library memory with the values previously loaded into the defaults
 *        structure for that channel.  The new parameters must be downloaded
 *		  to the channel after replacement.
 *
 *      dxp_start_run: Starts data taking for all channels in the system
 *
 *      dxp_stop_run: Stops data taking for all channels in the system
 *
 *static dxp_readout_run:  Reads out the parameter memory, baseline histogram
 *        memory and spectum memory for a single Board and channel.  Checks the 
 *        error word in the parameter memory and clears it if set.
 *
 *      dxp_readout_detector_run:  Reads out the parameter memory, baseline histogram
 *        memory and spectum memory for a detector.  Checks the 
 *        error word in the parameter memory and clears it if set.
 *
 *      dxp_write_spectra: reads out each DXP channel, displays some run
 *        statistics, and writes out all of the spectra to one file
 *        and all of the baselines to another.
 *
 *      dxp_fipconfig: Download FiPPI firmware to all modules.
 *        
 *      dxp_replace_fipconfig: Replaces the FiPPI firmware in a single
 *        detector channel.
 *        
 *      dxp_dspconfig: Download DSP firmware to all modules.
 *        
 *      dxp_replace_dspconfig: Replaces the DSP firmware in a single
 *        detector channel.
 *        
 *		dxp_put_dspparams:  Load all channels with parameters in library
 *		  memory.
 *
 *		dxp_replace_dspparams:  Load a single channel with parameters in 
 *		  library memory.
 *
 *		dxp_upload_dspparams:  Load library memory with parameters from a 
 *		  single channel.
 *
 *      dxp_download_params: Download the same set of parameters to all modules.
 *        Parameters to download are defined by address relative to parameter
 *        base memory and value array passed to routine.
 *
 *		dxp_download_one_params:  Download a set of addr-value pairs to a 
 *		  single channel.
 *
 *		dxp_upload_channel:  Load parameter memory from a single channel into
 *		  library memory.
 *
 *		dxp_get_statistics:  Read run statistics from a single channel.
 *
 *		dxp_nspec:  Retrieve the number of entries in spectrum memory for a 
 *		  given channel.
 *
 *		dxp_nbase:  Retrieve the number of entries in baseline memory for a 
 *		  given channel.
 *
 *		dxp_nevent:  Retrieve the number of entries in event memory for a 
 *		  given channel.
 *
 *		dxp_ndxp:  Retrieve the number of dxp modules in the system.
 *
 *		dxp_ndxpchan:  Retrieve the number of dxp channels in the system.
 *
 *		dxp_get_one_dspsymbol:  retrieve a single DSP parameter by name
 *		  from a specified channel.
 *
 *		dxp_set_one_dspsymbol:  set a single DSP parameter by name
 *		  in a specified channel.
 *
 *		dxp_set_dspsymbol:  set a single DSP parameter by name
 *		  in all channels.
 *
 *		dxp_symbolname_list:  returns an array containing all the symbol
 *		  names for the specified detector channel.
 *
 *		dxp_max_symbols:  returns the number of symbols in the specified
 *		  DSP parameter memory.
 *
 *		dxp_mem_dump:  dump the parameter memory to the screen for the 
 *		  specified channel.
 *
 *      dxp_check_decimation: Checks to see if any of the DXP channels has a
 *        value of the DECIMATION parameter different from function argument 
 *        value
 *
 *      dxp_calibrate: Performs one of the internal calibration functions
 *        for all channels: e.g. start and stop runs and check results
 *
 *      dxp_findpeak: Searches spectra for a maxima and returns the bin limits 
 *        that span the peak at some threshold value (e.g. fw at thresh*max)
 *
 *      dxp_fitgauss0: gaussian "fitter" based on lineaization method -- only
 *        computes peak position and FWHM -- no errors
 *
 *		dxp_save_dspparams:  save the DSP parameters in a single channel to 
 *		  a file as name-value pairs.
 *
 *      dxp_save_config: Writes to a file the DSP and FippI filenames for 
 *        each detector channel as well as information on how to communicate
 *        with each board.  Reads the parameter memory of all DXP channels and
 *        writes the parameters in hexidecimal format to a file along with the 
 *        parameter names.
 *
 *		dxp_restore_config:  Restore Fippi, DSP and parameter values previously
 *		  saved using dxp_save_config.
 *
 *      dxp_open_file: opens a file for reading or writing and assigns a logical
 *        unit number to the FILE pointer
 *
 *      dxp_close_file: closes a file opened with dxp_open_file
 *
 *      dxp_get_electronics: return number of DXP modules, list of CAMAC I/O 
 *         addresses and used DXP Channels.
 *
 *      dxp_get_detectors: returns number of DXP modules and list of detector
 *         channel numbers
 *
 *      dxp_modify_gains: scales the DXP amplifier gains by a factor for all 
 *         channels.
 *
 *      dxp_modify_one_gains: scales the DXP amplifier gains by a factor for 
 *         a single channel.
 *
 *      dxp_initialize_ASC: reads in the PREAMP configuration file that
 *         contains channel dependant preamp parameters and adjusts ASC
 *         DAC values (FINEGAIN,COARSEGAIN,OFFDACVAL,DACPERADC,TRACKRESET)
 *         based on these values
 *
 *      dxp_det_to_elec: utility to return CAMAC I/O address and DXP channel
 *         number for a given detector element.
 *
 *      dxp_elec_to_det: utility to return the detector number for a given
 *         CAMAC I/O address and DXP channel.
 *
 *		dxp_version: return version information for this library.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <ctype.h>
#include <limits.h>
#include <math.h> 
/* General DXP information */
#include <xerxes_structures.h>
#include <xia_xerxes_structures.h>
#include <xia_xerxes.h>
#include <xerxes_errors.h>
#include <xia_handel.h>
#include <handel_errors.h>

static char info_string[132],line[132],*token,*delim=" ,=\t\r\n";
/* Debug mode? */
/*static int debug=DXP_DEBUG;*/
static int debug=DXP_SUCCESS;
/* Register definition for zero and one */
static int zero=0;
/* Shorthand notation telling routines to act on all channels of the DXP (-1 currently). */
static int allChan=ALLCHAN;
/* Maximum number of files that can be open at a time */
#define MAXFILE 10
static FILE *filpnt[MAXFILE];
static int  dxpopen=0;

static int numDxpMod=0;
static int numDxpChan=0;
/* 
 * Define the utility routines used throughout this library
 */
static DXP_MD_ERROR_CONTROL dxp_md_error_control;
static DXP_MD_ERROR dxp_md_error;
static DXP_MD_ALLOC dxp_md_alloc;
static DXP_MD_FREE dxp_md_free;
static DXP_MD_PUTS dxp_md_puts;
static DXP_MD_WAIT dxp_md_wait;
/*
 *  Head of Linked list of Interfaces
 */
static Interface *iface_head = NULL;
/*
 *  Structure containing pointers to machine dependent routines
 */
static Utils *utils= NULL;
/*
 *  Structure to contain the board types in the system
 */
static Board_Info *btypes_head = NULL;
/*
 *	Define the Head of the linked list of items in the system
 */
static Board *system_head = NULL;  
/*
 *	Define a global structure for system level information
 */
static System_Info *info = NULL;  
/*
 *	Define the head of the DSP linked list
 */
static Dsp_Info *dsp_head = NULL;  
/*
 *	Define the head of the FIPPI linked list
 */
static Fippi_Info *fippi_head = NULL;  
/*
 *	Define the head of the DSP parameter defaults linked list
 */
static Dsp_Defaults *defaults_head = NULL;  

/*
 * Define global static variables to hold the information 
 * used to assign new board information.
 */
static Board_Info *working_btype = NULL;  
static Board *working_board = NULL;  
static Interface *working_iface = NULL;  
static Dsp_Info *working_dsp = NULL;  
static Fippi_Info *working_fippi = NULL;  
static Dsp_Defaults *working_defaults = NULL;  

/* Define variables for the HanDeL layer */

/*
 *	Define the head of the Detectors list
 */
static Detector *xiaDetectorsHead = NULL;  
/*
 *	Define the head of the Firmware Sets LL
 */
static FirmwareSet *xiaFirmwareSetHead = NULL;  
/*
 *	Define the head of the XiaDefaults LL
 */
static XiaDefaults *xiaDefaults = NULL;  


static Detector *workingDetector = NULL;
static FirmwareSet *workingFirmwareSet = NULL;
static XiaDefaults *workingXiaDefaults = NULL;


/********------******------*****------******-------******-------******------*****
 * Routines to perform global initialization functions.  Read in configuration 
 * files, download data to all modules, etc...
 *
 ********------******------*****------******-------******-------******------*****/

/******************************************************************************
 *
 * Routine to initial data structures and pointers for the XerXes 
 * library.
 *
 ******************************************************************************/
XERXES_EXPORT int XERXES_API dxp_init_library(void)
{
	int status=DXP_SUCCESS;

/* First thing is to load the utils structure.  This is needed 
 * to call allocation of any memory in general as well as any
 * error reporting.  It is absolutely essential that this work
 * with no errors.  */
 
/* Install the utility functions of the XerXes Library */
	if((status=dxp_install_utils("default"))!=DXP_SUCCESS){
		sprintf(info_string,"Unable to install utilities in XerXes");
		printf("%s\n",info_string);
		return status;
	}

/* Initialize the global system */
	if((status=dxp_init_ds())!=DXP_SUCCESS){
		sprintf(info_string,"Unable to initialize data structures.");
		dxp_md_error("dxp_init_library",info_string,&status);
		return status;
	}

	return status;
}
	
/******************************************************************************
 *
 * Global initialization routine.  Should be called at the start of all
 * programs.
 *
 ******************************************************************************/
XERXES_EXPORT int XERXES_API dxp_initialize(void)
{
	int status=DXP_SUCCESS;

/* First thing is to load the utils structure.  This is needed 
 * to call allocation of any memory in general as well as any
 * error reporting.  It is absolutely essential that this work
 * with no errors.  */
 
/* Initialize the data structures and pointers of the library */
	if((status=dxp_init_library())!=DXP_SUCCESS){
		sprintf(info_string,"Unable to initialize XerXes");
		printf("%s\n",info_string);
		return status;
	}

/* Read the configuration file */
	if((status=dxp_read_config("XIA_CONFIG"))!=DXP_SUCCESS){
		sprintf(info_string,"Error reading configuration file.");
		dxp_md_error("dxp_initialize",info_string,&status);
		return status;
	}

/* Setup the modules by assigning them IO and module channel numbers */
    if ((status=dxp_assign_channel())!=DXP_SUCCESS){
        status = DXP_INITIALIZE;
        dxp_md_error("dxp_initialize","Could not assign channels for this config.",&status);
        return status;
    }

/* Setup the modules by assigning them IO and module channel numbers */
    if ((status=dxp_user_setup())!=DXP_SUCCESS){
        status = DXP_INITIALIZE;
        dxp_md_error("dxp_initialize",
			"Could not perform user setup(downloading and calibrations)",&status);
        return status;
    }

	return status;
}

/******************************************************************************
 *
 * Routine to install pointers to the machine dependent
 * utility routines.
 *
 ******************************************************************************/
XERXES_EXPORT int XERXES_API dxp_install_utils(const char *utilname)
{
	Xia_Util_Functions util_funcs;

/* Call the MD layer init function for the utility routines */
	dxp_md_init_util(&util_funcs, (char *) utilname);
/* Now build up some information...slowly */
	dxp_md_alloc = util_funcs.dxp_md_alloc;
	dxp_md_free = util_funcs.dxp_md_free;
	dxp_md_error_control = util_funcs.dxp_md_error_control;
	dxp_md_error = util_funcs.dxp_md_error;
	dxp_md_wait = util_funcs.dxp_md_wait;
	dxp_md_puts = util_funcs.dxp_md_puts;

/* Now for the Utils structure */
	if (utils!=NULL) {
		dxp_md_free(utils->dllname);
		dxp_md_free(utils->funcs);
		dxp_md_free(utils);
	}
/* Allocate and assign memory for the Utils structure */
	utils = (Utils *) dxp_md_alloc(sizeof(Utils));
	utils->dllname = (char *) dxp_md_alloc((strlen(utilname)+1)*sizeof(char));
	strcpy(utils->dllname, utilname);
	utils->funcs = (Xia_Util_Functions *) dxp_md_alloc(sizeof(Xia_Util_Functions));
	utils->funcs->dxp_md_alloc = dxp_md_alloc;
	utils->funcs->dxp_md_free = dxp_md_free;
	utils->funcs->dxp_md_error_control = dxp_md_error_control;
	utils->funcs->dxp_md_error = dxp_md_error;
	utils->funcs->dxp_md_wait = dxp_md_wait;
	utils->funcs->dxp_md_puts = dxp_md_puts;

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Initialize the data structures to an empty state
 *
 ******************************************************************************/
XERXES_EXPORT int XERXES_API dxp_init_ds(void)
{
	int status = DXP_SUCCESS;

	Board_Info *current = btypes_head;
	Board_Info *next = NULL;

/* CLEAR BTYPES LINK LIST */
	
/* First make sure there is not an allocated btypes_head */
	while (current != NULL) {
		next = current->next;
/* Clear the memory first */
		dxp_free_binfo(current);
		current = next;
	}
/* And clear out the pointer to the head of the list(the memory is freed) */
	btypes_head = NULL;
	
/* NULL the current board type in the static variable, to ensure we dont point 
 * free memory */
	working_btype = NULL;
	
/* CLEAR INFO STRUCTURE */
	
/* Make sure the info structure is empty */
	if (info!=NULL) {
		if (info->preamp  != NULL) dxp_md_free(info->preamp);
		if (info->modules != NULL) dxp_md_free(info->modules);
		dxp_md_free(info);
	}
/* Now allocate memory for the system level information */
	info = (System_Info*) dxp_md_alloc(sizeof(System_Info));
	info->preamp = NULL;
	info->modules= NULL;
	info->btypes = NULL;
	info->utils = utils;
	info->status = 0;

/* CLEAR BOARDS,DSP,FIPPI,DEFAULTS LINK LIST */
	if((status=dxp_init_boards_ds())!=DXP_SUCCESS){
        sprintf(info_string,"Failed to initialize the boards structures");
        dxp_md_error("dxp_init_ds",info_string,&status);
        return status;
    }

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Initialize the BOARD data structures to an empty state
 *
 ******************************************************************************/
XERXES_EXPORT int XERXES_API dxp_init_boards_ds(void)
{
	int status=DXP_SUCCESS;
	
	Board *next;
	Board *current = system_head;

/* Initialize the system_head linked list */
	if (current!=NULL) {
/* Clear out the linked list...deallocating all memory */
		do {
			next = current->next;
			dxp_free_board(current);
			current = next;
		} while (next!=NULL);
		system_head = NULL;
	}

/* NULL the current board in the static variable, to ensure we dont point 
 * free memory */
	working_board = NULL;
	
/* CLEAR INTERFACE LINK LIST */
	if((status=dxp_init_iface_ds())!=DXP_SUCCESS){
        sprintf(info_string,"Failed to initialize the Interface structures");
        dxp_md_error("dxp_init_ds",info_string,&status);
        return status;
    }
	
/* CLEAR DSP LINK LIST */
	if((status=dxp_init_dsp_ds())!=DXP_SUCCESS){
        sprintf(info_string,"Failed to initialize the dsp structures");
        dxp_md_error("dxp_init_boards_ds",info_string,&status);
        return status;
    }

/* CLEAR FIPPI LINK LIST */
	if((status=dxp_init_fippi_ds())!=DXP_SUCCESS){
        sprintf(info_string,"Failed to initialize the fippi structures");
        dxp_md_error("dxp_init_boards_ds",info_string,&status);
        return status;
    }

/* CLEAR DEFAULTS LINK LIST */
	if((status=dxp_init_defaults_ds())!=DXP_SUCCESS){
        sprintf(info_string,"Failed to initialize the defaults structures");
        dxp_md_error("dxp_init_boards_ds",info_string,&status);
        return status;
    }

/* Initialize the number of modules and channels */
	numDxpMod = 0;
    numDxpChan= 0;

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Initialize the INTERFACE data structures to an empty state
 *
 ******************************************************************************/
static int XERXES_API dxp_init_iface_ds(VOID)
{
	Interface *next;
	Interface *current = iface_head;

/* Search thru the iface LL and clear them out */
	while (current != NULL) {
		next = current->next;
/* Clear the memory first */
		dxp_free_iface(current);
		current = next;
	}
/* And clear out the pointer to the head of the list(the memory is freed) */
	iface_head = NULL;
	
/* NULL the current board type in the static variable, to ensure we dont point 
 * free memory */
	working_iface = NULL;

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Initialize the DSP data structures to an empty state
 *
 ******************************************************************************/
static int XERXES_API dxp_init_dsp_ds(void)
{
	Dsp_Info *next;
	Dsp_Info *current = dsp_head;

/* Search thru the iface LL and clear them out */
	while (current != NULL) {
		next = current->next;
/* Clear the memory first */
		dxp_free_dsp(current);
		current = next;
	}
/* And clear out the pointer to the head of the list(the memory is freed) */
	dsp_head = NULL;
	
/* NULL the current dsp in the static variable, to ensure we dont point 
 * free memory */
	working_dsp = NULL;

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Initialize the FIPPI data structures to an empty state
 *
 ******************************************************************************/
static int XERXES_API dxp_init_fippi_ds(void)
{
	Fippi_Info *next;
	Fippi_Info *current = fippi_head;

/* Search thru the iface LL and clear them out */
	while (current != NULL) {
		next = current->next;
/* Clear the memory first */
		dxp_free_fippi(current);
		current = next;
	}
/* And clear out the pointer to the head of the list(the memory is freed) */
	fippi_head = NULL;
	
/* NULL the current fippi in the static variable, to ensure we dont point 
 * free memory */
	working_fippi = NULL;

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Initialize the DEFAULTS data structures to an empty state
 *
 ******************************************************************************/
static int XERXES_API dxp_init_defaults_ds(void)
{
	Dsp_Defaults *next;
	Dsp_Defaults *current = defaults_head;

/* Search thru the iface LL and clear them out */
	while (current != NULL) {
		next = current->next;
/* Clear the memory first */
		dxp_free_defaults(current);
		current = next;
	}
/* And clear out the pointer to the head of the list(the memory is freed) */
	defaults_head = NULL;
	
/* NULL the current defaults in the static variable, to ensure we dont point 
 * free memory */
	working_defaults = NULL;

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Read the global configuration file.
 *
 ******************************************************************************/
int XERXES_API dxp_read_config(char *cname)
{
	int status=DXP_SUCCESS;
	int len, start, finish;
	unsigned int i;

	FILE *fp;
    char *cstatus=NULL;
	char *filename;
	char newFile[MAXFILENAME_LEN];

	char *values[1];

/* Find the file that contains the XIA system configuration information */
    if((fp=dxp_find_file(cname,"r",newFile))==NULL){
/* Try to find the file by double indirection....just in case */
		if ((filename=getenv(cname))!=NULL) {
			if((fp=dxp_find_file(filename,"r",newFile))==NULL){ 
/* Could not find the XIA_CONFIG file! */
			    status = DXP_OPEN_FILE;
			    sprintf(info_string,
					"Can not find Config File: %s", filename);
			    dxp_md_error("dxp_read_config",info_string,&status);
				return status;
			}
			fclose(fp);
		} else {
			status = DXP_OPEN_FILE;
			sprintf(info_string,
				"Could not find the environment variable: %s", cname);
			dxp_md_error("dxp_read_config",info_string,&status);
			return status;
		}
	}

/*
 * Count the number of board types in the configuration file.  Look
 * for a 'dxp' string at the beginning of the line, this signifies
 * a new board type definition.  Use this count to allocate memory.
 */ 
	for(i=0;;i++){ 
		do {
			cstatus = fgets(line,132,fp); 
		} while (((line[0]=='*')||(line[0]=='\r')||(line[0]=='\n'))&&(cstatus!=NULL)); 
		if(cstatus==NULL) break;
/* What kind of entry do we have?
 * Supported types:  dxp* = new board type
 *				     library = DLL for the most recent board type
 */
/* Get the length of the line */
		len = strlen(line);
/* Find the first token deliminated by a space, this is the entry type */
		token = strtok(line, delim);
/* Unfort, need to allow spaces in the filenames.
 * so count forward in the line till out of spaces, this is the start of the 
 * file name */
		for(start=strlen(token)+1;start<len;start++) 
			if(isspace(line[start])==0) break;				/* find first character in path */ 
/* Now count backwards from the end of the line till you hit a non-space */
		for(finish=len;finish>start;finish--) 
			if(isgraph(line[finish])!=0) break;				/* find last character in name  */ 
/* Allocate memory for the 2nd token */
		len = finish-start+1;
/* Now copy the value associated with this item */
		values[0] = (char *) dxp_md_alloc((len+1)*sizeof(char));
/* Copy the value into the array. End the string with a NULL character. */
		strncpy(values[0], line+start, len);
		values[0][len] = '\0';

/* Add this item to the system information */
	    if ((status=dxp_add_system_item(token,values))!=DXP_SUCCESS){
			status = DXP_INITIALIZE;
			sprintf(info_string,"Problems adding item: %s",token);
			dxp_md_error("dxp_read_config",info_string,&status);
			return status;
		}
	}

	return status;
}

/******************************************************************************
 *
 * Add an item specified by token to the system configuration.
 *
 ******************************************************************************/
int XERXES_API dxp_add_system_item(char *token, char **values)
{
	int status=DXP_SUCCESS;
	int len;
	unsigned int j;

/* Now Proceed to process the entry */

/* Get the length of the value parameters used by this routine */
	len = strlen(values[0]);

/* Convert the token to lower case for comparison */
	for (j=0;j<strlen(token);j++) token[j] = (char) tolower(token[j]);
	if (strncmp(token,"dxp",3)==0) {
/* Load function pointers for local use */
		if ((status=dxp_add_btype(token, values[0], NULL))!=DXP_SUCCESS) {
			sprintf(info_string,"Problems adding board type %s",token);
			dxp_md_error("dxp_add_system_item",info_string,&status);
			return DXP_SUCCESS;
		}
	} else if (strcmp(token,"preamp")==0) {
/* Now copy the filename for the preamp configuration file */
		info->preamp = (char *) dxp_md_alloc((len+1)*sizeof(char));
/* Copy the filename into the structure.  End the string with 
 * a NULL character. */
		info->preamp = strncpy(info->preamp, values[0], len);
		info->preamp[len] = '\0';
	} else if (strcmp(token,"modules")==0) {
/* Now copy the filename for the modules configuration file */
		info->modules = (char *) dxp_md_alloc((len+1)*sizeof(char));
/* Copy the filename into the structure.  End the string with 
 * a NULL character. */
		info->modules = strncpy(info->modules, values[0], len);
		info->modules[len] = '\0';
	} else {
	    status = DXP_INPUT_UNDEFINED;
	    sprintf(info_string,"Unable to add token: %s", token);
	    dxp_md_error("dxp_add_system_item",info_string,&status);
		return status;
	}

	return status;
}

/******************************************************************************
 *
 * Add an item specified by token to the board configuration.
 *
 ******************************************************************************/
int XERXES_API dxp_add_board_item(char *token, char **values)
{
	int status=DXP_SUCCESS;
	unsigned int len, j;
	unsigned int nchan;

	unsigned short used;
	int ioChan,chanid;
	int *detChan;
	int found;

/* Get the length of the values parameter used by this routine */
	len = strlen(values[0]);

/* Make sure the token is all lower case */
	for (j=0;j<strlen(token);j++) token[j] = (char) tolower(token[j]);
/* 
 * Now Identify the entry type 
 */
	if (strcmp(token, "board_type")==0) {
		for (j=0;j<len;j++) values[0][j] = (char) tolower(values[0][j]);
/* Search thru board types to match find the structure 
   in the linked list */
		working_btype = btypes_head;
		found = 1;
		if (working_btype!=NULL) {
			found = strcmp(working_btype->name,values[0]);
		}
		while ((found!=0)&&(working_btype!=NULL)) {			
			working_btype = working_btype->next;
			if (working_btype!=NULL) {
				found = strcmp(working_btype->name,values[0]);
			}
		} 
		if (working_btype==NULL) {
	        status = DXP_UNKNOWN_BTYPE;
			sprintf(info_string,"Board type %s not recognized",	values[0]);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}

/* Define the interface tag, this will typically be the information 
 * that the DXP_MD_INITIALIZE() will use to determine what MD routines
 * are needed */
	} else if (strcmp(token, "interface")==0) {
/* Initialize the library for each Interface, if we have defined an IO tag */
		if ((values[0]==NULL)||(values[1]==NULL)) {
			status = DXP_INITIALIZE;
			sprintf(info_string,"Must define an interface type and tag");
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}
/* Add the interface */
		for (j=0;j<strlen(values[0]);j++) values[0][j] = (char) tolower(values[0][j]);
		if((status=dxp_add_iface(values[0], values[1], &working_iface))!=DXP_SUCCESS) {
            sprintf(info_string,"Unable to load Interface %s:%s",values[0],values[1]);
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
/* Got Interface, add to current board, or ignore until a module is defined */
		if (working_board!=NULL) {
			working_board->iface = working_iface;
		}

/* Is this a new module in the system */
	} else if (strcmp(token, "module")==0) {
/* Pack an information string for debugging purposes and error reporting */
/* Make sure we haven't tried to allocate too many modules */
        sprintf(info_string," mod %d: iface id:%s, #channels:%s",
			numDxpMod,values[0],values[1]);
		if(numDxpMod == MAXDXP){
			sprintf(info_string,"Too many modules in config:MAXDXP = %i",
				MAXDXP);
			status = DXP_ARRAY_TOO_SMALL;
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}
		dxp_md_error("dxp_add_board_items",info_string,&debug);
/* Check to make sure we have a board type defined */
		if (working_btype== NULL) {
			sprintf(info_string,"There is no board type defined");
			status = DXP_UNKNOWN_BTYPE;
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}

/* Set the NumDxpMod'th to have none of the channels used */
		used = 0;

/* Pull out the information for where in the system the module sits 
 * and then open the crate via the dxp_md_open() routine.  This routine is 
 * expected to return a number for ioChan that identifies this module's
 * location in the array used by the machine dependent functions to 
 * identify how to talk to this module. */
		if (working_iface==NULL) {
	        status = DXP_INITIALIZE;
            sprintf(info_string,"You must define an interface before any modules");
	        dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}
		if((status=working_iface->funcs->dxp_md_open(values[0],&ioChan))!=DXP_SUCCESS){
            sprintf(info_string,"Could not open Module %s",values[0]);
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }

/* Now loop over the next MAXCHAN numbers to define a detector element number
 * associated with each DXP channel.  If positive then the channel is 
 * considered instrumented and the number is stored in the Used[] array
 * for future reference.  Else if the number is negative or 0, the channel
 * is ignored */
        
		nchan = (unsigned int) strtol(values[1],NULL,10);
		detChan = (int *) dxp_md_alloc(nchan*sizeof(int));
		for(j=0;j<nchan;j++){
			chanid = (int) strtol(values[j+2],NULL,10);
			if (chanid>=0) {
				used |= 1<<j;
				detChan[j] = chanid;
				numDxpChan++;
			} else {
				detChan[j] = -1;
			}
        }

/* Find the last entry in the Linked list and add to the end */
		working_board = system_head;
		if (working_board!=NULL) {
			while (working_board->next!=NULL) {
				working_board = working_board->next;
			}
			working_board->next = (Board *) dxp_md_alloc(sizeof(Board));
			working_board = working_board->next;
		} else {
			system_head = (Board *) dxp_md_alloc(sizeof(Board));
			working_board = system_head;
		}
		working_board->ioChan= ioChan;
		working_board->used	= used;
/* Allocate memory for DetChan array */	
		working_board->detChan = (int *) dxp_md_alloc(nchan*sizeof(int));
		memcpy(working_board->detChan, detChan, nchan*sizeof(int));
/* Free Memory before i forget */
		dxp_md_free(detChan);

/* more assignments */
		working_board->mod	= numDxpMod;
		working_board->nchan	= nchan;
		working_board->btype	= working_btype;
		working_board->iface	= working_iface;
/* Allocate memory for the iostring containing slot information, etc... */
		working_board->iostring = (char *)
			dxp_md_alloc((strlen(values[0])+1)*sizeof(char));
		strcpy(working_board->iostring, values[0]);
		memset(working_board->state, 0, sizeof(working_board->state));
/* 
 * Initialize the interface.  Note that if multiple boards use the same
 * Interface, then this call is redundant on 2nd, etc...instances...but, 
 * make the call anyway, in case in the future we wish to enable different
 * interfaces for each board rather than the current limitation of essentially
 * one interface per board type...there are many implications to implementing 
 * this feature though...
 */
		working_board->btype->funcs->dxp_init_driver(working_iface);
/* This is the end of the linked list */
		working_board->next	= NULL;

/* More memory assignments */
		working_board->params = (unsigned short **) 
			dxp_md_alloc(nchan*sizeof(unsigned short *));
		for (j=0;j<working_board->nchan;j++) working_board->params[j] = NULL;
		working_board->dsp = (Dsp_Info **) 
			dxp_md_alloc(nchan*sizeof(Dsp_Info *));
		working_board->fippi= (Fippi_Info **) 
			dxp_md_alloc(nchan*sizeof(Fippi_Info *));
		working_board->defaults = (Dsp_Defaults **) 
			dxp_md_alloc(nchan*sizeof(Dsp_Defaults *));
		working_board->preamp = (Preamp_Info *) 
			dxp_md_alloc(nchan*sizeof(Preamp_Info));
/* Fill in some default information */
		for (j=0;j<working_board->nchan;j++) {
			if ((working_board->used)&&(1<<j)) {
/* Load the dsp structure */
				working_board->dsp[j] = working_dsp;
/* Load the fippi structure */
				working_board->fippi[j] = working_fippi;
/* Load the DSP defaults structure */
				working_board->defaults[j] = working_defaults;
/* Allocate memory for the params array, and zero the entries */
				if (working_dsp!=NULL) {
					working_board->params[j] = (unsigned short *) 
						dxp_md_alloc((working_board->dsp[j]->params->nsymbol)*sizeof(unsigned short));
					memset(working_board->params[j], 0, 
						working_board->dsp[j]->params->nsymbol*sizeof(unsigned short));
				}
			} else {
				working_board->dsp[j] = NULL;
				working_board->params[j] = NULL;
				working_board->fippi[j] = NULL;
				working_board->defaults[j] = NULL;
			}
				
		}

/* We seem to have a new module in the system...increment the global counter */
		numDxpMod++;
/* We need to define a new DSP default */
	} else if (strcmp(token, "default_dsp")==0) {
/* Make sure a board type is defined */
		if(working_btype==NULL){
			status = DXP_UNKNOWN_BTYPE;
            sprintf(info_string,"No board type defined, can not proceed");
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
/* Now load the configuration */
		if((status=dxp_add_dsp(values[0], working_btype, &working_dsp))!=DXP_SUCCESS) {
            sprintf(info_string,"Unable to load dsp %s",values[0]);
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
/* Now reload all the DSP pointers for the current board */
		if (working_board!=NULL) {
			for (j=0;j<working_board->nchan;j++) {
				working_board->dsp[j] = working_dsp;
/* Check if we are overwriting a previous DSP definition, if so, free the memory */
				if (working_board->params[j]!=NULL) dxp_md_free(working_board->params[j]);
/* Allocate memory for the params array */
				working_board->params[j] = (unsigned short *) 
					dxp_md_alloc((working_board->dsp[j]->params->nsymbol)*sizeof(unsigned short));
				memset(working_board->params[j], 0, working_board->dsp[j]->params->nsymbol*sizeof(unsigned short));
			}
		}
/* We need to define a new DSP default */
	} else if (strcmp(token, "dsp")==0) {
/* Make sure a board type is defined */
		if(working_btype==NULL){
			status = DXP_UNKNOWN_BTYPE;
            sprintf(info_string,"No board type defined, can not proceed");
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
/* Get the next token, it is the channel Number */
		chanid = (int) strtol(values[0],NULL,10);
		if ((chanid<0)||(chanid>=(int) working_board->nchan)) {
			status = DXP_BADCHANNEL;
            sprintf(info_string,"Channel #%i out of range", chanid);
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
		}

/* Now load the configuration */
		if((status=dxp_add_dsp(values[1], working_btype, 
						&(working_board->dsp[chanid])))!=DXP_SUCCESS) {
			sprintf(info_string,"Unable to locate %s",values[1]);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}
/* Check if we are overwriting a previous DSP definition, if so, free the memory */
		if (working_board->params[chanid]!=NULL) dxp_md_free(working_board->params[chanid]);
/* Allocate memory for the params array */
		working_board->params[chanid] = (unsigned short *) 
			dxp_md_alloc((working_board->dsp[chanid]->params->nsymbol)*sizeof(unsigned short));
		memset(working_board->params[chanid], 0, working_board->dsp[chanid]->params->nsymbol*sizeof(unsigned short));
/* We need to define a new FIPPI default */
	} else if (strcmp(token, "default_fippi")==0) {
/* Make sure a board type is defined */
		if(working_btype==NULL){
			status = DXP_UNKNOWN_BTYPE;
            sprintf(info_string,"No board type defined, can not proceed");
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
/* Now load the configuration */
		if((status=dxp_add_fippi(values[0], working_btype,
				&working_fippi))!=DXP_SUCCESS) {
			sprintf(info_string,"Unable to locate %s",values[0]);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}
/* Now reload all the FIPPI pointers for the current board type */
		if (working_board!=NULL) {
			for (j=0;j<working_board->nchan;j++)
				working_board->fippi[j] = working_fippi;
		}
/* We need to define a new FIPPI default */
	} else if (strcmp(token, "fippi")==0) {
/* Make sure a board type is defined */
		if(working_btype==NULL){
			status = DXP_UNKNOWN_BTYPE;
            sprintf(info_string,"No board type defined, can not proceed");
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
		chanid = (int) strtol(values[0],NULL,10);
		if ((chanid<0)||(chanid>=(int) working_board->nchan)) {
			status = DXP_BADCHANNEL;
			sprintf(info_string,"Channel #%i out of range", chanid);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}

/* Now load the configuration */
		if((status=dxp_add_fippi(values[1], working_btype,
				&(working_board->fippi[chanid])))!=DXP_SUCCESS) {
			sprintf(info_string,"Unable to locate %s",values[1]);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}
/* We need to define a new parameter default */
	} else if (strcmp(token, "default_param")==0) {
/* Make sure a board type is defined */
		if(working_btype==NULL){
			status = DXP_UNKNOWN_BTYPE;
            sprintf(info_string,"No board type defined, can not proceed");
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
/* Now load the configuration */
		if((status=dxp_add_defaults(values[0], working_btype,
				&working_defaults))!=DXP_SUCCESS) {
			sprintf(info_string,"Unable to locate %s",values[0]);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}
		for (j=0;j<working_board->nchan;j++) working_board->defaults[j] = working_defaults;
/* We need to define a new FIPPI default */
	} else if (strcmp(token, "param")==0) {
/* Make sure a board type is defined */
		if(working_btype==NULL){
			status = DXP_UNKNOWN_BTYPE;
            sprintf(info_string,"No board type defined, can not proceed");
            dxp_md_error("dxp_add_board_item",info_string,&status);
            return status;
        }
		chanid = (int) strtol(values[0],NULL,10);
		if ((chanid<0)||(chanid>=(int) working_board->nchan)) {
			status = DXP_BADCHANNEL;
			sprintf(info_string,"Channel #%i out of range", chanid);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
		}

/* Now load the configuration */
		if((status=dxp_add_defaults(values[1], working_btype,
				&(working_board->defaults[chanid])))!=DXP_SUCCESS) {
			sprintf(info_string,"Unable to locate %s",values[1]);
			dxp_md_error("dxp_add_board_item",info_string,&status);
			return status;
	       }
	} else if (strcmp(token, "end")==0) {
/* Clear out the working structures, to help protect ourselves from mistakes */
		working_board = NULL;
		working_btype = NULL;
		working_dsp = NULL;
		working_fippi = NULL;
		working_iface = NULL;
		working_defaults = NULL;
		return DXP_SUCCESS;
	} else {
		sprintf(info_string,"Unrecognized item to add to board definitions: %s",token);
		dxp_md_error("dxp_add_board_item",info_string,&status);
		return status;
	}

	return status;
}

/******************************************************************************
 *
 * Routine to open a new file.  
 * Try to open the file directly first.
 * Then try to open the file in the directory pointed to 
 *     by XIAHOME.
 * Finally try to open the file as an environment variable.
 *
 ******************************************************************************/
static FILE* XERXES_API dxp_find_file(const char* filename, const char* mode, char newFile[MAXFILENAME_LEN])
/* const char *filename;			Input: filename to open			*/
/* const char *mode;				Input: Mode to use when opening	*/
/* char *newFile;					Output: Full filename of file (translated env vars)	*/
{
	FILE *fp=NULL;
	char *name=NULL, *name2=NULL;
	char *home=NULL;
	
	unsigned int len = 0;

/* Try to open file directly */
	if((fp=fopen(filename,mode))!=NULL){
		len = MAXFILENAME_LEN>(strlen(filename)+1) ? strlen(filename) : MAXFILENAME_LEN;
		strncpy(newFile, filename, len);
		newFile[len] = 0;
		return fp;
	}
/* Try to open the file with the path XIAHOME */
	if ((home=getenv("XIAHOME"))!=NULL) {
		name = (char *) dxp_md_alloc(sizeof(char)*
							(strlen(home)+strlen(filename)+2));
		sprintf(name, "%s/%s", home, filename);
		if((fp=fopen(name,mode))!=NULL){
			len = MAXFILENAME_LEN>(strlen(name)+1) ? strlen(name) : MAXFILENAME_LEN;
			strncpy(newFile, name, len);
			newFile[len] = 0;
			dxp_md_free(name);
			return fp;
		}
		dxp_md_free(name);
		name = NULL;
	}
/* Try to open the file with the path DXPHOME */
	if ((home=getenv("DXPHOME"))!=NULL) {
		name = (char *) dxp_md_alloc(sizeof(char)*
							(strlen(home)+strlen(filename)+2));
		sprintf(name, "%s/%s", home, filename);
		if((fp=fopen(name,mode))!=NULL){
			len = MAXFILENAME_LEN>(strlen(name)+1) ? strlen(name) : MAXFILENAME_LEN;
			strncpy(newFile, name, len);
			newFile[len] = 0;
			dxp_md_free(name);
			return fp;
		}
		dxp_md_free(name);
		name = NULL;
	}
/* Try to open the file as an environment variable */
	if ((name=getenv(filename))!=NULL) {
		if((fp=fopen(name,mode))!=NULL){
			len = MAXFILENAME_LEN>(strlen(name)+1) ? strlen(name) : MAXFILENAME_LEN;
			strncpy(newFile, name, len);
			newFile[len] = 0;
			return fp;
		}
		name = NULL;
	}
/* Try to open the file with the path XIAHOME and pointing 
 * to a file as an environment variable */
	if ((home=getenv("XIAHOME"))!=NULL) {
		if ((name2=getenv(filename))!=NULL) {
		
			name = (char *) dxp_md_alloc(sizeof(char)*
								(strlen(home)+strlen(name2)+2));
			sprintf(name, "%s/%s", home, name2);
			if((fp=fopen(name,mode))!=NULL) {
				len = MAXFILENAME_LEN>(strlen(name)+1) ? strlen(name) : MAXFILENAME_LEN;
				strncpy(newFile, name, len);
				newFile[len] = 0;
				dxp_md_free(name);
				return fp;
			}
			dxp_md_free(name);
			name = NULL;
		}
	}
/* Try to open the file with the path DXPHOME and pointing 
 * to a file as an environment variable */
	if ((home=getenv("DXPHOME"))!=NULL) {
		if ((name2=getenv(filename))!=NULL) {
		
			name = (char *) dxp_md_alloc(sizeof(char)*
								(strlen(home)+strlen(name2)+2));
			sprintf(name, "%s/%s", home, name2);
			if((fp=fopen(name,mode))!=NULL) {
				len = MAXFILENAME_LEN>(strlen(name)+1) ? strlen(name) : MAXFILENAME_LEN;
				strncpy(newFile, name, len);
				newFile[len] = 0;
				dxp_md_free(name);
				return fp;
			}
			dxp_md_free(name);
			name = NULL;
		}
	}

	return NULL;
}

/******************************************************************************
 *
 * Routine to return the list of system files currently in use.
 * These filenames have decoded any environment variables used and included
 * them in the path names.
 *
 * The variable maxlen is the maximum length of a filename, this ensures that 
 * the routine will not overwrite memory space that it doesnt have access to
 * 
 * files[0] = XIA_CONFIG file
 * files[1] = modules configuration file
 * files[2] = preamp configuration file
 *
 ******************************************************************************/
int XERXES_API dxp_locate_system_files(unsigned int *maxlen, char **files)
/* unsigned short *maxlen;		Input: Maximum length of allocated memory	*/
/* char *files;					Output: List of filenames used by XerXes	*/
{
	int status = DXP_SUCCESS;
	FILE *fp=NULL;
	char *filename=NULL;
	char newFile[200];

	unsigned int index = 0;
	unsigned int len = 0;
	unsigned int start = 0;

/* Try to find the system configuration file	*/
	if ((filename=getenv("XIA_CONFIG"))!=NULL) {
	    if((fp=dxp_find_file(filename,"r",newFile))==NULL){ 
			status = DXP_OPEN_FILE;
		    sprintf(info_string,"Unable to locate the XIA_CONFIG File");
			dxp_md_error("dxp_locate_files",info_string,&status);
		} else {
			fclose(fp);
			if (strlen(newFile)>(*maxlen+1)) {
				len = *maxlen;
				start = strlen(newFile) - *maxlen;
			} else {
				len = strlen(newFile);
				start = 0;
			}
			strncpy(files[index], newFile+start, len);
			files[index][len] = 0;
			index++;
		}
	} else {
		status = DXP_OPEN_FILE;
		sprintf(info_string,"Unable to locate the XIA_CONFIG environment variable");
		dxp_md_error("dxp_locate_files",info_string,&status);
	}

/* Try to find the modules configuration file	*/
    if((fp=dxp_find_file(info->modules,"r",newFile))==NULL){ 
        status = DXP_OPEN_FILE;
	    sprintf(info_string,"Unable to locate the Modules configuration File");
		dxp_md_error("dxp_locate_files",info_string,&status);
	} else {
		fclose(fp);
		if (strlen(newFile)>(*maxlen+1)) {
			len = *maxlen;
			start = strlen(newFile) - *maxlen;
		} else {
			len = strlen(newFile);
			start = 0;
		}
		strncpy(files[index], newFile+start, len);
		files[index][len] = 0;
		index++;
	}

/* Try to find the preamp configuration file	*/
    if((fp=dxp_find_file(info->preamp,"r",newFile))==NULL){ 
        status = DXP_OPEN_FILE;
	    sprintf(info_string,"Unable to locate the preamp configuration File");
		dxp_md_error("dxp_locate_files",info_string,&status);
	} else {
		fclose(fp);
		if (strlen(newFile)>(*maxlen+1)) {
			len = *maxlen;
			start = strlen(newFile) - *maxlen;
		} else {
			len = strlen(newFile);
			start = 0;
		}
		strncpy(files[index], newFile+start, len);
		files[index][len] = 0;
		index++;
	}
	
	return status;
}

/******************************************************************************
 *
 * Routine to return the list of channel configuration files currently in use.
 * These filenames have decoded any environment variables used and included
 * them in the path names.
 *
 * The variable maxlen is the maximum length of a filename, this ensures that 
 * the routine will not overwrite memory space that it doesnt have access to
 * 
 * files[0] = fippi firmware file
 * files[1] = dsp firmware file
 * files[2] = default dsp parameter file
 *
 ******************************************************************************/
int XERXES_API dxp_locate_channel_files(int *detChan, unsigned int *maxlen, char **files)
/* int *detChan;				Input: Detector Channel to download new DSP code */
/* unsigned short *maxlen;		Input: Maximum length of allocated memory	*/
/* char *files;					Output: List of filenames used by XerXes	*/
{
	int status = DXP_SUCCESS;
	FILE *fp=NULL;
	char newFile[200];

	unsigned int index = 0;
	unsigned int len = 0;
	unsigned int start = 0;

	int modChan;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */
	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_nspec",info_string,&status);
		return status;
	}

/* Try to find the modules configuration file	*/
    if((fp=dxp_find_file(chosen->fippi[modChan]->filename,"r",newFile))==NULL){ 
        status = DXP_OPEN_FILE;
	    sprintf(info_string,"Unable to locate the FIPPI firmware file");
		dxp_md_error("dxp_locate_files",info_string,&status);
	} else {
		fclose(fp);
		if (strlen(newFile)>(*maxlen+1)) {
			len = *maxlen;
			start = strlen(newFile) - *maxlen;
		} else {
			len = strlen(newFile);
			start = 0;
		}
		strncpy(files[index], newFile+start, len);
		files[index][len] = 0;
		index++;
	}

/* Try to find the modules configuration file	*/
    if((fp=dxp_find_file(chosen->dsp[modChan]->filename,"r",newFile))==NULL){ 
        status = DXP_OPEN_FILE;
	    sprintf(info_string,"Unable to locate the DSP firmware file");
		dxp_md_error("dxp_locate_files",info_string,&status);
	} else {
		fclose(fp);
		if (strlen(newFile)>(*maxlen+1)) {
			len = *maxlen;
			start = strlen(newFile) - *maxlen;
		} else {
			len = strlen(newFile);
			start = 0;
		}
		strncpy(files[index], newFile+start, len);
		files[index][len] = 0;
		index++;
	}

/* Try to find the preamp configuration file	*/
    if((fp=dxp_find_file(chosen->defaults[modChan]->filename,"r",newFile))==NULL){ 
        status = DXP_OPEN_FILE;
	    sprintf(info_string,"Unable to locate the default DSP parameters file");
		dxp_md_error("dxp_locate_files",info_string,&status);
	} else {
		fclose(fp);
		if (strlen(newFile)>(*maxlen+1)) {
			len = *maxlen;
			start = strlen(newFile) - *maxlen;
		} else {
			len = strlen(newFile);
			start = 0;
		}
		strncpy(files[index], newFile+start, len);
		files[index][len] = 0;
		index++;
	}
	
	return status;
}

/******************************************************************************
 *
 * Routine to read the modules configuration file.
 *
 ******************************************************************************/
static int XERXES_API dxp_read_modules(FILE* fp)
/* FILE *fp;						Input: FILE pointer to modules configuration	*/
{
    int status;
    char *cstatus=" ";

	unsigned int j, len;
	unsigned int n, m, nchan;

	char strtemp[132];
	char temp_line[132];
	char ioname[132] = "";

	char *values[10];
	
/* Has the board types structure been filled? */
	if (btypes_head == NULL) {
        status = DXP_INITIALIZE;
        sprintf(info_string,"There is no board types structure");
        dxp_md_error("dxp_read_modules",info_string,&status);
        return status;
	}

/* Initialize the boards modules */
	if((status=dxp_init_boards_ds())!=DXP_SUCCESS){
        sprintf(info_string,"Failed to initialize the boards structures");
        dxp_md_error("dxp_read_modules",info_string,&status);
        return status;
    }

/* Start up reading the file */
	while (cstatus != NULL) { 
		do {
			cstatus = fgets(line,132,fp); 
		} while (((line[0]=='*')||(line[0]=='\r')||(line[0]=='\n'))&&(cstatus!=NULL)); 
		if(cstatus==NULL) continue;
/* What kind of entry do we have?
 * Supported types:  "module" = new module
 *				     "default_dsp" = default filename for DSP code
 *				     "default_fippi" = default filename for FIPPI code
 *				     "default_param" = default filename for parameters
 *				     "dsp" channel = DSP filename for this channel
 *				     "fippi" channel = FIPPI filename for this channel
 *				     "param" channel = default parameter setting for this filename
 */
/* Get the length of the line */
		len = strlen(line);
/* Make a copy of the line, since strtok buggers it up */
		strcpy(temp_line, line);
/* Find the first token deliminated by a space, this is the entry type */
		token = strtok(temp_line, delim);
/* Make sure the token is all lower case */
		for (j=0;j<strlen(token);j++) token[j] = (char) tolower(token[j]);
/* 
 * Now Identify the entry type 
 */
		if (strcmp(token, "board_type")==0) {
			values[0] = strtok(NULL, delim);
/* add board type */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add board type: %s",	values[0]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
/* Define the tag passed to DXP_MD_INITIALIZE(), typically used to 
 * specify what routines to use for IO with the CAMAC/EPP drivers */
		} else if (strcmp(token, "iolibrary")==0) {
/* Now search for the tag starting at m+1 */
			dxp_pick_filename(strlen(token)+1, line, ioname);
/* Define the interface tag, this will typically be the information 
 * that the DXP_MD_INITIALIZE() will use to determine what MD routines
 * are needed */
		} else if (strcmp(token, "interface")==0) {
/* Now search for the tag starting at m+1 */
			dxp_pick_filename(strlen(token)+1, line, strtemp);
			values[0] = strtemp;
/* Initialize the library for each Interface, if we have defined an IO tag */
			if (strcmp(ioname,"")==0) {
				status = DXP_INITIALIZE;
				sprintf(info_string,"Must define a iolibrary tag before every interface");
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
			values[1] = ioname;
/* Add the interface */
			if((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
	            sprintf(info_string,"Unable to load Interface %s",values[0]);
	            dxp_md_error("dxp_read_modules",info_string,&status);
	            return status;
	        }
/* Now to be safe, lets clear the ioname variable, this forces the user to define an
 * IO library for every interface specified... */
			strcpy(ioname,"");
/* Is this a new module in the system */
		} else if (strcmp(token, "module")==0) {
/* Pull out the information for where in the system the module sits 
 * and then open the crate via the dxp_md_open() routine.  This routine is 
 * expected to return a number for ioChan that identifies this module's
 * location in the array used by the machine dependent functions to 
 * identify how to talk to this module. */
			values[0] = strtok(NULL,delim);

/* Now loop over the remaining entries to determine channel ids and total 
 * number of channels
 */
			nchan = 0;
			values[2+nchan] = strtok(NULL,delim);
			while (values[2+nchan]!=NULL) {
				nchan++;
				values[2+nchan] = strtok(NULL,delim);
	        }
			sprintf(strtemp,"%i",nchan);
			values[1] = strtemp;
/* Add the board */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add board: %s",	values[0]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
/* We need to define a new DSP default */
		} else if (strcmp(token, "default_dsp")==0) {
/* Now search for the filename starting at end of token */
			dxp_pick_filename(strlen(token)+1, line, strtemp);
			values[0] = strtemp;
/* Add the Dsp */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add DSP: %s", values[0]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
/* We need to define a new DSP default */
		} else if (strcmp(token, "dsp")==0) {
/* Get the next token, it is the channel Number */
			values[0] = strtok(NULL, delim);

			n=0;
			dxp_strnstrm(line, token, &n, &m);
			n = m;
			dxp_strnstrm(line, values[0], &n, &m);
/* Now n contains the location in the line after the 2nd token */

/* Now search for the filename starting at m+1 */
			dxp_pick_filename(m+1, line, strtemp);
			values[1] = strtemp;
/* Add the Dsp */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add DSP to channel %s: %s", 
					values[0], values[1]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
/* We need to define a new FIPPI default */
		} else if (strcmp(token, "default_fippi")==0) {
/* Now search for the filename starting at end of token */
			dxp_pick_filename(strlen(token)+1, line, strtemp);
			values[0] = strtemp;
/* Add the Fippi */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add FIPPI: %s", values[0]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
/* We need to define a new FIPPI default */
		} else if (strcmp(token, "fippi")==0) {
/* Get the next token, it is the channel Number */
			values[0] = strtok(NULL, delim);

			n=0;
			dxp_strnstrm(line, token, &n, &m);
/* Now m contains the location in the line after the 2nd token */
			n = m;
			dxp_strnstrm(line, values[0], &n, &m);

/* Now search for the filename starting at m+1 */
			dxp_pick_filename(m+1, line, strtemp);
			values[1] = strtemp;
/* Add the Fippi */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add FIPPI to channel %s: %s", 
					values[0], values[1]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
/* We need to define a new parameter default */
		} else if (strcmp(token, "default_param")==0) {
/* Now search for the filename starting at end of token */
			dxp_pick_filename(strlen(token)+1, line, strtemp);
			values[0] = strtemp;
/* Add the Defaults */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add DSP Defaults: %s", values[0]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
/* We need to define a new FIPPI default */
		} else if (strcmp(token, "param")==0) {
/* Get the next token, it is the channel Number */
			values[0] = strtok(NULL, delim);

			n=0;
			dxp_strnstrm(line, token, &n, &m);
/* Now m contains the location in the line after the 2nd token */
			n = m;
			dxp_strnstrm(line, values[0], &n, &m);

/* Now search for the filename starting at m+1 */
			dxp_pick_filename(m+1, line, strtemp);
			values[1] = strtemp;
/* Add the Defaults */
			if ((status=dxp_add_board_item(token,values))!=DXP_SUCCESS) {
				sprintf(info_string,"Unable to add DSP Defaults to channel %s: %s", 
					values[0], values[1]);
				dxp_md_error("dxp_read_modules",info_string,&status);
				return status;
			}
		} else if (strcmp(token, "end")==0) {
			return DXP_SUCCESS;
		} else {
            sprintf(info_string,"Unrecognized entry in modules file : %s",token);
            dxp_md_error("dxp_read_modules",info_string,&status);
            return status;
		}
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Read the file pointed to by DXP_MODULE:  This should have one line
 * for each module in the system (excluding comment lines -- those that
 * begin with a *)  The first field in the line is a 4 digit number that 
 * points to branch, crate and slot number of the module.  The second thru fifth
 * fields are non-negative integers for those channnels that are instrumented
 * (or to be readout)  These numbers are interpreted as detector elements.
 * A 0 or negative number in the second thru fifth fields indicates that the DXP
 * is unused (or non-existant)   This routine also fills CamChan[],Used[],
 * DetChan[], NumDxpMod and NumDxpChan these are global to this file.
 *
 ******************************************************************************/
int XERXES_API dxp_assign_channel(VOID)
{
	int status;
    FILE *fp=NULL;

	int i;
	char strtmp[4];

	char newFile[MAXFILENAME_LEN];

/* Get a file pointer to the modules configuration file */

	for (i=0;i<4;i++) strtmp[i] = (char) toupper(info->modules[i]);

	if (strncmp(strtmp,"NULL",4)!=0) {

		if ((fp=dxp_find_file(info->modules,"r",newFile))==NULL){
			status = DXP_OPEN_FILE;
			sprintf(info_string,"could not open module configuration file : %s",
				info->modules);
			dxp_md_error("dxp_assign_channel",info_string,&status);
			return status;
		}

/* Read in the configuration */

	    if ((status=dxp_read_modules(fp))!=DXP_SUCCESS){
			status = DXP_INITIALIZE;
			sprintf(info_string,"Could not read in modules file : %s", info->modules);
			dxp_md_error("dxp_assign_channel",info_string,&status);
			return status;
		}
	}

    fclose(fp);
    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to a Board Structure
 *
 ******************************************************************************/
static int XERXES_API dxp_free_board(Board* board)
/* Board *board;							Input: pointer to structure to free	*/
{
	int status=DXP_SUCCESS;
	unsigned short i;

	if (board==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"Board object unallocated:  can not free");
        dxp_md_error("dxp_free_board",info_string,&status);
        return status;
	}

/* Free the detChan array */
	if (board->detChan!=NULL)
		dxp_md_free(board->detChan);

/* Free the iostring entry */
	if (board->iostring!=NULL) 
		dxp_md_free(board->iostring);

/* Free the params array */
	if (board->params!=NULL) {
		for (i=0; i<board->nchan; i++) {
			if (board->params[i]!=NULL) {
				dxp_md_free(board->params[i]);
				board->params[i] = NULL;
			}
		}
		dxp_md_free(board->params);
	}

/* Free the Dsp Array*/
	if (board->dsp!=NULL)
		dxp_md_free(board->dsp);

/* Free the Fippi Array*/
	if (board->fippi!=NULL)
		dxp_md_free(board->fippi);

/* Free the Defaults Array*/
	if (board->defaults!=NULL)
		dxp_md_free(board->defaults);

/* Free the DAQ Array*/
	if (board->daq!=NULL) {
		status = DXP_INITORDER;
		sprintf(info_string,"Daq Structure allocated: MUST initialize HanDeL first");
        dxp_md_error("dxp_free_board",info_string,&status);
        return status;
	}

/* Free the Board structure */
	dxp_md_free(board);
	board = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an Interface Structure
 *
 ******************************************************************************/
static int XERXES_API dxp_free_iface(Interface *iface)
/* Interface *iface;					Input: pointer to structure to free	*/
{
	int status=DXP_SUCCESS;

	if (iface ==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"Interface object unallocated:  can not free");
        dxp_md_error("dxp_free_iface",info_string,&status);
        return status;
	}

	if (iface->dllname != NULL) dxp_md_free(iface->dllname);
	if (iface->ioname  != NULL) dxp_md_free(iface->ioname);
	if (iface->funcs   != NULL) dxp_md_free(iface->funcs);

/* Free the Interface structure */
	dxp_md_free(iface);
	iface = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an Board_Info Structure
 *
 ******************************************************************************/
static int XERXES_API dxp_free_binfo(Board_Info *binfo)
/* Board_Info *binfo;					Input: pointer to structure to free	*/
{
	int status=DXP_SUCCESS;

	if (binfo==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"Board_Info object unallocated:  can not free");
        dxp_md_error("dxp_free_binfo",info_string,&status);
        return status;
	}

	if (binfo->pointer != NULL) dxp_md_free(binfo->pointer);
	if (binfo->name    != NULL) dxp_md_free(binfo->name);
	if (binfo->funcs   != NULL) dxp_md_free(binfo->funcs);

/* Free the Board_Info structure */
	dxp_md_free(binfo);
	binfo = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an Dsp Structure
 *
 ******************************************************************************/
static int XERXES_API dxp_free_dsp(Dsp_Info *dsp)
/* Dsp_Info *dsp;					Input: pointer to structure to free	*/
{
	int status=DXP_SUCCESS;

	if (dsp==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"Dsp object unallocated:  can not free");
        dxp_md_error("dxp_free_dsp",info_string,&status);
        return status;
	}

	if (dsp->filename!=NULL) 
			dxp_md_free(dsp->filename);
	if (dsp->data!=NULL) 
			dxp_md_free(dsp->data);
	if (dsp->params!=NULL)
		dxp_free_params(dsp->params);

/* Free the Dsp_Info structure */
	dxp_md_free(dsp);
	dsp = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an Fippi Structure
 *
 ******************************************************************************/
static int XERXES_API dxp_free_fippi(Fippi_Info *fippi)
/* Fippi_Info *fippi;					Input: pointer to structure to free	*/
{
	int status=DXP_SUCCESS;

	if (fippi==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"Fippi object unallocated:  can not free");
        dxp_md_error("dxp_free_fippi",info_string,&status);
        return status;
	}

	if (fippi->filename!=NULL) 
				dxp_md_free(fippi->filename);
	if (fippi->data!=NULL) 
				dxp_md_free(fippi->data);

/* Free the Fippi_Info structure */
	dxp_md_free(fippi);
	fippi = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an Fippi Structure
 *
 ******************************************************************************/
static int XERXES_API dxp_free_params(Dsp_Params *params)
/* Dsp_Params *params;					Input: pointer to structure to free	*/
{
	int j,status=DXP_SUCCESS;

	if (params==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"Dsp_Params object unallocated:  can not free");
        dxp_md_error("dxp_free_params",info_string,&status);
        return status;
	}

	if (params->pnames!=NULL) {
		for (j=0;j<params->nsymbol;j++) {
			if (params->pnames[j]!=NULL)
					dxp_md_free(params->pnames[j]);
		}
		dxp_md_free(params->pnames);
	}
	if (params->access!=NULL) 
				dxp_md_free(params->access);
	if (params->lbound!=NULL) 
				dxp_md_free(params->lbound);
	if (params->ubound!=NULL) 
				dxp_md_free(params->ubound);

/* Free the Dsp_Params structure */
	dxp_md_free(params);
	params = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an Dsp_Defaults Structure
 *
 ******************************************************************************/
static int XERXES_API dxp_free_defaults(Dsp_Defaults *defaults)
/* Dsp_Defaults *defaults;					Input: pointer to structure to free	*/
{
	int status=DXP_SUCCESS;

	if (defaults==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"Dsp_Defaults object unallocated:  can not free");
        dxp_md_error("dxp_free_defaults",info_string,&status);
        return status;
	}

	if (defaults->filename!=NULL)
				dxp_md_free(defaults->filename);
	if (defaults->data!=NULL)
				dxp_md_free(defaults->data);
	if (defaults->params!=NULL) {
		dxp_free_params(defaults->params);
	}

/* Free the Dsp_Defaults structure */
	dxp_md_free(defaults);
	defaults = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to remove a board from the linked list.
 *
 ******************************************************************************/
int XERXES_API dxp_del_board(char* type, char* iolib, char* ifacelib, char* iostring)
/* char *type;					Input: board type, must match btype entry	*/
/* char *iolib;					Input: Library to use to perform IO			*/
/* char *ifacelib;				Input: Library to use for user spec routines	*/
/* char *iostring;				Input: string to pass to the iface library	*/
{
	int status=DXP_SUCCESS;

	Board *current = system_head, *prev = NULL;

/* First search thru the linked list and see if this Board exists
 * already exists */
	while (current!=NULL) {

		if (strcmp(current->btype->name, type)==0) {
/* Do the interface and iostring match? */
			if (strcmp(current->iface->dllname, ifacelib)==0) {
				if (strcmp(current->iface->ioname, iolib)==0) {
					if (strcmp(current->iostring, iostring)==0) {
/* Found the board entry....now remove */
						if (current==system_head) {
							system_head = current->next;
						} else {
							prev->next = current->next;
						}
						dxp_free_board(current);
						return DXP_SUCCESS;
					}
				}
			}
		}
		prev = current;
		current = current->next;
	}

	status = DXP_NOMATCH;
	sprintf(info_string,"Board Type: %s using %s with id %s unknown",
				type, ifacelib, iostring);
	dxp_md_error("dxp_del_board",info_string,&status);
	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to remove a board type from the linked list.
 *
 ******************************************************************************/
int XERXES_API dxp_del_btype(char* name)
/* char *name;					Input: board type	*/
{
	int status=DXP_SUCCESS;

	Board_Info *current = btypes_head, *prev = NULL;

/* First search thru the linked list and find the board type */
	while (current!=NULL) {

		if (strcmp(current->name, name)==0) {
/* Found the board type entry....now remove */
			if (current==btypes_head) {
				btypes_head = current->next;
			} else {
				prev->next = current->next;
			}
			dxp_free_binfo(current);
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

	status = DXP_NOMATCH;
	sprintf(info_string,"Board Type %s is unknown", name);
	dxp_md_error("dxp_del_btype",info_string,&status);
	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to remove a DSP structure from the linked list.
 *
 ******************************************************************************/
int XERXES_API dxp_del_dsp(char* filename)
/* char *filename;			Input: filename associated with DPS	*/
{
	int status=DXP_SUCCESS;

	Dsp_Info *current = dsp_head, *prev = NULL;

/* First search thru the linked list and find the DSP structure */
	while (current!=NULL) {

		if (strcmp(current->filename, filename)==0) {
/* Found the dsp entry....now remove */
			if (current==dsp_head) {
				dsp_head = current->next;
			} else {
				prev->next = current->next;
			}
			dxp_free_dsp(current);
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

	status = DXP_NOMATCH;
	sprintf(info_string,"DSP %s is unknown", filename);
	dxp_md_error("dxp_del_dsp",info_string,&status);
	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to remove a FIPPI structure from the linked list.
 *
 ******************************************************************************/
int XERXES_API dxp_del_fippi(char* filename)
/* char *filename;			Input: filename associated with FIPPI	*/
{
	int status=DXP_SUCCESS;

	Fippi_Info *current = fippi_head, *prev = NULL;

/* First search thru the linked list and find the FIPPI structure */
	while (current!=NULL) {

		if (strcmp(current->filename, filename)==0) {
/* Found the fippi entry....now remove */
			if (current==fippi_head) {
				fippi_head = current->next;
			} else {
				prev->next = current->next;
			}
			dxp_free_fippi(current);
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

	status = DXP_NOMATCH;
	sprintf(info_string,"FIPPI %s is unknown", filename);
	dxp_md_error("dxp_del_fippi",info_string,&status);
	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to remove a Defaults structure from the linked list.
 *
 ******************************************************************************/
int XERXES_API dxp_del_defaults(char* filename)
/* char *filename;			Input: filename associated with Defaults	*/
{
	int status=DXP_SUCCESS;

	Dsp_Defaults *current = defaults_head, *prev = NULL;

/* First search thru the linked list and find the FIPPI structure */
	while (current!=NULL) {

		if (strcmp(current->filename, filename)==0) {
/* Found the fippi entry....now remove */
			if (current==defaults_head) {
				defaults_head = current->next;
			} else {
				prev->next = current->next;
			}
			dxp_free_defaults(current);
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

	status = DXP_NOMATCH;
	sprintf(info_string,"Defaults %s is unknown", filename);
	dxp_md_error("dxp_del_defaults",info_string,&status);
	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to Load a new board into the linked list.  
 *
 ******************************************************************************/
int XERXES_API dxp_add_board(char* type, char* iolib, char* ifacelib, char* iostring, 
				  unsigned short* nchan, int* detChan)
/* char *type;					Input: board type, must match btype entry	*/
/* char *iolib;					Input: Tag specifying how to do IO w/ driver	*/
/* char *ifacelib;				Input: Tag to specify what MD to use			*/
/* char *iostring;				Input: string to pass to the iface library	*/
/* unsigned short *nchan;		Input: Number of channels on board			*/
/* int *detChan;				Input: Array of detector channel numbers		*/
{
	int status=DXP_SUCCESS;
	unsigned short used;
	int ioChan;
	unsigned short j;

	Board_Info *btype = NULL;
	Interface *iface = NULL;
	Board *current = system_head, *prev = NULL;

/* First search thru the linked list and see if this Board exists
 * already exists */
	while (current!=NULL) {

/* Do the interface and iostring match? */
		if (strcmp(current->iface->dllname, ifacelib)==0) {
			if (strcmp(current->iface->ioname, iolib)==0) {
				if (strcmp(current->iostring, iostring)==0) {
/* If we get here then this Board is already defined */
					sprintf(info_string,"Board Type: %s using %s with id %s already exists",
								type, ifacelib, iostring);
					dxp_md_error("dxp_add_board",info_string,&status);
					return DXP_SUCCESS;
				}
			}
		}
		prev = current;
		current = current->next;
	}

/* Pack an information string for debugging purposes and error reporting */
/* Make sure we haven't tried to allocate too many modules */
    sprintf(info_string," mod %i: %s with %i channels",numDxpMod,iostring,*nchan);
	if(numDxpMod == MAXDXP){
		sprintf(info_string,"Too many modules in config:MAXDXP = %i",
				MAXDXP);
		status = DXP_ARRAY_TOO_SMALL;
		dxp_md_error("dxp_add_board",info_string,&status);
		return status;
	}
	dxp_md_error("dxp_add_board",info_string,&debug);

/* 
 * Retrieve Pointer to the Board_Type structure
 */
	if ((status=dxp_get_btype(type, &btype))!=DXP_SUCCESS) {
		sprintf(info_string,"Board Type %s not defined", type);
		status = DXP_UNKNOWN_BTYPE;
		dxp_md_error("dxp_add_board",info_string,&status);
		return status;
	}

/* 
 * Add or Retrieve Pointer to the Interface structure
 */
	if ((status=dxp_add_iface(ifacelib, iolib, &iface))!=DXP_SUCCESS) {
		sprintf(info_string,"Unable to allocate Interface %s and iolib %s", 
								ifacelib, iolib);
		status = DXP_UNKNOWN_BTYPE;
		dxp_md_error("dxp_add_board",info_string,&status);
		return status;
	}

/* Set the NumDxpMod'th to have none of the channels used */
	used = 0;

/* Pull out the information for where in the system the module sits 
 * and then open the crate via the dxp_md_open() routine.  This routine is 
 * expected to return a number for ioChan that identifies this module's
 * location in the array used by the machine dependent functions to 
 * identify how to talk to this module. */
	if((status=iface->funcs->dxp_md_open(iostring,&ioChan))!=DXP_SUCCESS){
		sprintf(info_string,"Could not open Module %s",iostring);
		dxp_md_error("dxp_add_board",info_string,&status);
		return status;
	}

/* Now loop over the detector numbers to determine how many active channels
 * we have.  If detector number is positive then the channel is 
 * considered instrumented and the bit is set in the used value
 * for future reference.  Else if the number is negative the channel
 * is ignored */
        
	for(j=0;j<*nchan;j++){
		if (*detChan>=0) {
			used |= 1<<j;
			numDxpChan++;
		}
	}

/* Create an entry in the linked list, checking to see if it is the first */
	if (system_head==NULL) {
		system_head = (Board *) dxp_md_alloc(sizeof(Board));
		current = system_head;
	} else {
		current->next = (Board *) dxp_md_alloc(sizeof(Board));
		current = current->next;
	}
	current->ioChan= ioChan;
	current->used	= used;
	memcpy(current->detChan, detChan, *nchan*sizeof(int));
	current->mod	= numDxpMod;
	current->btype	= btype;
	current->iface	= iface;
	current->iostring = (char *)
		dxp_md_alloc((strlen(iostring)+1)*sizeof(char));
	strcpy(current->iostring, iostring);
/* 
 * Initialize the interface.  Note that if multiple boards use the same
 * Interface, then this call is redundant on 2nd, etc...instances...but, 
 * make the call anyway, in case in the future we wish to enable different
 * interfaces for each board rather than the current limitation of essentially
 * one interface per board type...there are many implications to implementing 
 * this feature though...
 */
	current->btype->funcs->dxp_init_driver(iface);
	current->next	= NULL;

	for (j=0;j<*nchan;j++) {
		current->dsp[j] = NULL;
		current->params[j] = NULL;
		current->fippi[j] = NULL;
		current->defaults[j] = NULL;
	}

/* Initialize the DAQ structure */
	current->daq = (DaqValues *) dxp_md_alloc(sizeof(DaqValues)*(*nchan));
	for (j=0;j<*nchan;j++) {
		current->daq[j].peak = 0.0;
		current->daq[j].nroi = 0;
		current->daq[j].low_roi = NULL;
		current->daq[j].hi_roi = NULL;
	}

/* We seem to have a new module in the system...increment the global counter */
	numDxpMod++;

	return status;
}

/******************************************************************************
 *
 * Routine to Retrieve the correct Board_Info structure from the 
 * global linked list...match the type name.
 *
 ******************************************************************************/
static int XERXES_API dxp_get_btype(char* name, Board_Info** match)
/* char *name;					Input: name of the board type			*/
/* Board_Info **match;			Output: pointer the correct structure	*/
{
	int status=DXP_SUCCESS;

	Board_Info *current = btypes_head;

/* First search thru the linked list and see if this configuration 
 * already exists */
	while (current!=NULL) {
/* Does the filename match? */
		if (strcmp(current->name, name)==0) {
			*match = current;
			return status;
		}
		current = current->next;
	}

	status = DXP_UNKNOWN_BTYPE;
	sprintf(info_string,"Board Type %s unknown, please define",name);
	dxp_md_error("dxp_get_btype",info_string,&status);
	return status;
}

/******************************************************************************
 *
 * Routine to Load a new board Type into the linked list.  Return pointer
 * to the member corresponding to the new board type.
 *
 ******************************************************************************/
int XERXES_API dxp_add_btype(char* name, char* pointer, char* dllname)
/* char *name;							Input: name of the new board type	*/
/* char *pointer;						Input: extra information filename	*/
/* char *dllname;						Input: name of the board type library*/
{
	int status=DXP_SUCCESS;

	Board_Info *prev = NULL;
	Board_Info *current = btypes_head;
	char *ctemp;

/* Assign unused inputs to get rid of compiler warnings */
	ctemp = dllname;

/* First search thru the linked list and see if this configuration 
 * already exists */
	while (current!=NULL) {
/* Does the filename match? */
		if (strcmp(current->name, name)==0) {
			sprintf(info_string,"Board Type: %s already exists",name);
			dxp_md_error("dxp_add_btype",info_string,&status);
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

/* No match in the existing list, add a new entry */
	if (btypes_head==NULL) {
/* Initialize the Header entry in the btypes linked list */
		btypes_head = (Board_Info*) dxp_md_alloc(sizeof(Board_Info));
		btypes_head->type = 0;
		current = btypes_head;
/* Assign the head entry of board types to the global structure */
		info->btypes = btypes_head;
	} else {
/* Not First time, allocate memory for next member of linked list */
		prev->next = (Board_Info*) dxp_md_alloc(sizeof(Board_Info));
		current = prev->next;
		current->type = prev->type+1;
	}
/* Set funcs pointer to NULL, will be set by add_btype_library() */
	current->funcs = NULL;
/* Set the next pointer to NULL */
	current->next = NULL;
/* This a new board type. */
	current->name = (char *) dxp_md_alloc((strlen(name)+1)*sizeof(char));
/* Copy the tag into the structure.  End the string with 
 * a NULL character. */
	strcpy(current->name, name);

/* Check to make sure a pointer was defined */
	if (pointer!=NULL) {
/* Now copy the pointer to use for this type configuration */
		current->pointer = (char *) dxp_md_alloc((strlen(pointer)+1)*sizeof(char));
/* Copy the filename into the structure.  End the string with 
 * a NULL character. */
		strcpy(current->pointer, pointer);
	} else {
		current->pointer = NULL;
	}

/* Load function pointers for local use */
	if ((status=dxp_add_btype_library(current))!=DXP_SUCCESS) {
		sprintf(info_string,"Problems loading functions for board type %s",current->name);
		dxp_md_error("dxp_add_btype",info_string,&status);
		return DXP_SUCCESS;
	}
/* Initialize the utilty routines in the library */
	current->funcs->dxp_init_utils(utils);

/* All done */
	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Initialize the FIPPI data structures to an empty state
 *
 ******************************************************************************/
int XERXES_API dxp_get_board_type(int *detChan, char *name)
{
    int status;
	int modChan;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */
	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_get_board_type",info_string,&status);
		return status;
	}
	
	strncpy(name, chosen->btype->name, MAXBOARDNAME_LEN);
	
	return status;
}

/******************************************************************************
 *
 * Routine to Load a board library and create pointers to the 
 * routines in the board library.
 *
 ******************************************************************************/
static int XERXES_API dxp_add_btype_library(Board_Info* current)
/* Board_Info *current;			Input: pointer to structure	*/
{
	int status=DXP_SUCCESS;


/* Allocate the memory for the function pointer structure */
	current->funcs = (Functions *) dxp_md_alloc(sizeof(Functions));

/* Assign the function pointers */
#ifndef EXCLUDE_DXP4C
	if (strcmp(current->name,"dxp4c")==0) {
		dxp_init_dxp4c(current->funcs);
	} else 
#endif
#ifndef EXCLUDE_DXP4C2X
	if ((strcmp(current->name,"dxp4c2x")==0) || (strcmp(current->name,"dxp2x")==0)) {
		dxp_init_dxp4c2x(current->funcs);
	} else 
#endif
#ifndef EXCLUDE_DXPX10P
	if (strcmp(current->name,"dxpx10p")==0) {
		dxp_init_dxpx10p(current->funcs);
	} else
#endif
	{
	    status = DXP_UNKNOWN_BTYPE;
        sprintf(info_string,"Unknown board type %s: unable to load pointers", current->name);
        dxp_md_error("dxp_add_btype_library",info_string,&status);
		return status;
	}

/* All done */
	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to Load a DSP configuration file into the Dsp_Info linked list
 * or find the matching member of the list and return a pointer to it
 *
 ******************************************************************************/
static int XERXES_API dxp_add_dsp(char* filename, Board_Info* board, Dsp_Info** passed)
/* char *filename;						Input: filename of the DSP program	*/
/* Board_Info *board;					Input: need the type of board		*/
/* Dsp_Info **passed;					Output: Pointer to the structure		*/
{
	int status;
	unsigned int i;

	Dsp_Info *temp_dsp = NULL, *prev = NULL;
	Dsp_Info *current = dsp_head;

/* First search thru the linked list and see if this configuration 
 * already exists */
	while (current!=NULL) {
/* Does the filename match? */
		if (strcmp(current->filename, filename)==0) {
			*passed = current;
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

/* No match in the existing list, add a new entry */
	if (dsp_head==NULL) {
/* First time, allocate memory for head of linked list */
		dsp_head = (Dsp_Info *) dxp_md_alloc(sizeof(Dsp_Info));
		current = dsp_head;
	} else {
/* Not First time, allocate memory for next member of linked list */
		prev->next = (Dsp_Info *) dxp_md_alloc(sizeof(Dsp_Info));
		current = prev->next;
	}
/* Set the next pointer to NULL */
	current->next = NULL;
/* Must allocate memory for the filename in the structure */
	current->filename = (char *) dxp_md_alloc((strlen(filename)+1)*sizeof(char));
	strcpy(current->filename, filename);

/* Now allocate memory and fill the program information 
 * with the driver routine */
	temp_dsp = (Dsp_Info *) dxp_md_alloc(sizeof(Dsp_Info));
	temp_dsp->params = (Dsp_Params *) dxp_md_alloc(sizeof(Dsp_Params));
	temp_dsp->filename = (char *) dxp_md_alloc((strlen(filename)+1)*sizeof(char));
	strcpy(temp_dsp->filename, filename);

/* Fill in the default lengths */
	if ((status=board->funcs->dxp_get_dspinfo(temp_dsp))!=DXP_SUCCESS) {
/* Free up memory, since there was an error */
		dxp_free_dsp(temp_dsp);
		dxp_md_free(current->filename);
		dxp_md_free(current);
		if (prev!=NULL)	prev->next = NULL;
/* Write out error message */
		sprintf(info_string,"Unable to get DSP information");
        dxp_md_error("dxp_add_dsp",info_string,&status);
        return status;
	}

/* Now finish allocating memory */
	temp_dsp->data = (unsigned short *) dxp_md_alloc(temp_dsp->maxproglen*sizeof(unsigned short));
	temp_dsp->params->pnames = (char **) dxp_md_alloc(temp_dsp->params->maxsym*sizeof(char *));
	for (i=0;i<temp_dsp->params->maxsym;i++)
		temp_dsp->params->pnames[i] = (char *) dxp_md_alloc(temp_dsp->params->maxsymlen*sizeof(char));
	temp_dsp->params->access = (unsigned short *) dxp_md_alloc(temp_dsp->params->maxsym*sizeof(unsigned short));
	temp_dsp->params->lbound = (unsigned short *) dxp_md_alloc(temp_dsp->params->maxsym*sizeof(unsigned short));
	temp_dsp->params->ubound = (unsigned short *) dxp_md_alloc(temp_dsp->params->maxsym*sizeof(unsigned short));
	memset(temp_dsp->params->access, 0, temp_dsp->params->maxsym*sizeof(temp_dsp->params->access[0]));
	memset(temp_dsp->params->lbound, 0, temp_dsp->params->maxsym*sizeof(temp_dsp->params->lbound[0]));
	memset(temp_dsp->params->ubound, 0, temp_dsp->params->maxsym*sizeof(temp_dsp->params->ubound[0]));

/* Now read the configuration file(s) */
	if ((status=board->funcs->dxp_get_dspconfig(temp_dsp))!=DXP_SUCCESS) {
/* Free up memory, since there was an error */
		temp_dsp->params->nsymbol = temp_dsp->params->maxsym;
		dxp_free_dsp(temp_dsp);
		dxp_md_free(current->filename);
		dxp_md_free(current);
		if (prev!=NULL) prev->next = NULL;
/* Write out error message */
		sprintf(info_string,"Unable to load new DSP file");
        dxp_md_error("dxp_add_dsp",info_string,&status);
        return status;
	}

/* Now Copy the temp_dsp structure into the current structure, this will
 * optimize memory use by using actual lengths instead of maximum lengths
 */
	current->maxproglen = temp_dsp->maxproglen;
	current->proglen = temp_dsp->proglen;
	current->data = (unsigned short *) dxp_md_alloc((temp_dsp->proglen)*sizeof(unsigned short));
	memcpy(current->data, temp_dsp->data, (temp_dsp->proglen)*sizeof(unsigned short));

	current->params = (Dsp_Params *) dxp_md_alloc(sizeof(Dsp_Params));
	current->params->nsymbol   = temp_dsp->params->nsymbol;
	current->params->maxsymlen = temp_dsp->params->maxsymlen;
	current->params->maxsym    = temp_dsp->params->maxsym;
	current->params->pnames    = (char **) 
			dxp_md_alloc((current->params->nsymbol)*sizeof(char *));
	for (i=0;i<current->params->nsymbol;i++) {
		current->params->pnames[i] = (char *) 
			dxp_md_alloc((strlen(temp_dsp->params->pnames[i])+1)*sizeof(char));
		strcpy(current->params->pnames[i], temp_dsp->params->pnames[i]);
	}
	current->params->access    = (unsigned short *) 
			dxp_md_alloc((current->params->nsymbol)*sizeof(unsigned short));
	memcpy(current->params->access, temp_dsp->params->access, 
									(temp_dsp->params->nsymbol)*sizeof(unsigned short));
	current->params->lbound    = (unsigned short *) 
			dxp_md_alloc((current->params->nsymbol)*sizeof(unsigned short));
	memcpy(current->params->lbound, temp_dsp->params->lbound, 
									(temp_dsp->params->nsymbol)*sizeof(unsigned short));
	current->params->ubound    = (unsigned short *) 
			dxp_md_alloc((current->params->nsymbol)*sizeof(unsigned short));
	memcpy(current->params->ubound, temp_dsp->params->ubound, 
									(temp_dsp->params->nsymbol)*sizeof(unsigned short));

/* Ok, the memory is copied into the local library memory space, now free the memory
 * used by temp_dsp */
	temp_dsp->params->nsymbol = temp_dsp->params->maxsym;
	dxp_free_dsp(temp_dsp);

/* Now assign the passed variable to the proper memory */
	*passed = current;

/* All done */
	return DXP_SUCCESS;
}
/******************************************************************************
 *
 * Routine to Load a FIPPI configuration file into the Fippi_Info linked list
 * or find the matching member of the list and return a pointer to it
 *
 ******************************************************************************/
static int XERXES_API dxp_add_fippi(char* filename, Board_Info* board, Fippi_Info** passed)
/* char *filename;					Input: filename of the FIPPI program	*/
/* Board_Info *board;				Input: need the type of board		*/
/* Fippi_Info **passed;				Output: Pointer to the structure		*/
{
	int status;

	Fippi_Info *temp_fippi = NULL, *prev= NULL;
	Fippi_Info *current = fippi_head;

/* First search thru the linked list and see if this configuration 
 * already exists */
	while (current!=NULL) {
/* Does the filename match? */
		if (strcmp(current->filename, filename)==0) {
			*passed = current;
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

/* No match in the existing list, add a new entry */
	if (fippi_head==NULL) {
/* First time, allocate memory for head of linked list */
		fippi_head = (Fippi_Info *) dxp_md_alloc(sizeof(Fippi_Info));
		current = fippi_head;
	} else {
/* Not First time, allocate memory for next member of linked list */
		prev->next = (Fippi_Info *) dxp_md_alloc(sizeof(Fippi_Info));
		current = prev->next;
	}
/* Set the next pointer to NULL */
	current->next = NULL;
	current->filename = (char *) dxp_md_alloc((strlen(filename)+1)*sizeof(char));
	strcpy(current->filename, filename);

/* Now allocate memory and fill the program information 
 * with the driver routine */
	temp_fippi = (Fippi_Info *) dxp_md_alloc(sizeof(Fippi_Info));
	temp_fippi->filename = (char *) dxp_md_alloc((strlen(filename)+1)*sizeof(char));
	strcpy(temp_fippi->filename, filename);

/* Fill in the default lengths */
	if ((status=board->funcs->dxp_get_fipinfo(temp_fippi))!=DXP_SUCCESS) {
/* Free up memory, since there was an error */
		dxp_md_free(temp_fippi);
		dxp_md_free(current->filename);
		dxp_md_free(current);
		if (prev!=NULL) prev->next = NULL;
/* Write out error message */
		sprintf(info_string,"Unable to get FIPPI information");
        dxp_md_error("dxp_add_fippi",info_string,&status);
        return status;
	}

/* Finish allocating memory */
	temp_fippi->data = (unsigned short *) 
		dxp_md_alloc((temp_fippi->maxproglen)*sizeof(unsigned short));
	if ((status=board->funcs->dxp_get_fipconfig(temp_fippi))!=DXP_SUCCESS) {
/* Free up memory, since there was an error */
		dxp_free_fippi(temp_fippi);
		dxp_md_free(current->filename);
		dxp_md_free(current);
		if (prev!=NULL) prev->next = NULL;
/* Write out error message */
		sprintf(info_string,"Unable to load new FIPPI file");
        dxp_md_error("dxp_add_fippi",info_string,&status);
        return status;
	}

/* Now Copy the fippi_temp structure into the fippi structure, this will
 * move the memory allocated in the driver layer into the permanent
 * midlevel layer */
	current->maxproglen = temp_fippi->maxproglen;
	current->proglen = temp_fippi->proglen;
	current->data = (unsigned short *) dxp_md_alloc((temp_fippi->proglen)*sizeof(unsigned short));
	memcpy(current->data, temp_fippi->data, (temp_fippi->proglen)*sizeof(unsigned short));

/* Ok, the memory is copied into the local library memory space, now free the memory
 * used by temp_fippi */
	dxp_free_fippi(temp_fippi);

/* Now assign the passed variable to the proper memory */
	*passed = current;

/* All done */
	return DXP_SUCCESS;
}
/******************************************************************************
 *
 * Routine to Load a DSP parameter defaults file into the Dsp_Defaults linked 
 * list or find the matching member of the list and return a pointer to it
 *
 ******************************************************************************/
static int XERXES_API dxp_add_defaults(char* filename, Board_Info* board, Dsp_Defaults** passed)
/* char *filename;						Input: filename of the DSP program	*/
/* Board_Info *board;					Input: need the type of board		*/
/* Dsp_Defaults **passed;				Output: Pointer to the structure		*/
{
	int status;
	unsigned int i;

	Dsp_Defaults *temp_defaults = NULL, *prev = NULL;
	Dsp_Defaults *current = defaults_head;

	char strtmp[4];

/* First search thru the linked list and see if this configuration 
 * already exists */
	while (current!=NULL) {
/* Does the filename match? */
		if (strcmp(current->filename, filename)==0) {
			*passed = current;
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

/* No match in the existing list, add a new entry */
	if (defaults_head==NULL) {
/* First time, allocate memory for head of linked list */
		defaults_head = (Dsp_Defaults *) dxp_md_alloc(sizeof(Dsp_Defaults));
		current = defaults_head;
	} else {
/* Not First time, allocate memory for next member of linked list */
		prev->next = (Dsp_Defaults *) dxp_md_alloc(sizeof(Dsp_Defaults));
		current = prev->next;
	}
/* Set the next pointer to NULL */
	current->next = NULL;
	current->filename = (char *) dxp_md_alloc((strlen(filename)+1)*sizeof(char));
	current->filename = strcpy(current->filename, filename);

/* Now allocate memory and fill the program information 
 * with the driver routine */
	temp_defaults = (Dsp_Defaults *) dxp_md_alloc(sizeof(Dsp_Defaults));
	temp_defaults->params = (Dsp_Params *) dxp_md_alloc(sizeof(Dsp_Params));
	temp_defaults->filename = (char *) dxp_md_alloc((strlen(filename)+1)*sizeof(char));
	strcpy(temp_defaults->filename, filename);

/* Fill in the default lengths */
	if ((status=board->funcs->dxp_get_defaultsinfo(temp_defaults))!=DXP_SUCCESS) {
/* Free up memory, since there was an error */
		dxp_free_defaults(temp_defaults);
		dxp_md_free(current->filename);
		dxp_md_free(current);
		if (prev!=NULL) prev->next = NULL;
/* Write out error message */
		sprintf(info_string,"Unable to get DSP Defaults information");
		dxp_md_error("dxp_add_defaults",info_string,&status);
		return status;
	}

/* Finish allocating memory */
	temp_defaults->data = (unsigned short *) dxp_md_alloc(temp_defaults->params->maxsym*sizeof(unsigned short));
	temp_defaults->params->pnames = (char **) dxp_md_alloc(temp_defaults->params->maxsym*sizeof(char *));
	for (i=0;i<temp_defaults->params->maxsym;i++)
		temp_defaults->params->pnames[i] = (char *) dxp_md_alloc(temp_defaults->params->maxsymlen*sizeof(char));
	temp_defaults->params->access = NULL;
	temp_defaults->params->lbound = NULL;
	temp_defaults->params->ubound = NULL;

	for (i=0;i<4;i++) strtmp[i] = (char) toupper(filename[i]);
	if (strncmp(strtmp,"NULL",4)!=0) {
/* If not a NULL entry, then call the primitive function */
		if ((status=board->funcs->dxp_get_dspdefaults(temp_defaults))!=DXP_SUCCESS) {
/* Free up memory, since there was an error */
			temp_defaults->params->nsymbol = temp_defaults->params->maxsym;
			dxp_free_defaults(temp_defaults);
			dxp_md_free(current->filename);
			dxp_md_free(current);
			if (prev!=NULL)	prev->next = NULL;
/* Write out error message */
			sprintf(info_string,"Unable to load new Defaults file");
			dxp_md_error("dxp_add_defaults",info_string,&status);
			return status;
		}
	} else {
		temp_defaults->params->nsymbol = 0;
	}

/* Now Copy the temp_defaults structure into the defaults structure, this will
 * move the memory allocated in the driver layer into the permanent
 * midlevel layer */
	current->params = (Dsp_Params *) dxp_md_alloc(sizeof(Dsp_Params));
	current->params->nsymbol   = temp_defaults->params->nsymbol;
	current->params->maxsymlen = temp_defaults->params->maxsymlen;
	current->params->maxsym    = temp_defaults->params->maxsym;
	if (temp_defaults->params->nsymbol != 0) {
		current->params->pnames    = (char **) 
				dxp_md_alloc((temp_defaults->params->nsymbol)*sizeof(char *));
		for (i=0;i<current->params->nsymbol;i++) {
			current->params->pnames[i] = (char *) 
				dxp_md_alloc((strlen(temp_defaults->params->pnames[i])+1)*sizeof(char));
			strcpy(current->params->pnames[i], temp_defaults->params->pnames[i]);
		}
		current->params->access = NULL;
		current->params->lbound = NULL;
		current->params->ubound = NULL;
		current->data = (unsigned short *) dxp_md_alloc((temp_defaults->params->nsymbol)*sizeof(unsigned short));
		memcpy(current->data, temp_defaults->data, (temp_defaults->params->nsymbol)*sizeof(unsigned short));
	}

/* Ok, the memory is copied into the local library memory space, now free the memory
 * used by temp_defaults_dsp */
	temp_defaults->params->nsymbol = temp_defaults->params->maxsym;
	dxp_free_defaults(temp_defaults);

/* Now assign the passed variable to the proper memory */
	*passed = current;

/* All done */
	return DXP_SUCCESS;
}
/******************************************************************************
 *
 * Routine to Load a new DLL library and return a pointer to the proper
 * Libs structure
 *
 ******************************************************************************/
static int XERXES_API dxp_add_iface(char* dllname, char* iolib, Interface** iface)
/* char *dllname;					Input: filename of the interface DLL		*/
/* char *iolib;						Input: name of lib to talk to the board	*/
/* Interface **iface;				Output: Pointer to the structure			*/
{
	int status;
	unsigned int maxMod = MAXDXP;

	Interface *prev = NULL;
	Interface *current = iface_head;

/* First search thru the linked list and see if this configuration 
 * already exists */
	while (current!=NULL) {
/* Does the filename match? */
		if (strcmp(current->dllname, dllname)==0) {
			*iface = current;
			return DXP_SUCCESS;
		}
		prev = current;
		current = current->next;
	}

/* No match in the existing list, add a new entry */
	if (iface_head==NULL) {
/* First time, allocate memory for head of linked list */
		iface_head = (Interface *) dxp_md_alloc(sizeof(Interface));
		current = iface_head;
	} else {
/* Not First time, allocate memory for next member of linked list */
		prev->next = (Interface *) dxp_md_alloc(sizeof(Interface));
		current = prev->next;
	}
/* Set the next pointer to NULL */
	current->next = NULL;
/* Copy the name of the DLL into the structure */
	current->dllname = (char *) dxp_md_alloc((strlen(dllname)+1)*sizeof(char));
	strcpy(current->dllname, dllname);
/* Copy in the ioname variable for future reference */
	current->ioname = (char *) dxp_md_alloc((strlen(iolib)+1)*sizeof(char));
	strcpy(current->ioname, iolib);

/* Allocate memory for the function structure */
	current->funcs = (Xia_Io_Functions *) dxp_md_alloc(sizeof(Xia_Io_Functions));
/* Retrieve the function pointers from the MD appropriate MD routine */
	dxp_md_init_io(current->funcs, current->dllname);
	
/* Now initialize the library */
	if ((status = 
		current->funcs->dxp_md_initialize(&maxMod, 
							current->ioname))!=DXP_SUCCESS){
		status = DXP_INITIALIZE;
		sprintf(info_string,"Could not initialize %s",current->dllname);
		dxp_md_error("dxp_add_iface",info_string,&status);
		return status;
	}

/* Now assign the passed variable to the proper memory */
	*iface = current;

/* All done */
	return DXP_SUCCESS;
}
/******************************************************************************
 *
 * Routine to search out the filename in a string
 *
 ******************************************************************************/
static int XERXES_API dxp_pick_filename(unsigned int start, char* line, char* filename)
/* unsigned int start;				*/
/* char *line;						*/
/* char *filename;					*/
{
	unsigned int finish, len;
	int status=DXP_SUCCESS;

/* Get the length of the line */
	len = strlen(line);
/* Unfort, need to allow spaces in the filenames.
 * so count forward in the line till out of spaces, this is the start of the 
 * file name */
	for(;start<len;start++) 
		if(isspace(line[start])==0) break;				/* find first character in path */ 
/* Now count backwards from the end of the line till you hit a non-space */
	for(finish=len;finish>start;finish--) 
		if(isgraph(line[finish])!=0) break;				/* find last character in name  */ 
/* Determine Length so i can NULL terminate the string */
	len = finish-start+1;
/* Copy the filename into the structure.  End the string with 
 * a NULL character. */
	filename = strncpy(filename, line+start, len);
	filename[len] = '\0';

	return status;
}
/******************************************************************************
 * 
 * Routine to search for a token within string, starting at location
 * n and returning the location of the last character match in string
 * as m.
 * 
 ******************************************************************************/
static int XERXES_API dxp_strnstrm(char* string, char* token, unsigned int* n, 
						unsigned int* m)
/* char *string;			Input: String to be searched			*/
/* char *token;				Input: String to match				*/
/* unsigned int *n;			Input: Starting location of search	*/
/* unsigned int *m;			Output: Ending location of the search*/
{
	unsigned int i;
	int status = DXP_SUCCESS;

/* start the search */
	for (i=*n;i<strlen(string);i++) {
		if (string[i]==token[0]) {
/* Found a match for the first character, now match rest */
			if (strncmp(string+i, token, strlen(token))==0) {
/* Found match for string, return the ending position */
				*m = i + strlen(token);
				return status;
			}
		}
	}

	status = DXP_NOMATCH;
    sprintf(info_string,"Could not match string: %s",token);
    dxp_md_error("dxp_strnstrm",info_string,&status);
    return status;

}
/******************************************************************************
 *
 *  This routine resets all channels by downloading Fippi, DSP and default
 *  parameter settings.  Then all calibrations are performed.
 *
 ******************************************************************************/
int XERXES_API dxp_user_setup(VOID){

	int status=DXP_SUCCESS;
	int chan, detChan;
	float adcRule, energy;

/* Pointers to the linked list */
	Board *current = system_head;

/*
 *  Download the FiPPI firmware
 */
	dxp_md_error("dxp_user_setup","calling dxp_fipconfig",&debug);
	if((status=dxp_fipconfig())!=DXP_SUCCESS){
		dxp_md_error("dxp_user_setup"," ",&status);
		return status;
	}

/*
 *  Download the DSP firmware
 */
	dxp_md_error("dxp_user_setup","calling dxp_dspconfig",&debug);
	if((status=dxp_dspconfig())!=DXP_SUCCESS){
		dxp_md_error("dxp_user_setup"," ",&status);
		return status;
	}

/* Loop over all Boards in the system and perform initializations */
	while (current!=NULL) {
		
/* Time to read in the default set of DSP parameters from the config file */
    
/* Loop over the current->nchan channels */
		for (chan=0;chan<(int) current->nchan;chan++) {
			detChan = current->detChan[chan];
			if (detChan < 0) continue;
			if ((status=dxp_upload_dspparams(&detChan))!=DXP_SUCCESS){
				status = DXP_DSPPARAMS;
				dxp_md_error("dxp_user_setup",
						"Could not load parameters from detector channel",&status);
				return status;
			}

			if ((status=dxp_dspdefaults(&detChan))!=DXP_SUCCESS){
				status = DXP_DSPPARAMS;
				dxp_md_error("dxp_user_setup",
						"Could not load default parameters",&status);
				return status;
			}
/*
 * Download the defaults
 * It is necc to wait till after downloading the FiPPi config
 */
		    dxp_md_error("dxp_user_setup","downloading parameters",&debug);
			if((status=dxp_replace_dspparams(&detChan))!=DXP_SUCCESS){
				dxp_md_error("dxp_user_setup","Unable to download DSP parameters",&status);
				return status;
			}
		}
		current = current->next;
	}

/*
 *  Perform initialization of the ASC
 */
	adcRule = (float) 0.05;
    if((status=dxp_initialize_ASC(&adcRule, &energy))!=DXP_SUCCESS){
        dxp_md_error("dxp_user_setup","Unable to initialize ASC",&status);
        return status;
    }

    return status; 

}

/******************************************************************************
 *
 *  This routine loads the FIPPI, DSP and dsp default parameters to the 
 *  specified channel.  Finally calibrations are performed on this 
 *  channel.
 *
 ******************************************************************************/
int XERXES_API dxp_reset_channel(int* detChan)
/* int *detChan;					*/
{
	int status=DXP_SUCCESS, i;
	int modChan, ioChan;
	unsigned short used;
/* Temporary holding for the preamp file name */
	char strtmp[4];

	Board *chosen = NULL;

/* Get the ioChan and dxpChan matching the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Unknown Detector Channel %d",*detChan);
		dxp_md_error("dxp_reset_channel",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
/*
 *  Download the FiPPI firmware
 */
	dxp_md_error("dxp_reset_channel","Downloading FIPPI",&debug);
	if((status=dxp_reset_fipconfig(detChan))!=DXP_SUCCESS){
		dxp_md_error("dxp_reset_channel","Unable to load FIPPI",&status);
		return status;
	}

/*
 *  Download the DSP firmware
 */
	dxp_md_error("dxp_reset_channel","calling dxp_dspconfig",&debug);
	if((status=dxp_reset_dspconfig(detChan))!=DXP_SUCCESS){
		dxp_md_error("dxp_reset_channel","Unable to load DSP",&status);
		return status;
	}
/*
 * Load the library memory with the defaults at startup.
 */
	if ((status=dxp_upload_dspparams(detChan))!=DXP_SUCCESS){
		status = DXP_DSPPARAMS;
		dxp_md_error("dxp_reset_channel",
				"Could not load parameters from detector channel",&status);
		return status;
	}
/* 
 * Now transfer the default DSP parameter values to library memory.
 */
	if ((status=dxp_dspdefaults(detChan))!=DXP_SUCCESS){
		status = DXP_DSPPARAMS;
		dxp_md_error("dxp_reset_channel",
				"Could not load default parameters",&status);
		return status;
	}
/*
 * Download the defaults
 * It is necc to wait till after downloading the FiPPi config
 */
    dxp_md_error("dxp_reset_channel","downloading parameters",&debug);
	if((status=dxp_replace_dspparams(detChan))!=DXP_SUCCESS){
		dxp_md_error("dxp_reset_channel","Unable to download DSP parameters",&status);
		return status;
	}

/*
 *  Perform initialization of the ASC
 */
/* Convert 1st 4 characters to upper for comparison */   
	for (i=0;i<4;i++) strtmp[i] = (char) toupper(info->preamp[i]);
	if ((info->preamp != NULL) || (strncmp(strtmp,"NULL",4)!=0)) {
		if ((status=chosen->btype->funcs->dxp_setup_asc(&ioChan, &modChan, &(chosen->mod), 
								&(chosen->preamp[modChan].adcrule), &(chosen->preamp[modChan].gainmod), 
								&(chosen->preamp[modChan].polarity), &(chosen->preamp[modChan].vmin),
								&(chosen->preamp[modChan].vmax), &(chosen->preamp[modChan].vstep), 
								chosen->dsp[modChan]))!=DXP_SUCCESS) {
				sprintf(info_string, 
				"Error setting up the ASC for Detector %d", *detChan);
			dxp_md_error("dxp_reset_channel",info_string,&status);
			return status;
		}
	}
/*
 *  Perform calibrations neccessary to setup the ASC
 */
/* Fake the routine into thinking only one channel is in use */
	used = (unsigned short) (1<<modChan);
    if((status=chosen->btype->funcs->dxp_calibrate_asc(&(chosen->mod), 
					&ioChan, &used, chosen->dsp))!=DXP_SUCCESS){
		dxp_md_error("dxp_reset_channel",
			"Unable to calibrate the ASC",&status);
		return status;
	}

    return status; 
}

/*****************************************************************************
 *
 * Routine to transfer the default DSP parameters from the defaults structure
 * for this channel, to the downloadable parameters.
 *
 ******************************************************************************/
int XERXES_API dxp_dspdefaults(int* detChan)
/* int *detChan;					Input: detector channel to set defaults	*/
{
	unsigned int i;
    unsigned short addr;
    int status=DXP_SUCCESS;
	int ioChan, modChan;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and dxpChan matching the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Unknown Detector Channel %d",*detChan);
		dxp_md_error("dxp_dspdefaults",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
	

/* Be paranoid and check if the DSP configuration is downloaded.  
 * If not, do it */
	
	if ((chosen->dsp[modChan]->proglen)==0) {
		dxp_md_error("dxp_dspdefaults", "Failed to download DSP", &status);
		return status;
	}

/*
 * Now replace all the symbols loaded from the defaults file
 * in the database.
 */
	for (i=0;i<(chosen->defaults[modChan]->params->nsymbol);i++) {
		chosen->btype->funcs->dxp_loc(chosen->defaults[modChan]->params->pnames[i], 
			chosen->dsp[modChan], &addr);
		chosen->params[modChan][addr] = 
			chosen->defaults[modChan]->data[i];
	}

	return status;

}

/******************************************************************************
 *
 * Downloads FiPPI firmware to all modules using broadcast command.  Check
 * CSR bits to see if download was successful.  Use the version of the firmware
 * matched to decimation specified.
 *
 ******************************************************************************/
int XERXES_API dxp_fipconfig(VOID)
{

    int status;
	int i, found;
	int mod;
/* Pointer to the linked list of Boards */
	Board *current = system_head;
	int ioChan;
	unsigned short used;
/* create local refernence to the ALLCHAN variable */
	int allchan = ALLCHAN;
/* Store the fippi that was broadcast */
	Fippi_Info *broadcast_fippi=NULL;

/* Loop over all the modules in the system */

	while (current!=NULL) {
		mod    = current->mod;
		ioChan = current->ioChan;
		used   = current->used;

/* Disable the LAMs for each module before downloading */

		if((status=current->btype->funcs->dxp_look_at_me(&ioChan, &allChan))!=DXP_SUCCESS){
            sprintf(info_string,"Error disabling LAM for module %d",mod);
            dxp_md_error("dxp_fipconfig",info_string,&status);
            return status;
        }

/* Broadcast the first used channel to all channels */
		i = 0;
		found = 0;
		do {
			if (used&(1<<i)) {
		        if((status=current->btype->funcs->dxp_download_fipconfig(&ioChan,
												  &allchan, current->fippi[i]))!=DXP_SUCCESS){
					sprintf(info_string,"Error Broadcasting to module %d",mod);
		            dxp_md_error("dxp_fipconfig",info_string,&status);
		            return status;
				}
				broadcast_fippi = current->fippi[i];
				found = 1;
			}
			i++;
		} while((broadcast_fippi==NULL) && (i<(int) current->nchan));

/* Check if at least one channel was active on this board */
		if (found==0) {
			status = DXP_SUCCESS;
			sprintf(info_string,"No used channel was found in module %d \r\n Will Continue with other modules",mod);
            dxp_md_error("dxp_fipconfig",info_string,&status);
		}

/* Download the FiPPi program to all channels of a DXP */

/* peculiar start to loop, DO NOT initialize i=0 since the previous
 * while loop will stop with the correct starting point for i.  The
 * previous values of i would either point to a channel that is already
 * downloaded or one that is not implemented */
		for (;i<(int) current->nchan;i++) {
			if (current->fippi[i] != broadcast_fippi) {
				if (used&(1<<i)) {
			        if((status=current->btype->funcs->dxp_download_fipconfig(&ioChan,
													  &i, current->fippi[i]))!=DXP_SUCCESS){
						sprintf(info_string,"Error downloading to module %d",mod);
			            dxp_md_error("dxp_fipconfig",info_string,&status);
			            return status;
					}
				}
			}
        }

/* Determine if the FiPPi was downloaded successfully to all channels */

		if((status=current->btype->funcs->dxp_download_fippi_done(&ioChan, &mod, 
											&used))!=DXP_SUCCESS){
            sprintf(info_string,"Failed to download FiPPi succesfully for module %d",mod);
            dxp_md_error("dxp_fipconfig",info_string,&status);
            return status;
        }
		current = current->next;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Downloads FiPPI firmware to all modules using broadcast command.  Check
 * CSR bits to see if download was successful.
 *
 ******************************************************************************/
int XERXES_API dxp_replace_fipconfig(int* detChan, char* filename)
/* int *detChan;							Input: detector channel to load     */
/* char *filename;							Input: location of the new FiPPi    */
{

    int status;
	int ioChan, modChan;
	unsigned short used;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and dxpChan matching the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Unknown Detector Channel %d",*detChan);
		dxp_md_error("dxp_replace_fipconfig",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
	used = chosen->used;
	
/* Disable the LAMs for each module before downloading */

	if((status=chosen->btype->funcs->dxp_look_at_me(&ioChan, &allChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error disabling LAM for detector %d",*detChan);
		dxp_md_error("dxp_replace_fipconfig",info_string,&status);
		return status;
	}

/* Load the Fippi configuration into the structure */

    if((status=dxp_add_fippi(filename, chosen->btype, &(chosen->fippi[modChan])))!=DXP_SUCCESS){
		sprintf(info_string,"Error loading Fippi %s",filename);
		dxp_md_error("dxp_replace_fipconfig",info_string,&status);
		return status;
	}

/* Download the FiPPi program to the channel */

	if((status=chosen->btype->funcs->dxp_download_fipconfig(&ioChan,
					&modChan, chosen->fippi[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error downloading to detector %d", *detChan);
		dxp_md_error("dxp_replace_fipconfig",info_string,&status);
		return status;
	}

/* Determine if the FiPPi was downloaded successfully */

	if((status=chosen->btype->funcs->dxp_download_fippi_done(&ioChan, &(chosen->mod), &used))!=DXP_SUCCESS){
		sprintf(info_string,"Failed to replace FiPPi succesfully for detector %d",*detChan);
		dxp_md_error("dxp_replace_fipconfig",info_string,&status);
		return status;
	}
		
    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Download currently loaded FIPPI firmware to a single channel.  
 *
 ******************************************************************************/
int XERXES_API dxp_reset_fipconfig(int* detChan)
/* int *detChan;					Input: Detector Channel number	*/
{
	int status;
	int modChan, ioChan;
	unsigned short used;
	Board *chosen = NULL;
	
/* Get the ioChan and dxpChan matching the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Unknown Detector Channel %d",*detChan);
		dxp_md_error("dxp_reset_fipconfig",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Disable the LAMs for each module before downloading */

	if((status=chosen->btype->funcs->dxp_look_at_me(&ioChan, &allChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error disabling LAM for detector %d",*detChan);
		dxp_md_error("dxp_reset_fipconfig",info_string,&status);
		return status;
	}

/* Download the FiPPi program to the channel */

	if((status=chosen->btype->funcs->dxp_download_fipconfig(&ioChan,
					&modChan, chosen->fippi[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error downloading to detector %d", *detChan);
		dxp_md_error("dxp_reset_fipconfig",info_string,&status);
		return status;
	}

/* Determine if the FiPPi was downloaded successfully */
/* Fake routine into only checking this channel */
	used = (unsigned short) (1<<modChan);
	if((status=chosen->btype->funcs->dxp_download_fippi_done(&ioChan, &(chosen->mod), &used))!=DXP_SUCCESS){
		sprintf(info_string,"Failed to replace FiPPi succesfully for detector %d",*detChan);
		dxp_md_error("dxp_reset_fipconfig",info_string,&status);
		return status;
	}
	
    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Download DSP firmware to a single channel.  Test download with calls to 
 * dxp_test_spectrum_memory().
 *
 ******************************************************************************/
int XERXES_API dxp_reset_dspconfig(int* detChan)
/* int *detChan;					Input: Detector Channel number	*/
{
    int status;
	int modChan, ioChan;
	Board *chosen = NULL;
	
/* Get the ioChan and dxpChan matching the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Unknown Detector Channel %d",*detChan);
		dxp_md_error("dxp_reset_dspconfig",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Download the DSP code to this channel of this module. */
	if((status=chosen->btype->funcs->dxp_download_dspconfig(&ioChan,
								&modChan, chosen->dsp[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error downloading to module %d, channel %d",chosen->mod,modChan);
		dxp_md_error("dxp_reset_dspconfig",info_string,&status);
		return status;
	}
/* Read out the DSP parameter memory, these contain the defaults defined
 * at the DSP compile time. */
	if((status=chosen->btype->funcs->dxp_read_dspparams(&ioChan, &modChan, 
							chosen->dsp[modChan], chosen->params[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error reading parameters from module %d channel %d",chosen->mod,modChan);
		dxp_md_error("dxp_reset_dspconfig",info_string,&status);
		return status;
	}
/* Test some DSP memory to make sure the download worked */
	if((status=chosen->btype->funcs->dxp_test_spectrum_memory(&ioChan, &modChan, &zero,
				chosen->dsp[modChan], chosen->params[modChan])) !=DXP_SUCCESS){
		sprintf(info_string, "Failure testing memory of mod %d channel %d", chosen->mod, modChan);
        dxp_md_error("dxp_reset_dspconfig",info_string,&status);
        return status;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Download DSP firmware to all channels.  Test download with calls to 
 * dxp_test_spectrum_memory().
 *
 ******************************************************************************/
int XERXES_API dxp_dspconfig(VOID)
{

    int status,chan;
	int ioChan, used, mod;
/* Pointer to the current Board */
	Board *current = system_head;
/* Pointer to the DSP program that was broadcast */
	Dsp_Info *broadcast_dsp=NULL;
/* Local value for ALLCHAN */
	int allchan = ALLCHAN;
/* Time to wait for the initialization of the DSP program */
	float two = 2., ten = 10.;
/* What BUSY value are we waiting on? */
	unsigned short value=0;
	
/* Loop over all the modules in the system. */
    
	while (current != NULL) {
		ioChan = current->ioChan;
		used    = current->used;
		mod     = current->mod;

/* First download the DSP programs */
		chan = 0;
		do {
            if ((used&(1<<chan))!=0) {
/* Download the DSP code to all channels. */
				if((status=current->btype->funcs->dxp_download_dspconfig(&ioChan,
								&allchan, current->dsp[chan]))!=DXP_SUCCESS){
		            sprintf(info_string,"Error broadcasting DSP program to module %d",mod);
					dxp_md_error("dxp_dspconfig",info_string,&status);
					return status;
				}
				broadcast_dsp = current->dsp[chan];
			}
			chan++;
		} while ((broadcast_dsp==NULL) && (chan<(int) current->nchan));
/* peculiar start to loop, DO NOT initialize chan=0 since the previous
 * while loop will stop with the correct starting point for chan.  The
 * previous values of chan would either point to a channel that is already
 * downloaded or one that is not implemented */
        for(;chan<(int) current->nchan;chan++){
            if (((used&(1<<chan))==0) || 
				(current->dsp[chan]==broadcast_dsp)) continue;
/* Download the DSP code to this channel of this module. */
			if((status=current->btype->funcs->dxp_download_dspconfig(&ioChan,
                                        &chan, current->dsp[chan]))!=DXP_SUCCESS){
	            sprintf(info_string,"Error downloading to module %d, channel %d",mod,chan);
				dxp_md_error("dxp_dspconfig",info_string,&status);
				return status;
			}
		}

/* Now wait for 2 seconds to allow DSP initialization to complete */
		dxp_md_wait(&two);

/* Now check the status of the BUSY parameter for all channels */
        for(chan = 0;chan<(int) current->nchan;chan++){
            if ((used&(1<<chan))==0) continue;
/* Wait for maximum of 10 seconds for the DSP to go to BUSY state to indicate DSP init done. */
			if((status=current->btype->funcs->dxp_download_dsp_done(&ioChan,
							&chan, &mod, current->dsp[chan], &value, &ten))!=DXP_SUCCESS){
	            sprintf(info_string,"Error waiting for DSP to initialize for module %d, channel %d",mod,chan);
				dxp_md_error("dxp_dspconfig",info_string,&status);
				return status;
			}
		}

/* Now Read all the parameters and test spectrm memory */		
		for (chan = 0;chan<(int) current->nchan;chan++) {
			if((status=current->btype->funcs->dxp_read_dspparams(&ioChan, &chan, 
									current->dsp[chan], current->params[chan]))!=DXP_SUCCESS){
				sprintf(info_string,"Error reading parameters from module %d channel %d",mod,chan);
				dxp_md_error("dxp_dspconfig",info_string,&status);
				return status;
			}
/* Test some DSP memory to make sure the download worked */
            if((status=current->btype->funcs->dxp_test_spectrum_memory(&ioChan, &chan, &zero,
					current->dsp[chan], current->params[chan])) !=DXP_SUCCESS){
                sprintf(info_string, "Failure testing memory of mod %d channel %d",mod, chan);
                dxp_md_error("dxp_dspconfig",info_string,&status);
                return status;
            }
        }
		current = current->next;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Download DSP firmware to one specific detector channel from the given file.
 * It is currently essential that all symbol tables in all DSP programs be 
 * arranged in the same way.  Test download with calls to 
 * dxp_test_spectrum_memory().
 *
 ******************************************************************************/
int XERXES_API dxp_replace_dspconfig(int* detChan, char* filename)
/* int *detChan;			Input: Detector Channel to download new DSP code */
/* char *filename;			Input: Filename to obtain the new DSP config	 */
{

    int ioChan, modChan, status;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and dxpChan number that matches the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_replace_dspconfig",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
	

/* Load the Fippi configuration into the structure */

    if((status=dxp_add_dsp(filename, chosen->btype, &(chosen->dsp[modChan])))!=DXP_SUCCESS){
		sprintf(info_string,"Error loading Dsp %s",filename);
		dxp_md_error("dxp_replace_dspconfig",info_string,&status);
		return status;
	}

/* Download the DSP code to the IO channel and DXP channel. */
        
	if((status=chosen->btype->funcs->dxp_download_dspconfig(&ioChan, &modChan, chosen->dsp[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error downloading DSP to detector number %d",*detChan);
		dxp_md_error("dxp_replace_dspconfig",info_string,&status);
		return status;
	}

/* Now test reading and writing of data into the memory space of each DSP. */

	if((status=chosen->btype->funcs->dxp_test_spectrum_memory(&ioChan,&modChan,&zero, 
			chosen->dsp[modChan], chosen->params[modChan])) !=DXP_SUCCESS){
		sprintf(info_string, "Failure testing detector channel %d",*detChan);
		dxp_md_error("dxp_replace_dspconfig",info_string,&status);
		return status;
	}

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 *  Download parameters stored in memory for all channels
 *
 ******************************************************************************/
int XERXES_API dxp_put_dspparams(VOID)
{

    int status;
	int ioChan; 
	unsigned short used;
	unsigned int chan;
	int ichan;
/* Pointer to the current Board */
	Board *current = system_head;

/* Loop over all modules in the system */
	
	while (current!=NULL) {
		ioChan = current->ioChan;
		used = current->used;

        for(chan=0;chan<current->nchan;chan++){
            if ((used&(1<<chan))==0) continue;
/* Loop over the parameters requested, writing each value[] to addr[]. */
        
			ichan = (int) chan;
			if((status=current->btype->funcs->dxp_write_dspparams(&ioChan,&ichan, current->dsp[chan], current->params[chan]))!=DXP_SUCCESS){
				sprintf(info_string,"Error writing parameters to module %d",current->mod);
				dxp_md_error("dxp_put_params",info_string,&status);
				return status;
			}
		}
		current = current->next;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Download previously loaded by dxp_dspdefaults() params to designated detector 
 * channel.
 *
 ******************************************************************************/
int XERXES_API dxp_replace_dspparams(int* detChan)
/* int *detChan;					Input: detector channel to download to	*/
{

    int status;
	int ioChan, modChan;
/* Pointer to chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_replace_dspparams",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
	
/* Loop over the parameters requested, writing each value[] to addr[]. */
        
	if((status=chosen->btype->funcs->dxp_write_dspparams(&ioChan, &modChan, chosen->dsp[modChan], chosen->params[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error writing parameters to detector %d",*detChan);
		dxp_md_error("dxp_replace_params",info_string,&status);
		return status;
	}

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Upload DSP params from designated detector channel.
 *
 ******************************************************************************/
int XERXES_API dxp_upload_dspparams(int* detChan)
/* int *detChan;					Input: detector channel to download to	*/
{
    int status;
	int ioChan, modChan;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */
	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_upload_dspparams",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
	
/* Make sure that a params array in Board structure exists */
	if (chosen->params==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,
			"Something wrong: %d detector channel has no allocated parameter memory", *detChan);
		dxp_md_error("dxp_upload_dspparams",info_string,&status);
		return status;
	}

/* Loop over the parameters requested, writing each value[] to addr[]. */
        
	if((status=chosen->btype->funcs->dxp_read_dspparams(&ioChan, &modChan, 
									chosen->dsp[modChan], chosen->params[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error reading parameters from detector %d",*detChan);
		dxp_md_error("dxp_upload_dspparams",info_string,&status);
		return status;
	}

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Routine to echo the global static arrays.  Returns the number of DXP modules
 * in the system, the array of CAMAC modules, and the array designating which 
 * channels are in use.
 *
 ******************************************************************************/
int XERXES_API dxp_get_electronics(int* totMod, int ioChan[], int used[])
/* int *totMod;					*/
/* int ioChan[];				*/
/* int used[];					*/
{

    int mod,status;
/* Pointer to the current Board */
	Board *current = system_head;

/* If no channels, then nothing to report. */
	
	if (numDxpMod==0){
        status = DXP_NOCHANNELS;
        dxp_md_error("dxp_get_electronics",
          "No DXP modules defined: call dxp_assign_channel, define DXP_MODULE",
          &status);
        return status;
    }

/* Copy all the data for the user */

    *totMod = numDxpMod;
	mod = 0;
	while (current != NULL) {
        ioChan[mod] = current->ioChan;
        used[mod] = current->used;
/* Bump the pointers */
		mod++;
		current = current->next;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 * 
 * Return the number of DXP channels in the system and the array of 
 * user-defined detector channel numbers.
 * 
 ******************************************************************************/
int XERXES_API dxp_get_detectors(int* totChan,int detChan[])
/* int *totChan;						*/
/* int detChan[];						*/
{

    unsigned int i;
	int echan, status;
/* Pointer to the current Board */
	Board *current = system_head;

/* If there are no channels, then why are we here? */

    if (numDxpChan==0){
        status = DXP_NOCHANNELS;
        dxp_md_error("dxp_get_detectors",
          "No detectors defined: call dxp_assign_channel, define DXP_MODULE",
          &status);
        return status;
    }

/* Pull out the information and pass it back to the caller. */

	*totChan = numDxpChan;
	echan = 0;
	while (current != NULL) {
		for (i=0; i<current->nchan; i++) {
			if ((current->detChan[i])>-1) {
				detChan[echan] = current->detChan[i];
				echan++;
			}
		}
		current = current->next;
	}

    return DXP_SUCCESS;
}

/******************************************************************************
 * Routine to return the length of the spectrum memory.
 *
 * For 4C-2X boards, this value is stored in the DSP and dynamic.  
 * For 4C boards, it is fixed.
 * 
 ******************************************************************************/
int XERXES_API dxp_nspec(int* detChan, unsigned int* nspec)
/* int *detChan;				Input: Detector channel number				*/
/* unsigned int *nspec;			Output: The number of bins in the spectrum	*/
{
    int status=DXP_SUCCESS;
	int modChan;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */
	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_nspec",info_string,&status);
		return status;
	}

	*nspec = chosen->btype->funcs->dxp_get_spectrum_length(chosen->dsp[modChan], chosen->params[modChan]);

	return status;
}

/******************************************************************************
 * Routine to return the length of the baseline memory.
 *
 * For 4C-2X boards, this value is stored in the DSP and dynamic.  
 * For 4C boards, it is fixed.
 * 
 ******************************************************************************/
int XERXES_API dxp_nbase(int* detChan, unsigned int* nbase)
/* int *detChan;				Input: Detector channel number				*/
/* unsigned int *nbase;			Output: The number of bins in the baseline */
{
	int status = DXP_SUCCESS;
	int modChan;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */
	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_nspec",info_string,&status);
		return status;
	}

	*nbase = chosen->btype->funcs->dxp_get_baseline_length(chosen->dsp[modChan], chosen->params[modChan]);

	return status;
}

/******************************************************************************
 * Routine to return the length of the event memory.
 *
 * For 4C-2X boards, this value is stored in the DSP and dynamic.  
 * For 4C boards, it is fixed.
 * 
 ******************************************************************************/
int XERXES_API dxp_nevent(int* detChan, unsigned int* nevent)
/* int *detChan;					Input: Detector channel number				*/
/* unsigned int *nevent;			Output: The number of samples in event memory*/
{
	int status = DXP_SUCCESS;
	int modChan;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */
	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_nspec",info_string,&status);
		return status;
	}

	*nevent = chosen->btype->funcs->dxp_get_event_length(chosen->dsp[modChan], chosen->params[modChan]);

	return status;
}

/******************************************************************************
 * Routine to return the number of DXP modules in the system.
 *
 ******************************************************************************/
int XERXES_API dxp_ndxp(int* ndxp)
/* int *ndxp;						Output: The number of DXP modules */
{
	int status = DXP_SUCCESS;

	*ndxp = numDxpMod;

	return status;
}

/******************************************************************************
 * Routine to return the number of DXP modules in the system.
 *
 ******************************************************************************/
int XERXES_API dxp_ndxpchan(int* ndxpchan)
/* int *ndxpchan;					Output: The number of detector channels */
{
	int status = DXP_SUCCESS;

	*ndxpchan = numDxpChan;

	return status;
}

/********------******------*****------******-------******-------******------*****
 * Routines to control the taking of data and get the data from the modules.
 * These include starting and stopping runs and reading spectra from the system.
 *
 ********------******------*****------******-------******-------******------*****/

/******************************************************************************
 *
 * Enable the LAM (Look At Me) mechanism for this channel.  
 *
 ******************************************************************************/
int XERXES_API dxp_enable_LAM(int* detChan)
/* int *detChan;					Input: Detector Channel number	*/
{
	int status=DXP_SUCCESS;
	int modChan, ioChan;
	Board *chosen = NULL;
	

/* If the detector channel is not -1, then process only that channel */
	if (*detChan != -1) {	
/* Get the ioChan and dxpChan matching the detChan */

		if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
			sprintf(info_string,"Unknown Detector Channel %d",*detChan);
			dxp_md_error("dxp_enable_LAM",info_string,&status);
			return status;
		}
		ioChan = chosen->ioChan;

/* Enable the LAM for this module */

		if((status=chosen->btype->funcs->dxp_look_at_me(&ioChan, &modChan))!=DXP_SUCCESS){
			sprintf(info_string,"Error enabling the LAM for detector %d",*detChan);
			dxp_md_error("dxp_enable_LAM",info_string,&status);
			return status;
		}
	} else {
/* Else enable LAMs on all boards */
		chosen = system_head;
/* Check that there are any boards defined in the system */
		if (chosen==NULL) {
			status = DXP_INITIALIZE;
			sprintf(info_string,"No Boards defined in the system, please initialize first");
			dxp_md_error("dxp_enable_LAM",info_string,&status);
			return status;
		}
/* Loop over all boards */
		while (chosen) {
			if((status=chosen->btype->funcs->dxp_look_at_me(&(chosen->ioChan), &allChan))!=DXP_SUCCESS){
				sprintf(info_string,"Error enabling the LAM for module %d",chosen->mod);
				dxp_md_error("dxp_enable_LAM",info_string,&status);
				return status;
			}
			chosen = chosen->next;
		}
	}


	return status;
}
	
/******************************************************************************
 *
 * Disable the LAM (Look At Me) mechanism for this channel.  
 *
 ******************************************************************************/
int XERXES_API dxp_disable_LAM(int* detChan)
/* int *detChan;					Input: Detector Channel number	*/
{
	int status=DXP_SUCCESS;
	int modChan, ioChan;
	Board *chosen = NULL;
	
/* If the detector channel is not -1, then process only that channel */
	if (*detChan != -1) {	
/* Get the ioChan and dxpChan matching the detChan */

		if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
			sprintf(info_string,"Unknown Detector Channel %d",*detChan);
			dxp_md_error("dxp_disable_LAM",info_string,&status);
			return status;
		}
		ioChan = chosen->ioChan;

/* Disable the LAM for this module */

		if((status=chosen->btype->funcs->dxp_ignore_me(&ioChan, &modChan))!=DXP_SUCCESS){
			sprintf(info_string,"Error disabling the LAM for detector %d",*detChan);
			dxp_md_error("dxp_disable_LAM",info_string,&status);
			return status;
		}
	} else {
/* Else disable LAMs on all boards */
		chosen = system_head;
/* Check that there are any boards defined in the system */
		if (chosen==NULL) {
			status = DXP_INITIALIZE;
			sprintf(info_string,"No Boards defined in the system, please initialize first");
			dxp_md_error("dxp_disable_LAM",info_string,&status);
			return status;
		}
/* Loop over all boards */
		while (chosen) {
			if((status=chosen->btype->funcs->dxp_ignore_me(&(chosen->ioChan), &allChan))!=DXP_SUCCESS){
				sprintf(info_string,"Error disabling the LAM for module %d",chosen->mod);
				dxp_md_error("dxp_disable_LAM",info_string,&status);
				return status;
			}
			chosen = chosen->next;
		}
	}

	return status;
}

/******************************************************************************
 *
 * Clear the LAM (Look At Me) mechanism for this channel.  
 *
 ******************************************************************************/
int XERXES_API dxp_reset_LAM(int* detChan)
/* int *detChan;					Input: Detector Channel number	*/
{
	int status=DXP_SUCCESS;
	int modChan, ioChan;
	Board *chosen = NULL;
	

/* If the detector channel is not -1, then process only that channel */
	if (*detChan != -1) {	
/* Get the ioChan and dxpChan matching the detChan */

		if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
			sprintf(info_string,"Unknown Detector Channel %d",*detChan);
			dxp_md_error("dxp_reset_LAM",info_string,&status);
			return status;
		}
		ioChan = chosen->ioChan;

/* Clear the LAM for this module */

		if((status=chosen->btype->funcs->dxp_clear_LAM(&ioChan, &modChan))!=DXP_SUCCESS){
			sprintf(info_string,"Error reseting the LAM for detector %d",*detChan);
			dxp_md_error("dxp_reset_LAM",info_string,&status);
			return status;
		}
	} else {
/* Else enable LAMs on all boards */
		chosen = system_head;
/* Check that there are any boards defined in the system */
		if (chosen==NULL) {
			status = DXP_INITIALIZE;
			sprintf(info_string,"No Boards defined in the system, please initialize first");
			dxp_md_error("dxp_reset_LAM",info_string,&status);
			return status;
		}
/* Loop over all boards */
		while (chosen) {
			if((status=chosen->btype->funcs->dxp_clear_LAM(&(chosen->ioChan), &allChan))!=DXP_SUCCESS){
				sprintf(info_string,"Error reseting the LAM for module %d",chosen->mod);
				dxp_md_error("dxp_reset_LAM",info_string,&status);
				return status;
			}
			chosen = chosen->next;
		}
	}


	return status;
}
	
/******************************************************************************
 *
 * This routine initiates a data-taking run on all DXP channels.  If *gate=0,
 * the run does not actually begin until the external gate goes to a TTL high
 * state.  If *resume=0, the MCA is first cleared; if *resume=1, the spectrum
 * is not cleared.
 *
 ******************************************************************************/
int XERXES_API dxp_start_run(unsigned short* gate, unsigned short* resume)
/* unsigned short *gate;			Input: 0 --> use extGate 1--> ignore gate*/
/* unsigned short *resume;			Input: 0 --> clear MCA first; 1--> update*/
{

	int status=DXP_SUCCESS;
	int ioChan, detChan, runactive;
/* Pointer to the current Board */
	Board *current = system_head;

/* gate value, if NULL is passed, use the XerXes stored value */
	unsigned short my_gate;

/* Check for active runs in the system */
	runactive = 0;
	if((status=dxp_isrunning_any(&detChan, &runactive))!=DXP_SUCCESS){
		sprintf(info_string,"Error checking run active status");
		dxp_md_error("dxp_start_run",info_string,&status);
		return status;
	}
	if (runactive != 0) {
		status = DXP_RUNACTIVE;
		if ((runactive&0x1)!=0) {
			sprintf(info_string,"Must stop run in detector %d before beginning a new run.", detChan);
		} else if ((runactive&0x2)!=0) {
			sprintf(info_string,"XerXes thinks there is an active run.  Please execute dxp_stop_run().");
		} else {
			sprintf(info_string,"Unknown reason why run is active.  This should never occur.");
		}
		dxp_md_error("dxp_start_run",info_string,&status);
		return status;
	}

	
/* Nice simple wrapper routine to loop over all modules in the system */
	while (current!=NULL) {
		ioChan = current->ioChan;
/* Check on gate, if not defined, pick up from XerXes */
		if (gate == NULL) {
			my_gate = (unsigned short) current->state[1];
		} else {
			my_gate = *gate;
		}
/* Start the run */
		if((status=current->btype->funcs->dxp_begin_run(&ioChan, &allChan, &my_gate, resume))!=DXP_SUCCESS){
			sprintf(info_string,"error beginning run for module %d",current->mod);
			dxp_md_error("dxp_start_run",info_string,&status);
			return status;
		}
/* Change the system status */
		current->state[0] = 1;				/* Run is active		*/
		current->state[1] = my_gate;		/* Store gate status	*/
/* Point to the next board */
		current = current->next;
    }

    return status;
}

/******************************************************************************
 *
 * This routine allows the user to resume a run that was previously stopped.
 * This will not clear the currently stored histogram in the firmware.
 *
 * If a run is active, no action is performed.
 *
 ******************************************************************************/
int XERXES_API dxp_resume_run()
{
	int status=DXP_SUCCESS;
/* Set resume to true */
	unsigned short resume = 1;

/* Nice simple wrapper routine to loop over all modules in the system */

	if((status=dxp_start_run(NULL, &resume))!=DXP_SUCCESS){
		sprintf(info_string,"Error resuming run");
		dxp_md_error("dxp_resume_run",info_string,&status);
		return status;
	}

    return status;
}

/******************************************************************************
 *
 * Stops runs on all DXP channels.
 *
 ******************************************************************************/
int XERXES_API dxp_stop_run(VOID)
{

    int status=DXP_SUCCESS;
	int ioChan;
/* Pointer to the current Board */
	Board *current = system_head;
    
/* Wrapper to loop over all modules and stop the runs. */
    
	while (current!=NULL) {
		if (current->state[0] == 1) {
/* Change the run active status */
			current->state[0] = 0;		/* Run is not active now */
/* Get the ioChan pointer */
			ioChan = current->ioChan;
		    if((status=current->btype->funcs->dxp_end_run(&ioChan, &allChan))!=DXP_SUCCESS){
				sprintf(info_string,"error ending run for module %d",current->mod);
				dxp_md_error("dxp_stop_run",info_string,&status);
				return status;
			}
		}
		current = current->next;
    }

    return status;

}

/******************************************************************************
 *
 * This routine initiates a data-taking run on all DXP channels.  If *gate=0,
 * the run does not actually begin until the external gate goes to a TTL high
 * state.  If *resume=0, the MCA is first cleared; if *resume=1, the spectrum
 * is not cleared.
 *
 ******************************************************************************/
int XERXES_API dxp_start_one_run(int *detChan, unsigned short* gate, unsigned short* resume)
/* int *detChan;					Input: Detector channel to write to			*/
/* unsigned short *gate;			Input: 0 --> use extGate 1--> ignore gate*/
/* unsigned short *resume;			Input: 0 --> clear MCA first; 1--> update*/
{

	int status=DXP_SUCCESS;
	int ioChan, modChan;		/* IO channel and module channel numbers */
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;
/* Run status */
	int active;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_start_one_run", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Determine if there is already a run active */
	if ((status  = dxp_isrunning(detChan, &active))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to determine run status of detector channel %d", *detChan);
		dxp_md_error("dxp_start_one_run", info_string, &status);
		return status;
	}

	if(active == 0) {
/* Continue with starting new run */
		ioChan = chosen->ioChan;
		if((status=chosen->btype->funcs->dxp_begin_run(&ioChan, &allChan, gate, resume))!=DXP_SUCCESS){
			sprintf(info_string,"error beginning run for module %d",chosen->mod);
			dxp_md_error("dxp_start_one_run",info_string,&status);
			return status;
		}
/* Change the system status */
		chosen->state[0] = 1;				/* Run is active		*/
		chosen->state[1] = (short) *gate;	/* Store gate status	*/
/* Point to the next board */
	} else if ((active&0x1)!=0) {
		status = DXP_RUNACTIVE;
        sprintf(info_string,"Run already active in module %d, please end before starting new run", 
												chosen->mod);
		dxp_md_error("dxp_start_one_run",info_string,&status);
	} else if ((active&0x2)!=0) {
		status = DXP_RUNACTIVE;
        sprintf(info_string,"Xerxes thinks a run is active in module %d, please end before starting new run", 
												chosen->mod);
		dxp_md_error("dxp_start_one_run",info_string,&status);
	} else {
		status = DXP_RUNACTIVE;
        sprintf(info_string,"Unknown reason run is active in module %d.  This should never occur.", 
												chosen->mod);
		dxp_md_error("dxp_start_one_run",info_string,&status);
	}

    return status;
}

/******************************************************************************
 *
 * This routine allows the user to resume a run that was previously stopped.
 * This will not clear the currently stored histogram in the firmware.
 *
 ******************************************************************************/
int XERXES_API dxp_resume_one_run(int *detChan)
{
	int status=DXP_SUCCESS;
/* Set resume to true */
	unsigned short resume = 1;

/* Nice simple wrapper routine to loop over all modules in the system */

	if((status=dxp_start_one_run(detChan, NULL, &resume))!=DXP_SUCCESS){
		sprintf(info_string,"Error resuming run for detector %d", *detChan);
		dxp_md_error("dxp_resume_one_run",info_string,&status);
		return status;
	}

    return status;
}

/******************************************************************************
 *
 * Stops runs on all DXP channels.
 *
 ******************************************************************************/
int XERXES_API dxp_stop_one_run(int *detChan)
/* int *detChan;					Input: Detector channel to write to			*/
{

    int status=DXP_SUCCESS;
	int ioChan, modChan;
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_stop_one_run", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

	if (chosen->state[0] == 1) {
/* Change the run active status */
		chosen->state[0] = 0;		/* Run is not active now */
/* Get the ioChan pointer */
		ioChan = chosen->ioChan;
	    if((status=chosen->btype->funcs->dxp_end_run(&ioChan, &allChan))!=DXP_SUCCESS){
			sprintf(info_string,"error ending run for module %d",chosen->mod);
			dxp_md_error("dxp_stop_one_run",info_string,&status);
			return status;
		}
	}

    return status;

}

/******************************************************************************
 *
 * This routine returns non-zero if the specified module has a run active.
 *
 * Return value has bit 1 set if the board thinks a run is active
 * bit 2 is set if XerXes thinks a run is active
 *
 ******************************************************************************/
int XERXES_API dxp_isrunning(int *detChan, int *ret)
{

	int status=DXP_SUCCESS;
	int ioChan, modChan, runactive;
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Initialize */
	*ret = 0;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_isrunning", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

	if (chosen->state[0] == 1) *ret |= 2;

/* Query the hardware for its status */

    if((status=chosen->btype->funcs->dxp_run_active(&ioChan, &modChan, &runactive))!=DXP_SUCCESS){
		sprintf(info_string,"Error checking run status for module %d",chosen->mod);
		dxp_md_error("dxp_isrunning",info_string,&status);
		return status;
	}

	if (runactive == 1) *ret |= 1;


    return status;
}

/******************************************************************************
 *
 * This routine returns the first detector channel that registers as running.
 * If detChan=-1, then no channel was running, ret will contain if a XerXes thinks
 * any board has an active run.
 *
 * Return value has bit 1 set if the board thinks a run is active
 * bit 2 is set if XerXes thinks a run is active
 *
 ******************************************************************************/
int XERXES_API dxp_isrunning_any(int *detChan, int *ret)
{

	int status=DXP_SUCCESS;
	int runactive, i;
/* Pointer to the Chosen Channel */
	Board *current=system_head;

/* Initialize */
	*detChan = -1;
	*ret = 0;

	while (current!=NULL) {
		for (i=0;i < (int) current->nchan;i++) {
			if ((current->used&(1<<i))!=0) {
				runactive = 0;
/* Get the run status for this detector channel */
				if((status=dxp_isrunning(&(current->detChan[i]), &runactive))!=DXP_SUCCESS){
					sprintf(info_string,"Error determining run status for detector %d",current->detChan[i]);
					dxp_md_error("dxp_isrunning_any",info_string,&status);
					return status;
				}
				*ret |= runactive;
/* Does this channel think a run is active on the physical board.  Return from the routine on the first
 * detector that has an active run */
				if ((runactive&0x1)!=0) {
					*detChan = current->detChan[i];
					return status;
				}
			}
		}
		current = current->next;
	}

    return status;
}

/******************************************************************************
 *
 * Get the index of the symbol defined by name, from within the list of 
 * DSP parameters.  This index can be used to reference the symbol value
 * in the parameter array returned by various routines
 *
 ******************************************************************************/
int XERXES_API dxp_get_symbol_index(int* detChan, char* name, unsigned short* index)
/* int *detChan;				Input: Detector channel to write to			*/
/* char *name;					Input: user passed name of symbol			*/
/* unsigned short *index;		Output: Index of the symbol within parameter array	*/
{
	
	int status;
	int ioChan, modChan;		/* IO channel and module channel numbers */
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_set_one_dspsymbol", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Retrieve the index number */

    if((status=chosen->btype->funcs->dxp_loc(name, chosen->dsp[modChan], index))!=DXP_SUCCESS){
		sprintf(info_string, "Error searching for parameter %s", name);
        dxp_md_error("dxp_get_symbol_index",info_string,&status);
        return status;
    }

    return status;

}

/******************************************************************************
 *
 * Get a parameter value from the DSP .  Given a symbol name, return its
 * value from a given detector number.
 *
 ******************************************************************************/
int XERXES_API dxp_get_one_dspsymbol(int* detChan, char* name, unsigned short* value)
/* int *detChan;					Input: Detector channel to write to			*/
/* char *name;						Input: user passed name of symbol			*/
/* unsigned short *value;			Output: returned value of the parameter		*/
{
	
	int status;
	int ioChan, modChan;		/* IO channel and module channel numbers */
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;
/* Temporary Storage for the long return value */
	unsigned long temp;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_get_one_dspsymbol", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Allow driver to do neede ops to prepare channel for readout */

    if((status=chosen->btype->funcs->dxp_prep_for_readout(&(chosen->ioChan), &modChan))!=DXP_SUCCESS){
        dxp_md_error("dxp_get_one_dspsymbol","Error Preparing board for readout",&status);
        return status;
    }

/* Read the value of the symbol from DSP memory */

    if((status=chosen->btype->funcs->dxp_read_dspsymbol(&ioChan, &modChan, 
								name, chosen->dsp[modChan], &temp))!=DXP_SUCCESS){
		sprintf(info_string, "Error reading parameter %s", name);
        dxp_md_error("dxp_get_one_dspsymbol",info_string,&status);
        return status;
    }

/* Allow driver to return board to previous state */

    if((status=chosen->btype->funcs->dxp_done_with_readout(&(chosen->ioChan), 
											&modChan, chosen->state))!=DXP_SUCCESS){
        dxp_md_error("dxp_get_one_dspsymbol","Error Preparing board for readout",&status);
        return status;
    }

/* Convert to short */
	if (temp>USHRT_MAX) {
		sprintf(info_string, "This routine is deprecated for reading parameters >16bits \nPlease use dxp_get_dspsymbol()");
        dxp_md_error("dxp_get_one_dspsymbol",info_string,&status);
        return status;
	}
	*value = (unsigned short) temp;

    return status;

}

/******************************************************************************
 *
 * Get a parameter value from the DSP .  Given a symbol name, return its
 * value (as a long) from a given detector number.
 *
 ******************************************************************************/
int XERXES_API dxp_get_dspsymbol(int* detChan, char* name, unsigned long* value)
/* int *detChan;					Input: Detector channel to write to			*/
/* char *name;						Input: user passed name of symbol			*/
/* unsigned long *value;			Output: returned value of the parameter		*/
{
	
	int status;
	int ioChan, modChan;		/* IO channel and module channel numbers */
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_get_dspsymbol", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Allow driver to do neede ops to prepare channel for readout */

    if((status=chosen->btype->funcs->dxp_prep_for_readout(&(chosen->ioChan), &modChan))!=DXP_SUCCESS){
        dxp_md_error("dxp_get_dspsymbol","Error Preparing board for readout",&status);
        return status;
    }

/* Read the value of the symbol from DSP memory */

    if((status=chosen->btype->funcs->dxp_read_dspsymbol(&ioChan, &modChan, 
								name, chosen->dsp[modChan], value))!=DXP_SUCCESS){
		sprintf(info_string, "Error reading parameter %s", name);
        dxp_md_error("dxp_get_dspsymbol",info_string,&status);
        return status;
    }

/* Allow driver to return board to previous state */

    if((status=chosen->btype->funcs->dxp_done_with_readout(&(chosen->ioChan), 
											&modChan, chosen->state))!=DXP_SUCCESS){
        dxp_md_error("dxp_get_dspsymbol","Error Preparing board for readout",&status);
        return status;
    }

    return status;

}

/******************************************************************************
 *
 * Set a parameter of the DSP.  Pass the symbol name, value to set and detector
 * channel number.
 *
 ******************************************************************************/
int XERXES_API dxp_set_one_dspsymbol(int* detChan, char* name, unsigned short* value)
/* int *detChan;				Input: Detector channel to write to			*/
/* char *name;					Input: user passed name of symbol			*/
/* unsigned short *value;		Input: Value to set the symbol to			*/
{
	
	int status;
	int ioChan, modChan;		/* IO channel and module channel numbers */
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;
	int runstat = 0;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_set_one_dspsymbol", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Check if there is a run in progress on this board */
	if ((status = dxp_isrunning(detChan, &runstat))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to determine the run status of detChan %d", *detChan);
		dxp_md_error("dxp_set_one_dspsymbol", info_string, &status);
		return status;
	}
	if (runstat!=0) {
		status = DXP_RUNACTIVE;
		sprintf(info_string, "You must stop the run before modifying DSP parameters for detChan %d", 
																			*detChan);
		dxp_md_error("dxp_set_one_dspsymbol", info_string, &status);
		return status;
	}
	
/* Write the value of the symbol into DSP memory */

    if((status=chosen->btype->funcs->dxp_modify_dspsymbol(&ioChan, &modChan, name, value, chosen->dsp[modChan]))!=DXP_SUCCESS){
		sprintf(info_string, "Error writing parameter %s", name);
        dxp_md_error("dxp_set_one_dspsymbol",info_string,&status);
        return status;
    }

    return status;

}

/******************************************************************************
 *
 * Set a parameter of the DSP in all channels.  Pass the symbol name 
 * and value to set.
 *
 ******************************************************************************/
int XERXES_API dxp_set_dspsymbol(char* name, unsigned short* value)
/* char *name;					Input: user passed name of symbol			*/
/* unsigned short *value;		Input: Value to set the symbol to			*/
{

	int status=DXP_SUCCESS;
	int i;
/* Pointer to the Current Channel */
	Board *current = system_head;
	int runstat = 0;

/* Loop over the detector channels and set the value */

	while (current!=NULL) {
		for (i=0; i<(int) current->nchan; i++) {
/* Check if there is a run in progress on this board */
			if ((status = dxp_isrunning(&(current->detChan[i]), &runstat))!=DXP_SUCCESS) {
				sprintf(info_string, "Failed to determine the run status of detChan %d", 
																	current->detChan[i]);
				dxp_md_error("dxp_set_one_dspsymbol", info_string, &status);
				return status;
			}
			if (runstat!=0) {
				status = DXP_RUNACTIVE;
				sprintf(info_string, "You must stop the run before modifying DSP parameters for detChan %d", 
																	current->detChan[i]);
				dxp_md_error("dxp_set_one_dspsymbol", info_string, &status);
				return status;
			}
	
			if ((status  = dxp_set_one_dspsymbol(&(current->detChan[i]), 
												name, value))!=DXP_SUCCESS) {
				sprintf(info_string, "Failed to set %s in detector channel %d", 
												name, current->detChan[i]);
				dxp_md_error("dxp_set_dspsymbol", info_string, &status);
				return status;
			}
		}
		current = current->next;
	}

    return status;

}

/******************************************************************************
 * Routine to return a list of DSP symbol names to the user.
 *
 * The user must allocate the memory for the list of symbols and the 
 * integer containing the number of symbols.  If dynamic allocation 
 * is desired, then a call to dxp_num_symbols() should be made and 
 * the memory allocated accordingly.  All symbols have a maximum size
 * of MAXSYMBOL_LEN characters.
 *
 ******************************************************************************/
int XERXES_API dxp_symbolname_list(int* detChan, char** list)
/* int *detChan;					Input: detector channel number	*/
/* char **list;						Output: List of DSP symbol names */
{
	int status=DXP_SUCCESS;
	int modChan;
	int i;	
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_symbolname_list", info_string, &status);
		return status;
	}
	
	if (list==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"No Memory Allocated for symbolnames");
		dxp_md_error("dxp_symbolname_list",info_string,&status);
		return status;
	}

/* Copy the list of parameter names */
	for (i=0;i<chosen->dsp[modChan]->params->nsymbol;i++) {
		strcpy(list[i], chosen->dsp[modChan]->params->pnames[i]);
	}

	return status;
}

/******************************************************************************
 * Routine to return the DSP symbol name at index i.
 *
 * All symbols have a maximum size of MAXSYMBOL_LEN characters.
 *
 ******************************************************************************/
int XERXES_API dxp_symbolname_by_index(int* detChan, unsigned short *index, char* name)
/* int *detChan;					Input: detector channel number			*/
/* unsigned short *index;			Input: index of the symbol of interest	*/
/* char *name;						Output: Name of DSP symbol				*/
{
	int status=DXP_SUCCESS;
	int modChan;
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_symbolname_by_index", info_string, &status);
		return status;
	}

/* Make sure there is allocated memory for the name */

	if (name==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"No Memory Allocated for symbolname");
		dxp_md_error("dxp_symbolname_by_index",info_string,&status);
		return status;
	}

/* Check if the supplied index is greater than the number of symbols.  Symbols
 * are indexed starting at 0. */

	if (*index >= chosen->dsp[modChan]->params->nsymbol) {
		status = DXP_INDEXOOB;
		sprintf(info_string,"Index %i greater than allowed 0-%i",*index, 
								chosen->dsp[modChan]->params->nsymbol);
		dxp_md_error("dxp_symbolname_by_index",info_string,&status);
		return status;
	}

/* Copy the symbolname */
	strcpy(name, chosen->dsp[modChan]->params->pnames[*index]);

	return status;
}

/******************************************************************************
 * Routine to return a list of DSP symbol information to the user.
 *
 * The user must allocate the memory for the list of read-write codes,
 * lower and uppper bounds.  If dynamic allocation is desired, then a 
 * call to dxp_num_symbols() should be made and 
 * the memory allocated accordingly.  
 *
 ******************************************************************************/
int XERXES_API dxp_symbolname_limits(int* detChan, unsigned short *access, 
						  unsigned short *lbound, unsigned short *ubound)
/* int *detChan;				Input: detector channel number						*/
/* unsigned short *access;		Output: Access information for each DSP parameter	*/
/* unsigned short *lbound;		Output: lower bound for each DSP parameter			*/
/* unsigned short *ubound;		Output: upper bound for each DSP parameter			*/
{
	int status=DXP_SUCCESS;
	int modChan;
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_symbolname_limits", info_string, &status);
		return status;
	}

/* Was Memory allocated for the access list? */
	if (access==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"No Memory Allocated for access information");
		dxp_md_error("dxp_symbolname_limits",info_string,&status);
		return status;
	}

/* Copy the list of access information */
	memcpy(access, chosen->dsp[modChan]->params->access, 
		chosen->dsp[modChan]->params->nsymbol*sizeof(unsigned short));

/* Was Memory allocated for the lower bounds list? */
	if (lbound==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"No Memory Allocated for lower bounds information");
		dxp_md_error("dxp_symbolname_limits",info_string,&status);
		return status;
	}

/* Copy the list of lower bounds information */
	memcpy(lbound, chosen->dsp[modChan]->params->lbound, 
		chosen->dsp[modChan]->params->nsymbol*sizeof(unsigned short));

/* Was Memory allocated for the upper bounds list? */
	if (ubound==NULL) {
		status = DXP_NOMEM;
		sprintf(info_string,"No Memory Allocated for upper bounds information");
		dxp_md_error("dxp_symbolname_limits",info_string,&status);
		return status;
	}

/* Copy the list of upper bounds information */
	memcpy(ubound, chosen->dsp[modChan]->params->ubound, 
		chosen->dsp[modChan]->params->nsymbol*sizeof(unsigned short));

	return status;
}

/******************************************************************************
 *
 * Routine to return the number of DSP symbols.
 *
 ******************************************************************************/
int XERXES_API dxp_max_symbols(int* detChan, unsigned short* nsymbols)
/* int *detChan;						Input: Detector Channel Number			 */
/* unsigned short *nsymbols;			Output: The number of symbols in the list */
{
	int status=DXP_SUCCESS;
	int modChan;
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_max_symbols", info_string, &status);
		return status;
	}

	*nsymbols = chosen->dsp[modChan]->params->nsymbol;

	return status;
}

/******************************************************************************
 * Routine to dump the memory contents of the DSP.
 * 
 * This routine prints the memory contents of the DSP in detector 
 * channel detChan
 *
 ******************************************************************************/
int XERXES_API dxp_mem_dump(int* detChan)
/* int *detChan;						Input: detector channel of to dump        */
{
	int status;
	int ioChan, modChan;		/* IO channel and module channel numbers */
/* Pointer to the Chosen Board */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status  = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_mem_dump", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Write the value of the symbol into DSP memory */

    if((status=chosen->btype->funcs->dxp_dspparam_dump(&ioChan, &modChan, chosen->dsp[modChan]))!=DXP_SUCCESS){
		sprintf(info_string, "Error dumping DSP memory of detector %d", *detChan);
        dxp_md_error("dxp_mem_dump",info_string,&status);
        return status;
    }

    return status;

}
/******************************************************************************
 *
 * This routine reads out the paramter memory, the baseline memory and the
 * MCA spectrum for a single DXP channel.  The error words in the parameter
 * memory are first decoded and checked.  If OK, the baseline and spectrum
 * memories are read, otherwise the error is cleared and this routine is
 * exited.  Note: spectrum words are 32 bits, wheras params and baseline are
 * only 16 bits.
 *
 ******************************************************************************/
int XERXES_API dxp_readout_detector_run(int* detChan, unsigned short params[],
							 unsigned short baseline[], unsigned long spectrum[])
/* int *detChan;						Input: detector channel number      */
/* unsigned short params[];				Output: parameter memory            */
/* unsigned short baseline[];			Output: baseline histogram          */
/* unsigned long spectrum[];			Output: spectrum                    */
{

	int status;

	int ioChan, modChan;
/* Pointer to the Chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_readout_detector_run",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
	
/* wrapper for the real readout routine */

	if((status=dxp_readout_run(chosen, &modChan, params, baseline, spectrum))!=DXP_SUCCESS){
		dxp_md_error("dxp_readout_detector_run","error reading out run data",&status);
		return status;
	}

    return status;
}

/******************************************************************************
 *
 * This routine reads out the paramter memory, the baseline memory and the
 * MCA spectrum for a single DXP channel.  The error words in the parameter
 * memory are first decoded and checked.  If OK, the baseline and spectrum
 * memories are read, otherwise the error is cleared and this routine is
 * exited.  Note: spectrum words are 32 bits, wheras params and baseline are
 * only 16 bits.
 *
 ******************************************************************************/
static int XERXES_API dxp_readout_run(Board* board, int* modChan, unsigned short params[],
						   unsigned short baseline[], unsigned long spectrum[])
/* Board *board;						Input: I/O channel number of module */
/* int *modChan;						Input: DXP channel number 0,1,2,3   */
/* unsigned short params[];				Output: parameter memory            */
/* unsigned short baseline[];			Output: baseline histogram          */
/* unsigned long spectrum[];			Output: spectrum                    */
{

	int status;

/* Allow driver to do neede ops to prepare channel for readout */

    if((status=board->btype->funcs->dxp_prep_for_readout(&(board->ioChan), &allChan))!=DXP_SUCCESS){
        dxp_md_error("dxp_readout_run","Error Preparing board for readout",&status);
        return status;
    }

/* Perform the read */

    if((status=dxp_do_readout(board, modChan, params, baseline, spectrum))!=DXP_SUCCESS){
        dxp_md_error("dxp_readout_run","error reading data",&status);
        return status;
    }


/* Allow driver to return board to previous state */

    if((status=board->btype->funcs->dxp_done_with_readout(&(board->ioChan), 
											&allChan, board->state))!=DXP_SUCCESS){
        dxp_md_error("dxp_readout_run","Error Preparing board for readout",&status);
        return status;
    }

    return status;
}

/******************************************************************************
 *
 * This routine reads out the paramter memory, the baseline memory and the
 * MCA spectrum for a single DXP channel.  The error words in the parameter
 * memory are first decoded and checked.  If OK, the baseline and spectrum
 * memories are read, otherwise the error is cleared and this routine is
 * exited.  Note: spectrum words are 32 bits, wheras params and baseline are
 * only 16 bits.
 *
 ******************************************************************************/
static int XERXES_API dxp_do_readout(Board* board, int* modChan, unsigned short params[],
						   unsigned short baseline[], unsigned long spectrum[])
/* Board *board;						Input: I/O channel number of module */
/* int *modChan;						Input: DXP channel number 0,1,2,3   */
/* unsigned short params[];				Output: parameter memory            */
/* unsigned short baseline[];			Output: baseline histogram          */
/* unsigned long spectrum[];			Output: spectrum                    */
{

	int status, error;
	unsigned short runerror,errinfo; 

	unsigned short *temp_params=NULL;

/* Read out the parameters from the DSP memory */

/* Must always read parameters, if the user doesn't want them, then we have to allocate 
	for ourselves. */
	if (params==NULL) {
		temp_params = (unsigned short *) dxp_md_alloc(sizeof(unsigned short)*board->dsp[*modChan]->params->nsymbol);
	} else {
		temp_params = params;
	}

    if(temp_params==NULL) {
		status = DXP_NOMEM;
		dxp_md_error("dxp_do_readout","Unable to allocate memory for parameter array",&status);
		return status;
	}

/* Now we are sure to have allocated memory */
	
	if((status=board->btype->funcs->dxp_read_dspparams(&(board->ioChan), modChan, board->dsp[*modChan], temp_params))!=DXP_SUCCESS){
/* Clean up after ourselves */
		if (params==NULL) {
			dxp_md_free(temp_params);
		}
		dxp_md_error("dxp_do_readout","error reading parameters",&status);
		return status;
	}

/* Pull out the RUNERROR and ERRINFO parameters from the DSP parameter list
 * and act on it accordingly */

    if((status=board->btype->funcs->dxp_decode_error(temp_params, board->dsp[*modChan], 
										&runerror, &errinfo))!=DXP_SUCCESS){
/* Clean up after ourselves */
		if (params==NULL) {
			dxp_md_free(temp_params);
		}
		dxp_md_error("dxp_do_readout"," ",&status);
		return status;
	}
	if(runerror!=0){
/* Clean up after ourselves */
		if (params==NULL) {
			dxp_md_free(temp_params);
		}
/* Continue */
		error = DXP_DSPRUNERROR;
		sprintf(info_string,"RUNERROR = %d ERRINFO = %d",runerror,errinfo);
		dxp_md_error("dxp_readout_run",info_string,&error);
		if((status=board->btype->funcs->dxp_clear_error(&(board->ioChan), modChan, board->dsp[*modChan]))!=DXP_SUCCESS){
			dxp_md_error("dxp_do_readout"," ",&status);
			return status;
		}
		return error;
	}

/* Read out the spectrum memory now (also known as MCA memory) */

/* Only read spectrum if there is allocated memory */
	if (spectrum!=NULL) {
	    if((status=board->btype->funcs->dxp_read_spectrum(&(board->ioChan),modChan,
				board->dsp[*modChan], temp_params, spectrum))!=DXP_SUCCESS){
/* Clean up after ourselves */
			if (params==NULL) {
				dxp_md_free(temp_params);
			}
			dxp_md_error("dxp_do_readout","error reading out spectrum",&status);
			return status;
		}
	}

/* Read out the basline histogram. */

/* Only read baseline if there is allocated memory */
	if (baseline!=NULL) {
	    if((status=board->btype->funcs->dxp_read_baseline(&(board->ioChan),modChan,
				board->dsp[*modChan], temp_params, baseline))!=DXP_SUCCESS){
/* Clean up after ourselves */
			if (params==NULL) {
				dxp_md_free(temp_params);
			}
			dxp_md_error("dxp_do_readout","error reading out baseline",&status);
			return status;
		}
	}

/* Clean up after ourselves */
	if (params==NULL) {
		dxp_md_free(temp_params);
	}

    return status;
}

/******************************************************************************
 * 
 * Preform internal gain calibration or internal TRACKDAC reset point 
 * for all DXP channels:
 *    o save the current value of RUNTASKS for each channel
 *    o start run
 *    o wait
 *    o stop run
 *    o readout parameter memory
 *    o check for errors, clear errors if set
 *    o check calibration results..
 *    o restore RUNTASKS for each channel
 *
 ******************************************************************************/
int XERXES_API dxp_calibrate(int* calibtask)
/* int *calibtask;							Input: which calibration function */
{
    int status=DXP_SUCCESS;
/* Pointer to the Current Board */
	Board *current=system_head;

/* Perform Calibration on each module in the system */
	while (current!=NULL) {
		if ((status=current->btype->funcs->dxp_calibrate_channel(&(current->mod), 
							&(current->ioChan), &(current->used), calibtask, current->dsp))!=DXP_SUCCESS) {
			sprintf(info_string,"Failed to calibrate module %d",current->mod);
			dxp_md_error("dxp_calibrate",info_string,&status);
			return status;
		}
		current = current->next;
	}

    return status;
}

/******************************************************************************
 * 
 * Preform internal gain calibration or internal TRACKDAC reset point 
 * for a single DXP channels (module):
 *    o save the current value of RUNTASKS for each channel
 *    o start run
 *    o wait
 *    o stop run
 *    o readout parameter memory
 *    o check for errors, clear errors if set
 *    o check calibration results..
 *    o restore RUNTASKS for each channel
 *
 ******************************************************************************/
int XERXES_API dxp_calibrate_one_channel(int *detChan, int* calibtask)
/* int *detChan;							Input: which channel to perform calibration */
/* int *calibtask;							Input: which calibration function */
{
    int status=DXP_SUCCESS;
	int modChan;
/* Pointer to the Chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and modChan number that matches the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_calibrate_one_channel",info_string,&status);
		return status;
	}

/* Perform the calibration on this board */
	if ((status=chosen->btype->funcs->dxp_calibrate_channel(&(chosen->mod), 
						&(chosen->ioChan), &(chosen->used), calibtask, chosen->dsp))!=DXP_SUCCESS) {
		sprintf(info_string,"Failed to calibrate module %d",chosen->mod);
		dxp_md_error("dxp_calibrate_one_channel",info_string,&status);
		return status;
	}

    return status;
}

/********------******------*****------******-------******-------******------*****
 * Routines to save and restore data.  This involves writing spectra to files, 
 * opening and closing of files, etc...
 *
 ********------******------*****------******-------******-------******------*****/

/******************************************************************************
 *
 * Read out parameter, baseline, and spectrum memories for all DXP channels
 * Writes all spectra to file associated with luns and all baselines to file
 * associated with lunb.  The output format of these files is N+1 columns,
 * where N is the total number of channels.  The first column is the bin
 * number and the remaining columns are the histograms, one each for the 
 * different channels.
 *
 * NOTE: must call dxp_open_file (or DXP_OPEN_FILE_FORTRAN) prior to calling
 * this routine.  if *luns<0, no spectrum will be written; if *lunb<0, 
 * no baseline will be written;
 *
 ******************************************************************************/
int XERXES_API dxp_write_spectra(int* luns, int* lunb)
/* int *luns;							Input: logical unit no. for spectra  */
/* int *lunb;							Input: logical unit no. for baseline */
{
    int mod=0,chan=0,status=DXP_SUCCESS,echan;
	int allChan=ALLCHAN;

	unsigned int i, j;

	unsigned int maxspec=0,maxbase=0;
    unsigned long **spectra=NULL;
    unsigned short **baseline=NULL;

    unsigned int evts=0,under=0,over=0,fast=0,base=0;
    double live=0.,icr=0.,ocr=0.;
    FILE *fp;
/* Pointer to the Current Board */
	Board *current=system_head;

/*
 * check to see if files have been opened for writing:
 */
    if(*luns>=0){
        if ((*luns>=MAXFILE)||(filpnt[*luns]==NULL)){
            status=DXP_BAD_PARAM;
            dxp_md_error("write_spectra","Output file for spectra not opened",
                &status);
            return status;
        }
    }
    if(*lunb>=0){
        if ((*lunb>=MAXFILE)||(filpnt[*lunb]==NULL)){
            status=DXP_BAD_PARAM;
            dxp_md_error("write_spectra","Output file for baseline not opened",
                &status);
            return status;
        }
    }

/* Allocate memory for the baseline and spectrum memory */

    if ((spectra=(unsigned long **)dxp_md_alloc(numDxpChan*sizeof(unsigned long *)))==NULL) {
		status=DXP_NOMEM;
		dxp_md_error("write_spectra","Out of Memory",&status);
		return status;
	}
    if ((baseline=(unsigned short **)dxp_md_alloc(numDxpChan*sizeof(unsigned short *)))==NULL) {
		status=DXP_NOMEM;
		dxp_md_error("write_spectra","Out of Memory",&status);
		return status;
	}

/* Readout the all information for each run: parameter list, baseline histogram
 * and spectrum memory.  Do this for each channel on every module in the system. */

	current = system_head;
    echan = 0;    
	while (current!=NULL) {
		mod = current->mod;

/* Allow driver to do neede ops to prepare channel for readout */

	    if((status=current->btype->funcs->dxp_prep_for_readout(&(current->ioChan), &allChan))!=DXP_SUCCESS){
		    dxp_md_error("dxp_write_spectra","Error Preparing board for readout",&status);
			return status;
		}

/* Loop over the channels */
        for(chan=0;chan<(int) current->nchan;chan++){
            if(((current->used)&(1<<chan))!=0){

/* Get the length of the spectrum and baseline memories from the primitive lib */
				maxspec = current->btype->funcs->dxp_get_spectrum_length(current->dsp[chan], 
											current->params[chan]);
				maxbase = current->btype->funcs->dxp_get_baseline_length(current->dsp[chan], 
											current->params[chan]);
	
				if ((spectra[echan]=(unsigned long *)
								dxp_md_alloc(maxspec*sizeof(unsigned long)))==NULL) {
					status=DXP_NOMEM;
					dxp_md_error("write_spectra","Out of Memory",&status);
					return status;
				}
				if ((baseline[echan]=(unsigned short *)
								dxp_md_alloc(maxbase*sizeof(unsigned short)))==NULL) {
					status=DXP_NOMEM;
					dxp_md_error("write_spectra","Out of Memory",&status);
					return status;
				}

/* Readout the run data.  Parameter list, baseline histograms and spectrum memory */

                if((status=dxp_do_readout(current, &chan, current->params[chan],
								baseline[echan], spectra[echan]))!=DXP_SUCCESS){
                    sprintf(info_string,"module %d chan %d",mod,chan);
                    dxp_md_error("dxp_write_spectra",info_string,&status);

/* If readout fails, clean up after myself.  */
    
					echan = 0;
					current = system_head;
					while (current!=NULL) {
					    for(i=0;i<current->nchan;i++){
							if ((current->detChan[i])<0) continue;
						    dxp_md_free(spectra[echan]);
							dxp_md_free(baseline[echan]);
							echan++;
						}
						current = current->next;
				    }
					dxp_md_free(spectra);
					dxp_md_free(baseline);

                    return status;
                }

/* Parse the parameter memory to retrieve the event rates and live time */

				if((status=current->btype->funcs->dxp_get_runstats(current->params[chan], 
						current->dsp[chan],&evts,&under,&over,&fast,&base,&live))!=DXP_SUCCESS){
					dxp_md_error("dxp_write_spectra"," ",&status);

/* Clean up after myself.  */
    
					echan = 0;
					current = system_head;
					while (current!=NULL) {
					    for(i=0;i<current->nchan;i++){
							if ((current->detChan[i])<0) continue;
						    dxp_md_free(spectra[echan]);
							dxp_md_free(baseline[echan]);
							echan++;
						}
						current = current->next;
				    }
					dxp_md_free(spectra);
					dxp_md_free(baseline);

					return status;
				}
				if(live>0.){
					icr = 0.001*(fast)/live;
					ocr = 0.001*(under + over + evts )/live;
				} else icr = ocr = -999.;

/* Print the event rates and livetime to the screen */

				sprintf(info_string,
					"mod %d chan %d live %8.3f ICR %8.2f OCR %8.2f",
					mod,chan,
					live,                       /* livetime (s)      */
					icr,			            /* ICR in kcps       */
					ocr);						/* OCR in kcps       */
				dxp_md_error("dxp_write_spectra",info_string,&debug);

				echan++;
			}
		}
/* Allow driver to return board to previous state */

		if((status=current->btype->funcs->dxp_done_with_readout(&(current->ioChan), 
									&allChan, current->state))!=DXP_SUCCESS){
	        dxp_md_error("dxp_write_spectra","Error Preparing board for readout",&status);
		    return status;
		}

/* Proceed to next board in the system */
		current = current->next;
	}


/* If requested, write the spectrum memory to a file */

	if(*luns>=0){
        fp=filpnt[*luns];

/* Print a little header with the number of channels and 
 * number of words per channel, followed by the detector channel
 * numbers. */

		fprintf(fp, "Number of Channels = %d Maximum Number of Bins = %d", numDxpChan, maxspec);
		fprintf(fp, "\r\n");
		echan = 0;
		current = system_head;
		while (current!=NULL) {
		    for(i=0;i<current->nchan;i++){
				if ((current->detChan[i])<0) continue;
				fprintf(fp, "Detector Channel %d ", current->detChan[i]);
			}
			current = current->next;
		}
		fprintf(fp, "\r\n");
        
/* Now write out the spectra for each channel */

		for(i=0;i < maxspec;i++){
            fprintf(fp,"%d",i);
			echan = 0;
			current = system_head;
			while (current!=NULL) {
			    for(j=0;j<current->nchan;j++){
					if ((current->detChan[j])<0) continue;
		            fprintf(fp," %lu",spectra[echan][i]);
					echan++;
				}
				current = current->next;
			}
            fprintf(fp,"\r\n");
        }
    }

/* If requested, write the baseline histograms to a file */

	if(*lunb>=0){
        fp=filpnt[*lunb];

/* Print a little header with the number of channels and 
 * number of words per channel, followed by the detector channel
 * numbers. */

		fprintf(fp, "Number of Channels = %d Maximum Number of Bins = %d", numDxpChan, maxbase);
		fprintf(fp, "\r\n");
		current = system_head;
		while (current!=NULL) {
		    for(i=0;i<current->nchan;i++){
				if ((current->detChan[i])<0) continue;
				fprintf(fp, "Detector Channel %d ", current->detChan[i]);
			}
			current = current->next;
		}
		fprintf(fp, "\r\n");
        
/* Now write out the baseline for each channel */

        for(i=0;i < maxbase;i++){
            fprintf(fp,"%d",i);
			echan = 0;
			current = system_head;
			while (current!=NULL) {
			    for(j=0;j<current->nchan;j++){
					if ((current->detChan[j])<0) continue;
	                fprintf(fp," %d", ((unsigned int) baseline[echan][i]));
					echan++;
				}
				current = current->next;
			}
            fprintf(fp,"\r\n");
        }
    }

/* Clean up after myself.  */
    
	echan = 0;
	current = system_head;
	while (current!=NULL) {
	    for(i=0;i<current->nchan;i++){
			if ((current->detChan[i])<0) continue;
	        dxp_md_free(spectra[echan]);
			dxp_md_free(baseline[echan]);
			echan++;
		}
		current = current->next;
    }
    dxp_md_free(spectra);
    dxp_md_free(baseline);

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * This routine writes the DSP parameter list read from the specified 
 * detector channel to the file pointed to by lun.  
 *
 * WARNING: dxp_open_file (DXP_OPEN_FILE_FORTRAN) must be called prior to 
 * this routine.  
 *
 ******************************************************************************/
int XERXES_API dxp_save_dspparams(int* detChan, int* lun)
/* int *detChan;						Input: detector channel to get params  */
/* int *lun;							Input: logical unit number to write to */
{
	int status = DXP_SUCCESS;
	int ioChan, modChan, nsymb;
	char *symbolname=NULL;
	FILE *fout;

	unsigned short par, i;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the ioChan and dxpChan number that matches the detChan */

	if((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS){
		sprintf(info_string,"Error finding detector number %d",*detChan);
		dxp_md_error("dxp_save_dspparams",info_string,&status);
		return status;
	}
	ioChan = chosen->ioChan;
	
/* Read the parameters into memory */
        
	if((status=chosen->btype->funcs->dxp_read_dspparams(&ioChan, &modChan, 
					chosen->dsp[modChan], chosen->params[modChan]))!=DXP_SUCCESS){
		sprintf(info_string,"Error writing parameters to detector %d",*detChan);
		dxp_md_error("dxp_save_dspparams",info_string,&status);
		return status;
	}

/* Write the parameters to the file */

/*
 * Make sure the file is previously opened
 */
    fout = filpnt[*lun];
    if(fout==NULL){
        status=DXP_OUTPUT_UNDEFINED;
        dxp_md_error("dxp_save_dspparams","output file not opened",&status);
        return status;
    }
/* 
 * Now attach a header to the file 
 */
    fprintf(fout,"* \r\n");
    fprintf(fout,"* Save file for detector channel %d\r\n", *detChan);
    fprintf(fout,"* \r\n");
    fprintf(fout,"* This file contains the contents of all DSP\r\n");
    fprintf(fout,"* parameters.\r\n");
    fprintf(fout,"* \r\n");
/*
 * Now write out the list
 */
/* Allocate memory for the symbolname */
	symbolname = (char *) dxp_md_alloc(MAXSYMBOL_LEN*sizeof(char));

	nsymb = chosen->dsp[modChan]->params->nsymbol;
	for (i=0; i < nsymb; i++) {

/* Get the symbolname */
		if((status=chosen->btype->funcs->dxp_symbolname(&i, chosen->dsp[modChan], symbolname))!=DXP_SUCCESS){
			dxp_md_free(symbolname);
			sprintf(info_string,"Error getting symbol #%d", i);
			dxp_md_error("dxp_save_dspparams",info_string,&status);
			return status;
		}
/* Write the symbolname-value pair to the file */
		par = chosen->dsp[modChan]->data[i];
		fprintf(fout,"%s\t\t%du\r\n", symbolname, par);
	}

/* Free memory */
	dxp_md_free(symbolname);

/* Write an ender section */
    fprintf(fout,"* \r\n");
    fprintf(fout,"* This file was generated by the XIA DXP Library\r\n");
    fprintf(fout,"* END OF FILE\r\n");

	return status;
}

/******************************************************************************
 *
 * This routine reads the parameter memory of all DXP channels and writes
 * the data in hexidecimal format to the file pointed to by lun.  
 *
 * WARNING: dxp_open_file (DXP_OPEN_FILE_FORTRAN) must be called prior to 
 * this routine.  
 *
 ******************************************************************************/
int XERXES_API dxp_save_config(int* lun)
/* int *lun;							Input: logical unit number to write to */
{

    int chan;
    int status;
	unsigned short i, nsymb;
	unsigned short nsymbols;
    char symbol[MAXSYMBOL_LEN];
    FILE *fout;

	Board_Info *current_btype = NULL;
	char current_ifacename[132];
	char current_dllname[132];
	Fippi_Info *current_fippi = NULL;
	Dsp_Info *current_dsp = NULL;

	unsigned short par;
/* Pointer to the current Board */
	Board *current=system_head;

/*
 * Make sure the file is previously opened
 */
    fout = filpnt[*lun];
    if(fout==NULL){
        status=DXP_OUTPUT_UNDEFINED;
        dxp_md_error("dxp_save_config","output file not opened",&status);
        return status;
    }
/* First put the preamp configuration file at the top of the save file */
	fprintf(fout, "preamp          %s\r\n", info->preamp);

/* Now loop over boards in the system to define all entries */
	while (current!=NULL) {
		if (current->btype != current_btype) {
			current_btype = current->btype;
			fprintf(fout, "board_type      %s\r\n", current_btype->name);
		}
		if (strcmp(current->iface->ioname,current_dllname)!=0) {
			strcpy(current_dllname, current->iface->ioname);
			fprintf(fout, "iolibrary       %s\r\n", current_dllname);
		}
		if (strcmp(current->iface->dllname,current_ifacename)!=0) {
			strcpy(current_ifacename, current->iface->dllname);
			fprintf(fout, "interface       %s\r\n", current_ifacename);
		}
/* Print out the module number and information */
		fprintf(fout, "module          %s ", current->iostring);
		for (i=0;i<current->nchan;i++) {
			fprintf(fout,"%i ", current->detChan[i]);
		}
		fprintf(fout, "\r\n");
/* Now do the FIPPI information */
		current_fippi = current->fippi[0];
		fprintf(fout, "default_fippi   %s\r\n", current->fippi[0]->filename);
		for (i=1;i<current->nchan;i++) {
			if (current_fippi != current->fippi[i]) {
				fprintf(fout, "fippi           %i %s\r\n", i, current->fippi[i]->filename);
			}
		}
/* Now do the DSP information */
		current_dsp = current->dsp[0];
		fprintf(fout, "default_dsp     %s\r\n", current->dsp[0]->filename);
		for (i=1;i<current->nchan;i++) {
			if (current_dsp != current->dsp[i]) {
				fprintf(fout, "dsp             %i %s\r\n", i, current->dsp[i]->filename);
			}
		}
/* Now do the DSP information */
/*		current_defaults = current->defaults[0];
		fprintf(fout, "default_param   %s\r\n", current->defaults[0]->filename);
		for (i=1;i<current->nchan;i++) {
			if (current_defaults != current->defaults[i]) {
				fprintf(fout, "param           %i %s\r\n", i, current->defaults[i]->filename);
			}
		}*/
		fprintf(fout, "default_param   NULL\r\n");
		current = current->next;
	}
    fprintf(fout,"END\r\n");
/*
 *    Now, loop over channels and write parameter memory to configuration
 */
	current = system_head;
	while (current!=NULL) {
        for(chan=0;chan<(int) current->nchan;chan++){
            if( ((current->used)&(1<<chan))==0) continue;
            fprintf(fout,"********  detector channel %d ********\r\n",
                current->detChan[chan]);
/* Now write all the symbols in the DSP memory to the file. */
			nsymbols = current->dsp[chan]->params->nsymbol;
			fprintf(fout,"Symbols = %d\r\n",nsymbols);
			nsymb = 0;
			for(i = 0;i < nsymbols;i++){
				if (nsymb==0) {
					fprintf(fout,"%2i:",i/6);
				}
				current->btype->funcs->dxp_symbolname(&i, current->dsp[chan], symbol);
				fprintf(fout,"%12s ",symbol);
				nsymb++;
				if(nsymb==6){
					fprintf(fout,"\r\n");
					nsymb=0;
				}
			}
			if(nsymb!=0) fprintf(fout,"\r\n");

/* Read in the parameter values from the hardware */
            if((status=current->btype->funcs->dxp_read_dspparams(&(current->ioChan),
                                      &chan, current->dsp[chan], current->params[chan]))!=DXP_SUCCESS){
                sprintf(info_string,
                   "Error reading out parameters for mod %d chan %d",current->mod,chan);
                dxp_md_error("dxp_save_config",info_string,&status);
                return status;
            }
			nsymb = 0;
            for(i=0;i<nsymbols;i++){
				if (nsymb==0) {
					fprintf(fout,"%2i:",i/6);
				}
/* Equate to an unsigned short to avoid sign-extending the parameter */
				par = current->params[chan][i];
                fprintf(fout,"%12X ", par);
				current->btype->funcs->dxp_symbolname(&i, current->dsp[chan], symbol);
				nsymb++;
				if(nsymb==6){
					fprintf(fout,"\r\n");
					nsymb=0;
				}
            }
			if(nsymb!=0) fprintf(fout,"\r\n");
        }
		current = current->next;
    }

	return DXP_SUCCESS;
}

/******************************************************************************
 *
 * This routine reads a configuration file (written by dxp_save_config) and  
 * restores the parameter memory of all DXP channels   
 *
 * WARNING: dxp_open_file (DXP_OPEN_FILE_FORTRAN) must be called prior to 
 * this routine.  
 *
 ******************************************************************************/
int XERXES_API dxp_restore_config(int* lun)
/* int *lun;				Input: logical unit number configuration file  */
{
   
    FILE *fp;  
    char *cstatus,line[132], temp_line[132];
	char strtmp[132];
    int status;
	unsigned short i, j;
    unsigned short addr[MAXSYM],nsymbols,nparam;
	int ioChan, modChan, detChan=0;

	unsigned short par;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

    fp = filpnt[*lun];    
    if(fp==NULL){  
        status=DXP_INPUT_UNDEFINED;  
        dxp_md_error("dxp_restore_config","Input file not opened",&status);  
        return status;  
    }  
/*  
 *   Read configuration file:  num mods    
 */        
    do  
        cstatus=fgets(line,132,fp);    /* skip comments and blank lines*/  
    while((line[0]=='*')||(line[0]=='\r')||(line[0]=='\n'));  
/* Make a copy of the line, since strtok buggers it up */
	strcpy(temp_line, line);

    token=strtok(line,delim);  
	for (i=0;i<strlen(token);i++) 
		token[i] = (char) tolower(token[i]);
	if (strcmp(token, "preamp")!=0) {
        status = DXP_INPUT_UNDEFINED;
        sprintf(info_string,"Save files must start with a preamp entry");
        dxp_md_error("dxp_restore_config",info_string,&status);
        return status;
	}
/* Now search for the filename starting at m+1 */
	dxp_pick_filename(strlen(token)+1, temp_line, strtmp);
/* Now assign the preamp file name to the global structure */
	if (info->preamp != NULL) {
		dxp_md_free(info->preamp);
	}
	info->preamp = (char *) dxp_md_alloc((strlen(strtmp)+1)*sizeof(char));
	strcpy(info->preamp, strtmp);
/*  
 *   Read previously saved modules portion of the configuration file
 */  
    if ((status=dxp_read_modules(fp))!=DXP_SUCCESS){
        status = DXP_INITIALIZE;
        sprintf(info_string,"Could not read in modules file : %s", info->modules);
        dxp_md_error("dxp_restore_config",info_string,&status);
        return status;
    }

/*
 *  Download the FiPPI firmware
 */
	dxp_md_error("dxp_restore_config","calling dxp_fipconfig",&debug);
	if((status=dxp_fipconfig())!=DXP_SUCCESS){
		dxp_md_error("dxp_restore_config","Error Downloading FiPPi(s)",&status);
		return status;
	}

/*
 *  Download the DSP firmware
 */
	dxp_md_error("dxp_restore_config","calling dxp_dspconfig",&debug);
	if((status=dxp_dspconfig())!=DXP_SUCCESS){
		dxp_md_error("dxp_restore_config","Error Downloading DSP(s)",&status);
		return status;
	}

    for(i=0;i<numDxpChan;i++){  
/*  
 *   Read configuration file: Symbols  
 *      store their in-DXP addresses.  
 *    addr[file symbol position] = (symbol in dxp) ? symbol address in DXP: -1  
 */  
/* Get the Detector number */
        fgets(line,132,fp);  
        token=strtok(line,delim);
        for(j=0;j<3;j++) token=strtok(NULL,delim);  
        detChan = (int) strtol(token,NULL,10);  

/* Find Detector number in the system list */
		if((status=dxp_det_to_elec(&detChan, &chosen, &modChan))!=DXP_SUCCESS){
			sprintf(info_string,"Error finding detector number %d", detChan);
			dxp_md_error("dxp_restore_config",info_string,&status);
			return status;
		}
		ioChan = chosen->ioChan;

/*  
 * Read the param block and substitite the values from the file.  Then write the params  
 * back in.
 */  

/* Read in the parameter names */
		fgets(line,132,fp);  
		token=strtok(line,delim);  
		token=strtok(NULL,delim);  
		nsymbols = (unsigned short) strtoul(token,NULL,10);  
		memset(addr, 0, sizeof(addr));
		nparam = 0;
		while(fgets(line,132,fp)!=NULL){  
			token=strtok(line,delim);  
			token=strtok(NULL,delim);  
			while(token!=NULL) {  
				if(isalnum(*token)!=0){  
					chosen->btype->funcs->dxp_loc(token, chosen->dsp[modChan], &addr[nparam]);
					nparam++;
				}  
				token=strtok(NULL,delim);  
			}
			if(nparam==nsymbols) break;  
		}  

/*
 * Read configuration file:  params  
 * Load the parameter values
 */
        nparam=0;
        while(fgets(line,132,fp)!=NULL){  
            token=strtok(line,delim);
            token = strtok(NULL,delim);  
            while(token!=NULL){
                if(isxdigit(*token)!=0) {
					sscanf(token,"%hX",&par);
					chosen->params[modChan][addr[nparam]] = par;
					nparam++;
				}
                token = strtok(NULL,delim);  
            }  
            if(nparam==nsymbols) break;  
        }  
/* Now write the parameters back to the board */
        if((status=chosen->btype->funcs->dxp_write_dspparams(&ioChan, &modChan, 
										chosen->dsp[modChan], chosen->params[modChan]))!=DXP_SUCCESS) break;  
    }  
    
	if (status!=DXP_SUCCESS){  
        sprintf(info_string,"Failure occurred restoring detector chan %d", detChan);  
        dxp_md_error("restore_config",info_string,&status);  
        return status;  
    }

	info->status |= 0x1;

    return status;  
}  

/******************************************************************************
 *
 * Open a new file: return *lun that points to an entry in the fout[] array
 *
 ******************************************************************************/
int XERXES_API dxp_open_file(int* lun, char* name, int* mode)
/* int *lun;						Output: logical unit number to refer to file */
/* char *name;						Input: file name of file to open             */
/* int *mode;						Input: 0: open new file for output           */
								/*         1: open new file or append to existing*/
								/*               file for output                 */
								/*         2: open existing file for input       */
{
    int status,i;
    char *type=NULL;
    static int first=1;
/*
 *    initialize file pointers on first call
 */
    if(first==1){
        first=0;
        dxpopen=0;
        for(i=0;i<MAXFILE;i++) filpnt[i]=NULL;
    }
/*
 *     make sure file mode is defined...
 */
    if((*mode<0)||(*mode>2)){
        sprintf(info_string,"mode = %d unrecognized; %s not opened",*mode,name);
        status=DXP_BAD_PARAM;
        dxp_md_error("dxp_open_file",info_string,&status);
        return status;
    }
    if (*mode==0) 
        type = "w";
    else if(*mode==1)
        type = "a";
    else
        type = "r";
/*
 *      make sure not too many files are open...
 */
    dxpopen++;
    if(dxpopen==MAXFILE){
		dxpopen--;
        sprintf(info_string,"too many files open, limit = %d",MAXFILE);
        status = DXP_OPEN_FILE;
        dxp_md_error("dxp_open_file",info_string,&status);
        return status;
    }
/*
 *       find an unused pointer
 */
    for(i=0;i<MAXFILE;i++) if(filpnt[i]==NULL) break;
    *lun=i;
    if((filpnt[*lun]=fopen(name,type))==NULL){
        sprintf(info_string,"Error opening %s ",name);
        status=DXP_OPEN_FILE;
        dxp_md_error("dxp_open_file",info_string,&status);
        return status;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Closes the file pointed to by *lun.  If *lun=-1, then close all files.
 *
 ******************************************************************************/
int XERXES_API dxp_close_file(int* lun)
/* int *lun;					*/
{

    int i,status;

    if(((*lun)<-1)||((*lun)>=MAXFILE)){
        status = DXP_BAD_PARAM;
        dxp_md_error("dxp_close_file","Illegal logical unit number",&status);
        return status;
    }
    if(*lun==-1){
        for(i=0;i<MAXFILE;i++){
            if(filpnt[i]!=NULL){
                if(fclose(filpnt[i])==EOF){
                    sprintf(info_string,"Error closing lun = %d",i);
                    status=DXP_CLOSE_FILE;
                    dxp_md_error("dxp_close_file",info_string,&status);
                }
            }
        }
        dxpopen=0;
        return DXP_SUCCESS;
    }
    if(filpnt[*lun]==NULL){
        sprintf(info_string,"lun %d was not opened -- no action taken",*lun);
        dxp_md_error("dxp_close_file",info_string,&debug);
    } else {
        if(fclose(filpnt[*lun])==EOF){
            sprintf(info_string,"Error closing lun = %d",*lun);
            status=DXP_CLOSE_FILE;
            dxp_md_error("dxp_close_file",info_string,&status);
            return status;
        }
        filpnt[*lun]=NULL;
        dxpopen--;
    }

    return DXP_SUCCESS;
}

/********------******------*****------******-------******-------******------*****
 * Routines to calculate values associated with modules.  Also routines to
 * perform limited IO to the modules.
 *
 ********------******------*****------******-------******-------******------*****/

/******************************************************************************
 *
 * Download nparam values to one dectector channel's DSP memory space 
 * specified by addr[]
 *
 ******************************************************************************/
int XERXES_API dxp_download_one_params(int* detChan, int* nparam, unsigned short addr[], 
							unsigned short value[])
/* int *detChan;						Input: detector channel number	  */
/* int *nparam;							Input: number of params to dnload  */
/* unsigned short addr[];				Input: address of each param       */
/* unsigned short value[];				Input: parameter values            */
{

    int ioChan, modChan, status;
	unsigned short i;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Find the detector number in the system */

	if ((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Unknown detector element: %d", *detChan);
		dxp_md_error("dxp_download_one_params", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Loop over the parameters requested, writing each value[] to addr[]. */
        
	for(i=0;i<*nparam;i++){
		if((status=chosen->btype->funcs->dxp_write_dsp_param_addr(&ioChan,&modChan,
									&addr[i],&value[i]))!=DXP_SUCCESS){
			sprintf(info_string,"error writing parameter %d to detector %d",i,*detChan);
			dxp_md_error("dxp_download_one_params",info_string,&status);
			return status;
		}
	}

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 *  Download nparam values to DSP memory space specified by addr[]
 *
 ******************************************************************************/
int XERXES_API dxp_download_params(int* nparam, unsigned short addr[],
						unsigned short value[])
/* int *nparam;							Input: number of params to dnload  */
/* unsigned short addr[];				Input: address of each param       */
/* unsigned short value[];				Input: parameter values            */
{

    int status;
	unsigned short i;
/* Pointer to the current Board */
	Board *current=system_head;

/* Loop over all modules in the system */    
	
	while (current!=NULL) {

/* Loop over the parameters requested, writing each value[] to addr[]. */
        
		for(i=0;i<*nparam;i++){
            if((status=current->btype->funcs->dxp_write_dsp_param_addr(&(current->ioChan),&allChan,
											&addr[i],&value[i]))!=DXP_SUCCESS){
                sprintf(info_string,"error writing parameter %d to module %d",
                    i,current->mod);
                dxp_md_error("dxp_download_params",info_string,&status);
                return status;
            }
        }
		current = current->next;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Upload nparam values from DSP memory space of a single detector channel
 * starting at a specified address.
 *
 ******************************************************************************/
int XERXES_API dxp_upload_channel(int* detChan, unsigned short* nparam, 
					   unsigned short* addr, unsigned long value[])
/* int *detChan;							Input: detector channel to readout	*/
/* unsigned short *nparam;					Input: number of params to dnload	*/
/* unsigned short *addr;					Input: address of 1st param			*/
/* unsigned long value[];					Output: parameter values				*/
{

    int status;
	unsigned short i;
	char name[20];

	int ioChan, modChan;
/* Pointer to the chosen Channel */
	Board *chosen=NULL;

/* Get the IO channel number and module channel number that matches the detChan */

	if ((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Unknown detector element: %d", *detChan);
		dxp_md_error("dxp_upload_channel", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

	for (i = *addr; i < (*addr + *nparam); i++) {
		if ((status=chosen->btype->funcs->dxp_symbolname(&i, chosen->dsp[modChan], name))!=DXP_SUCCESS) {
			sprintf(info_string,"Error retrieving symbol name for address %d", i);
			dxp_md_error("dxp_upload_channel",info_string,&status);
			return status;
		}

		if ((status=chosen->btype->funcs->dxp_read_dspsymbol(&ioChan, &modChan, 
										name, chosen->dsp[modChan], &value[i]))!=DXP_SUCCESS){
			sprintf(info_string,"Error reading parameters from det chan %d",*detChan);
			dxp_md_error("dxp_upload_channel",info_string,&status);
			return status;
		}
	}

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Loop over all DXP channels and compare downloaded DECIMATION value with
 * *decimation argument.  If all the downloaded values are the same as 
 * the *decimation argument, *different is set to 0; otherwise it is set to 1
 *
 ******************************************************************************/
int XERXES_API dxp_check_decimation(unsigned short* decimation, unsigned short* different)
/* unsigned short *decimation;		Input: decimation value           */
/* unsigned short *different;		Output: 0 if all are same, else 1 */
{

    unsigned int chan;
	int status, ichan;
    unsigned long value;
/* Pointer to the current Board */
	Board *current=system_head;

/* Loop over all the modules, checking that the decimation is correct */
    
	*different = 1;
	while (current!=NULL) {
        for(chan=0;chan<current->nchan;chan++){
            if(((current->used)&(1<<chan))==0) continue;

			ichan = (int) chan;
            if((status=current->btype->funcs->dxp_read_dspsymbol(&(current->ioChan),&ichan,
								"DECIMATION", current->dsp[chan], &value))!=DXP_SUCCESS){
                sprintf(info_string,
                    "Error reading decimation for module %d channel %d",
                    current->mod,chan);
                dxp_md_error("dxp_check_decimation",info_string,&status);
                return status;
            }
            if(((unsigned short) value) != *decimation) return status;
        }
		current = current->next;
    }

/* No differences found, set to 0 and return. */

    *different = 0;
    return DXP_SUCCESS;
}

/******************************************************************************
 *
 *  Determine the run statistics.  Livetime, ICR, OCR.
 *
 ******************************************************************************/
int XERXES_API dxp_get_statistics(int* detChan, double* livetime, double* icr, 
					   double* ocr, unsigned long* nevents)
/* int *detChan;					Input: Detector Channel number			*/
/* double *livetime;				Output: livetime for last run (sec)		*/
/* double *icr;						Output: input count rate (Kcounts/sec)	*/
/* double *ocr;						Output: output count rate (Kcounts/sec)	*/
/* unsigned long *nevents;				Output: number of events in last run		*/
{

    int status;

	unsigned int evts, under, over, fast, base;

	int ioChan, modChan;
/* Pointer to Chosen Board */
	Board *chosen=NULL;

/* Get the IO channel number and module channel number that matches the detChan */

	if ((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Unknown detector element: %d", *detChan);
		dxp_md_error("dxp_get_statistics", info_string, &status);
		return status;
	}
	ioChan = chosen->ioChan;

/* Allow driver to do neede ops to prepare channel for readout */

    if((status=chosen->btype->funcs->dxp_prep_for_readout(&(chosen->ioChan), &modChan))!=DXP_SUCCESS){
        dxp_md_error("dxp_get_statistics","Error Preparing board for readout",&status);
        return status;
    }

/* Read out the parameter memory of the specified detector */

	if ((status=chosen->btype->funcs->dxp_read_dspparams(&ioChan, &modChan, 
						chosen->dsp[modChan], chosen->params[modChan]))!=DXP_SUCCESS) {
		sprintf(info_string, "Unable to read parameters from detector element: %d", 
							*detChan);
		dxp_md_error("dxp_get_statistics", info_string, &status);
		return status;
	}

/* Get the statistics data from the parameter array */

	if((status=chosen->btype->funcs->dxp_get_runstats(chosen->params[modChan], 
								chosen->dsp[modChan], &evts, &under, &over, &fast, 
								&base, livetime))!=DXP_SUCCESS){
		dxp_md_error("dxp_get_statistics","Error parsing statistics",&status);
		return status;
	}

/* Calculate the number of events in the run */

	*nevents = evts + under + over;

/* Calculate the event rates */

	if(*livetime>0.){
		*icr = 0.001 * (fast) / (*livetime);
		*ocr = 0.001 * (*nevents) / (*livetime);
	} else *icr = *ocr = -999.;

/* Allow driver to return board to previous state */

    if((status=chosen->btype->funcs->dxp_done_with_readout(&(chosen->ioChan), 
											&modChan, chosen->state))!=DXP_SUCCESS){
        dxp_md_error("dxp_get_statistics","Error restarting board for data taking",&status);
        return status;
    }

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Search the spectrum with nbins bins for the maximum entry.  Then search on
 * either side of the maximum till the spectrum entry falls below thresh.percent
 * of the peak value.  Return the lower and upper limits of the "peak".
 *
 ******************************************************************************/
int XERXES_API dxp_findpeak(long spectrum[], int* nbins, float* thresh, 
				 int* lower, int* upper)
/* long spectrum[];					Input: MCA spectrum                    */
/* int *nbins;						Input: number of bins                  */
/* float *thresh;					Input: Fraction of peak value          */
/* int *lower;						Output: lower bin limit                */
/* int *upper;						Output: upper bin limit                */
{

    int i,pkpos= -1;
    long max=0,min;

/* Search the spectrum for the bin with the most entries. */

    for(i=0;i<(*nbins);i++){
        if(spectrum[i]>max){
            max=spectrum[i];
            pkpos = i;
        }
    }
/*
 * find upper bin limit
 */
    min = (long) ((*thresh)*((float) max));
    i=pkpos;
    while(spectrum[i]>min) i++;
    *upper = i;
/*
 *  find lower bin limit
 */
    i = pkpos;
    while(spectrum[i]>min) i--;
    *lower = i;

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * Perform a quick and dirty gaussian fit to a spectrum using following 
 * linearization method: (from Knoll, "Radiation Detection and Measurement"
 * 2nd Edition, John Wiley & Sons, 1989, p677-8...)
 *
 *   y(x) = y0*exp(-(x-x0)^2/(2*sig^2))
 *
 *   define Q(x) = y(x-1)/y(x+1), then
 *
 *   log(Q(x)) = 2*(x-x0)/sigma^2
 *
 *   Thus, fit log(Q) vs. x to a straight line...
 *
 ******************************************************************************/
int XERXES_API dxp_fitgauss0(long spectrum[], int* lower, int* upper, float* pos,
				   float* fwhm)
/* long spectrum[];					Input: spectrum to fit                  */
/* int *lower;						Input: starting bin to use in fit       */
/* int *upper;						Input: ending bin to use in fit         */
/* float *pos;						Output: fit result - peak position      */
/* float *fwhm;						Output: fit result - full width half max*/
{

    int s0,status=DXP_SPECTRUM,i;
    double s1,sx,sy,sxx,sxy,det,a,b,fact,x,y,wt;
    long cut1=10;					/* minummm number of events in bin			*/
    int cut2=3;						/* minimum number of bins used in fit		*/ 

    s0=0;
	s1=sx=sy=sxx=sxy=0.;
    for(i= *lower;i<= *upper;i++){
        if((spectrum[i-1]<cut1)||(spectrum[i+1]<cut1))continue;
        s0++;
        x = (float)i + 0.5;
        y = log((spectrum[i-1]+0.0)/spectrum[i+1]);
        wt= 1.0/(1./spectrum[i+1]+1./spectrum[i-1]);
        s1+=wt;
        sx+=wt*x;
        sy+=wt*y;
        sxx+=wt*x*x;
        sxy+=wt*x*y;
    }
    if(s0<cut2){
        sprintf(info_string,"Too few bins: start = %d stop = %d bins = %d",
            *lower,*upper,s0);
        dxp_md_error("dxp_fitgauss0",info_string,&status);
        return status;
    }
    det = s1*sxx-sx*sx;
    if(det==0.){
        dxp_md_error("dxp_fitgauss0","Singular determinant",&status);
        return status;
    }
    fact = 1./det;
    a = fact*(sy*sxx-sx*sxy);
    b = fact*(s1*sxy-sx*sy);
    if(b==0.){
        dxp_md_error("dxp_fitgauss0","vanishing slope",&status);
        return status;
    }
    *fwhm = (float) (2.355*sqrt(2./b));
    *pos = (float) (-a/b);

    return DXP_SUCCESS;
}

/******************************************************************************
 *
 * This routine changes all DXP channel gains:
 *
 *            initial         final
 *            gain_i -> gain_i*gainchange
 *  
 * THRESHOLD is also modified to keep the energy threshold fixed.
 * Note: DACPERADC must be modified -- this can be done via calibration
 *       run or by scaling DACPERADC --> DACPERADC/gainchange.  This routine
 *       does not do either!
 *  
 ******************************************************************************/
int XERXES_API dxp_modify_gains(float* gainchange)
/* float *gainchange;			Input: desired gain change    */
{

	unsigned int chan;
	int status=DXP_SUCCESS;
	int ichan;
/* Pointer to the current Board */
	Board *current=system_head;

/* Now loop over all modules and channels in use */

	while (current!=NULL) {
        for(chan=0;chan<current->nchan;chan++){
            if(((current->used)&(1<<chan))==0) continue;

			ichan = (int) chan;
			if ((status=current->btype->funcs->dxp_change_gains(&(current->ioChan), &ichan, 
										&(current->mod), gainchange, current->dsp[chan]))!=DXP_SUCCESS) {
				sprintf(info_string, 
					"Error modifying Gain for Detector %d", current->detChan[chan]);
				dxp_md_error("dxp_modify_gains",info_string,&status);
				return status;
			}
		}
		current = current->next;
	}

	return status;
}

/******************************************************************************
 *
 * This routine changes gains in one DXP channel :
 *
 *            initial         final
 *            gain_i -> gain_i*gainchange
 *  
 * THRESHOLD is also modified to keep the energy threshold fixed.
 * Note: DACPERADC must be modified -- this can be done via calibration
 *       run or by scaling DACPERADC --> DACPERADC/gainchange.  This routine
 *       does not do either!
 *  
 ******************************************************************************/
int XERXES_API dxp_modify_one_gains(int* detChan, float* gainchange)
/* int *detChan;					Input: channel to modify		*/
/* float *gainchange;				Input: desired gain change	*/
{

	int modChan;
	int status=DXP_SUCCESS;
/* Pointer to the chosen Board */
	Board *chosen=NULL;

/* Get the IO channel number and module channel number that matches the detChan */

	if ((status=dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Unknown detector element: %d", *detChan);
		dxp_md_error("dxp_modify_one_gains", info_string, &status);
		return status;
	}

/* Now modify the gain on this channel */

	if ((status=chosen->btype->funcs->dxp_change_gains(&(chosen->ioChan), &modChan, 
								&(chosen->mod), gainchange, chosen->dsp[modChan]))!=DXP_SUCCESS) {
		sprintf(info_string, 
			"Error modifying Gain for Detector %d", chosen->detChan[modChan]);
		dxp_md_error("dxp_modify_one_gains",info_string,&status);
		return status;
	}

	return status;
}

/******************************************************************************
 *
 * Read file pointed to by PREAMP_CONFIG and adjust each DXP channel's gain 
 *    parameters to achieve the desired ADC rule.  The ADC rule is the fraction
 *    of the ADC full scale that a single x-ray step of energy *energy contributes.
 *    PREAMP_CONFIG has measurements taken for each detector channel at the 
 *    specified energy:
 *
 *     channel number (same as in DXP_MODULE file)
 *     gainmod (=1.0 for standard model) front end gain factor
 *     polarity (0 for negative, 1 for positive)
 *     vmin = minimum voltage of preamp in volts
 *     vmax = maximum voltage of preamp in volts
 *     vstep= voltage step size (in mV)
 *
 * set following, detector specific parameters for all DXP channels:
 *       COARSEGAIN
 *       FINEGAIN
 *       VRYFINGAIN (default:128)
 *       POLARITY
 *       OFFDACVAL
 *
 * apply internal gain calibrations (TRACKDAC and RESET)
 *
 ******************************************************************************/
int XERXES_API dxp_initialize_ASC(float* adcRule, float* energy)
/* float *adcRule;							Input: desired ADC rule           */
/* float *energy;							Output: energy from PREAMP_CONFIG */
										/*          in eV                     */
{

	int status, i;
    int detChan,ioChan,dxpChan;
    char *fstatus=" ";
    FILE *fp;
    unsigned short polarity;
    float vstep,vmin,vmax,gainmod;
/* Pointer to the chosen and current Board */
	Board *chosen=NULL;
	Board *current = system_head;
/* Temporary holding for the preamp file name */
	char strtmp[4];
	char newFile[MAXFILENAME_LEN];

/* Convert 1st 4 characters to upper for comparison */   
	for (i=0;i<4;i++) strtmp[i] = (char) toupper(info->preamp[i]);
/* If preamp is undefined on purpose, then return with no changes */
	if (strncmp(strtmp,"NULL",4)==0) {
		status = DXP_SUCCESS;
		return status;
	}
/* Else try to open the file pointed to by info->preamp */
	if((fp=dxp_find_file(info->preamp,"r",newFile))==NULL){
        status = DXP_OPEN_FILE;
        sprintf(info_string,"could not open preamp configuration file : %s",
											info->preamp);
        dxp_md_error("dxp_initialize_ASC",info_string,&status);
        return status;
    }


    while(fstatus!=NULL){
/* Read lines in the file until we get a non comment or EOF line */

        do
            fstatus = fgets(line,132,fp);
        while((line[0]=='*')&&(fstatus!=NULL));
        if(fstatus==NULL) continue;

/* Pull out the information from the file.  If there is a label called ENERGY, then
 * echo that information back to the caller, else continue with the detector channel
 * number as the first entry in the file. */

        token=strtok(line,delim);
        if(strcmp(token,"ENERGY")==0){
            token=strtok(NULL,delim);
            *energy = (float) strtod(token,NULL);
            continue;
        }

/* Get the detector channel number.  This is the same number stored in the detchan array
 * and represents the user defined number associated with a given detector in the system.
 * It also points to a member of the camChan array that allows communication. */
        
		detChan= (int) strtol(token,NULL,10);

/* Confirm that this detchan is in the system configuration. And return the CAMAC access
 * path for that channel. */

		if((status=dxp_det_to_elec(&detChan, &chosen, &dxpChan))!=DXP_SUCCESS){
            dxp_md_error("dxp_initialize_ASC","no action taken",&status);
            continue;
        }
		ioChan = chosen->ioChan;

/* Now pull the rest of the information in order:  GAINMOD, POLARITY, VMIN, VMAX, VSTEP */

		chosen->preamp[dxpChan].energy  = *energy;
		chosen->preamp[dxpChan].adcrule = *adcRule;

		token=strtok(NULL,delim);
        gainmod = (float) strtod(token,NULL);
		chosen->preamp[dxpChan].gainmod = gainmod;

        token=strtok(NULL,delim);
        polarity = (short) strtol(token,NULL,10);
		chosen->preamp[dxpChan].polarity = polarity;

        token=strtok(NULL,delim);
        vmin = gainmod*((float) strtod(token,NULL));
		chosen->preamp[dxpChan].vmin = vmin;

        token=strtok(NULL,delim);
        vmax = gainmod*((float) strtod(token,NULL));
		chosen->preamp[dxpChan].vmax = vmax;

        token=strtok(NULL,delim);
        vstep = gainmod*((float) strtod(token,NULL));
		chosen->preamp[dxpChan].vstep = vstep;

/* Now perform the ASC initialization on this detector channel */

		if ((status=chosen->btype->funcs->dxp_setup_asc(&ioChan, &dxpChan, &(chosen->mod), 
								adcRule, &gainmod, &polarity, &vmin,
								&vmax, &vstep, chosen->dsp[dxpChan]))!=DXP_SUCCESS) {
/* Close the preamp file */
			fclose(fp);
			sprintf(info_string, 
				"Error setting up the ASC for Detector %d", detChan);
			dxp_md_error("dxp_initialize_ASC",info_string,&status);
			return status;
		}
	}

/* Close the preamp file */
	fclose(fp);
	
/*
 *  Perform calibrations neccessary to setup the ASC
 */
	while (current!=NULL) {
	    if((status=current->btype->funcs->dxp_calibrate_asc(&(current->mod), 
						&(current->ioChan), &(current->used),current->dsp))!=DXP_SUCCESS){
			dxp_md_error("dxp_initialize_ASC",
				"Unable to calibrate the ASC",&status);
			return status;
		}
		current = current->next;
	}

    return DXP_SUCCESS;
}

/******************************************************************************
 * 
 * Returns the electronic channel numbers (e.g. module number and channel
 * within a module) for a given detector channel.
 *
 ******************************************************************************/
int XERXES_API dxp_det_to_elec(int* detChan, Board** passed, int* dxpChan)
/* int *detChan;						Input: detector channel               */
/* Board **passed;						Output: Pointer to the correct Board   */
/* int *dxpChan;						Output: DXP channel number             */
{

	unsigned int chan;
    int status;

/* Point the current Board to the first Board in the system */
	Board *current = system_head;

/* Loop through the Boards to find which module we desire */

    while (current != NULL) {
		for (chan=0;chan<current->nchan;chan++) {
			if (*detChan==current->detChan[chan]) {
				*dxpChan = chan;
				*passed = current;
				return DXP_SUCCESS;
			}
		}
		current = current->next;
	}

	*passed = NULL;
	sprintf(info_string,"detector channel %d is unknown",*detChan);
	status=DXP_NODETCHAN;
    dxp_md_error("dxp_det_to_elec",info_string,&status);
    return status;
}

/******************************************************************************
 * 
 * Returns the electronic channel numbers (e.g. module number and channel
 * within a module) for a given detector channel.
 *
 ******************************************************************************/
int XERXES_API dxp_elec_to_det(int* ioChan, int* dxpChan, int* detChan)
/* int *ioChan;						Input: CAMAC I/O number				*/
/* int *dxpChan;					Input: DXP channel number			*/
/* int *detChan;					Output: detector channel			*/
{

    int status;
/* Pointer to current Board */
	Board *current=system_head;

/* Only support up to 4 channels per module. */
	
	if (((unsigned int) *dxpChan) >= current->nchan) {
        sprintf(info_string,"No support for %d channels per module",*dxpChan);
        status=DXP_NOMODCHAN;
        dxp_md_error("dxp_elec_to_det",info_string,&status);
        return status;
	}

/* Now count forward through the module and channel list till we 
 * find the matching detector number */
    
	while (current!=NULL) {
		if (*ioChan==(current->ioChan)) {
			*detChan = current->detChan[*dxpChan];
			return DXP_SUCCESS;
        }
		current = current->next;
	}

/* Should never exit loop normally => no match */
	
	sprintf(info_string,"IO Channel %d unknown",*ioChan);
    status=DXP_NOIOCHAN;
    dxp_md_error("dxp_elec_to_det",info_string,&status);
    return status;
}

/******************************************************************************
 *
 * Routine to report the software version number
 *
 ******************************************************************************/
void XERXES_API dxp_version(VOID)
{
	int version=CODE_VERSION;
	int revision = CODE_REVISION;
/*	char type[20] = LATEST_BOARD_TYPE;
	char brevision[5] = LATEST_BOARD_VERSION;
*/
	sprintf(info_string, "\r\nCode version %d.%d\r\n", version, revision);
	dxp_md_puts(info_string);
/*	sprintf(info_string, "Works with board type %s\r\n", type);
	dxp_md_puts(info_string);
	sprintf(info_string, "Works with board revision %s\r\n", brevision);
	dxp_md_puts(info_string);
*/
	return;
}

/******************************************************************************
 *
 * Routine to tell the MD layer to lock resources associated with detChan
 *
 ******************************************************************************/
int XERXES_API dxp_lock_resource(int *detChan, short *lock)
{
	int status = DXP_SUCCESS;
	int modChan;
/* Pointer to the Chosen Channel */
	Board *chosen=NULL;

/* Translate the detector channel number to IO and channel numbers for writing. */

	if ((status = dxp_det_to_elec(detChan, &chosen, &modChan))!=DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_lock_resource", info_string, &status);
		return status;
	}

/* Set the state of this resource to be locked iff *lock != 0 */

	chosen->state[3] = (short) ((*lock != 0) ? 1 : 0);
	
/* Check to see if the lock function is implemented at the MD layer */
	if (chosen->iface->funcs->dxp_md_lock_resource != NULL) {
		if ((status = chosen->iface->funcs->dxp_md_lock_resource(&(chosen->ioChan), 
														&modChan, lock))!=DXP_SUCCESS) {
			sprintf(info_string, "Failed to lock resource for detector %d", *detChan);
			dxp_md_error("dxp_lock_resource", info_string, &status);
			return status;
		}
	} else {
		status = DXP_MDLOCK;
		sprintf(info_string, "No Resource Locking function implemented for detector %d", *detChan);
		dxp_md_error("dxp_loc_resource", info_string, &status);
	}

	return status;
}

/******************************************************************************
 * 
 * Routine to readout the ADC trace (if possible) for a given channel
 *
 ******************************************************************************/
int XERXES_API dxp_read_adc(int *detChan, unsigned short *data)
{
	int status = DXP_SUCCESS;
	int modChan;
	int isRunning;

	Board *chosen = NULL;

/* Get pointer to the channel */
	if ((status = dxp_det_to_elec(detChan, &chosen, &modChan)) != DXP_SUCCESS) {
		sprintf(info_string, "Failed to locate detector channel %d", *detChan);
		dxp_md_error("dxp_read_adc", info_string, &status);
		return status;
	}

/* Make sure a run isn't active on the channel of interest */
	if ((status = dxp_isrunning(detChan, &isRunning)) != DXP_SUCCESS) {
		sprintf(info_string, "Failed to determine run status for %d", *detChan);
		dxp_md_error("dxp_read_adc", info_string, &status);
		return status;
	}
	if (isRunning != 0) {
		status = DXP_ADC_RUNACTIVE;
		sprintf(info_string, "Detector channel %d is already running", *detChan);
		dxp_md_error("dxp_read_adc", info_string, &status);
		return status;
	}

/* Now call the actual driver level routine */
	if (chosen->btype->funcs->dxp_get_adc != NULL) {
		if ((status = chosen->btype->funcs->dxp_get_adc(&(chosen->ioChan), &modChan, chosen->dsp[modChan], data)) != DXP_SUCCESS) {
			sprintf(info_string, "Failed reading ADC from detector channel %d", *detChan);
			dxp_md_error("dxp_read_adc", info_string, &status);
			return status;
		}
	} else {
		status = DXP_ADC_READ;
		sprintf(info_string, "No function to read ADC implemented for detector channel %d", *detChan);
		dxp_md_error("dxp_read_adc", info_string, &status);
		return status;
	}

	return status;
}


/******************************************************************************
 * Begin the routines for HanDeL (Hardware Definition Layer).  This is intended 
 * to act as a higher level interface to the hardware for users, accepting 
 * peaking times in microsecs, thresholds in KeV, etc...
 ******************************************************************************/

/******************************************************************************
 *
 * Routine to initialize HanDeL
 *
 ******************************************************************************/
int HANDEL_API xiaInit()
{
	int status = DXP_SUCCESS;

/* Initialize the XerXes Library */
	if((status=dxp_init_library())!=XIA_SUCCESS){
		dxp_md_error("xiaInit","Unable to Initiailize XerXes",&status);
		return status;
	}

/* Initialize the HanDeL Library */
	if((status=xiaInitHandel())!=XIA_SUCCESS){
		dxp_md_error("xiaInit","Unable to Initiailize HanDeL",&status);
		return status;
	}

/* Now read in the xia.ini file */
	if((status=xiaReadIniFile())!=XIA_SUCCESS){
		dxp_md_error("xiaInit","Unable to read .ini file",&status);
		return status;
	}

	return status;
}
/******************************************************************************
 *
 * Routine to initialize the Detector linked list.
 *
 ******************************************************************************/
int HANDEL_API xiaInitHandel()
{
	int status = DXP_SUCCESS;

/* Clear the HanDeL data structures */
	if((status=xiaInitHandelDS())!=XIA_SUCCESS){
		dxp_md_error("xiaInitHandel","Unable to clear Data Structures",&status);
		return status;
	}

	return status;
}
/******************************************************************************
 *
 * Routine to initialize the Detector linked list.
 *
 ******************************************************************************/
static int HANDEL_API xiaInitHandelDS()
{
	int status = DXP_SUCCESS;
	
/* Clear the HanDeL Detector structure */
	if((status=xiaInitDetectorDS())!=XIA_SUCCESS){
		dxp_md_error("xiaInitHandelDS","Unable to clear the Detector LL",&status);
		return status;
	}

/* Clear the HanDeL FirmwareSet structure */
	if((status=xiaInitFirmwareSetDS())!=XIA_SUCCESS){
		dxp_md_error("xiaInitHandelDS","Unable to clear the FirmwareSet LL",&status);
		return status;
	}

/* Clear all the HanDeL Daq Values structures */
	if((status=xiaInitDaqValuesDS())!=XIA_SUCCESS){
		dxp_md_error("xiaInitHandelDS","Unable to clear DaqValues from Boards",&status);
		return status;
	}

	return status;
}
/******************************************************************************
 *
 * Routine to initialize the Detector linked list.
 *
 ******************************************************************************/
static int HANDEL_API xiaReadIniFile()
{
	int status = DXP_SUCCESS;
/* 
 * Pointers to keep track of the xia.ini file
 */
	FILE *fp = NULL;
	char newFile[MAXFILENAME_LEN];
	char name[132],value[132];
	fpos_t start=0, end=0, current=0;

/* Open the .ini file */
	if ((fp=dxp_find_file("xia.ini","r",newFile))==NULL){
		status = XIA_OPEN_FILE;
		dxp_md_error("xiaReadIniFile","Could not open xia.ini",&status);
		return status;
	}
	
/* First thing is to check if there is a [xerxes definitions] for XerXes 
 * startup of HanDeL.
 */
	status = xiaFindEntryLimits(fp, "xerxes definitions", &start, &end);
/* Check to see if we have a "xerxes definitions" section in the 
 * file */
	if (status==XIA_SUCCESS) {
/* Set the current file position back to the start of the section */
		if ((status=fsetpos(fp, &start))!=0) {
			status = XIA_FILEERR;
			dxp_md_error("xiaReadIniFile","Unable to set File Position",&status);
			return status;
		}
		do {
/* Get the next non-blank line from the file */	
			if ((status=xiaGetLine(fp, line))!=0) {
				dxp_md_error("xiaReadIniFile","Error searching for line in xia.ini",&status);
				return status;
			}
/* Check to see if we went past the end of this section */
			if ((status=fgetpos(fp, &current))!=0) {
				status = XIA_FILEERR;
				dxp_md_error("xiaReadIniFile","Unable to get File Position",&status);
				return status;
			}
			if (current<=end) {
/* Get the name value pair from this line */
				if ((status=xiaGetLineData(line, name, value))!=XIA_SUCCESS) {
					dxp_md_error("xiaReadIniFile","Error searching line for name-value pair",&status);
					return status;
				}
/* Did we find a XerXes configuration tag? */
				if (strcmp(name, "xia_config")==0) {
/* If so, then configure XerXes */
/* Read the configuration file */
					if((status=dxp_read_config(value))!=DXP_SUCCESS){
						sprintf(info_string,"Error reading configuration file.");
						dxp_md_error("xiaReadIniFile",info_string,&status);
						return status;
					}

/* Setup the modules by assigning them IO and module channel numbers */
					if ((status=dxp_assign_channel())!=DXP_SUCCESS){
						status = XIA_INITIALIZE;
						dxp_md_error("xiaReadIniFile","Could not assign channels for this config.",&status);
						return status;
					}

/* Setup the modules by assigning them IO and module channel numbers */
					if ((status=dxp_user_setup())!=DXP_SUCCESS){
						status = XIA_INITIALIZE;
						dxp_md_error("xiaReadIniFile",
							"Could not perform user setup(downloading and calibrations)",&status);
						return status;
					}
				}
			}
		} while (current<end);
	
	}


	return status;
}

/******************************************************************************
 *
 * Routine that gets the first line with text after the current file 
 * position.
 *
 ******************************************************************************/
static int HANDEL_API xiaGetLineData(char *line, char *name, char *value)
{
	int status = XIA_SUCCESS;
	unsigned int j;

	size_t loc, begin, end, len;

/* Start by finding the '=' within the line */
	loc = strcspn(line, "=");
	if (loc==0) {
		status = XIA_FORMAT_ERROR;
		sprintf(info_string, "No = present in xia.ini line: \n %s", line);
		dxp_md_error("xiaGetLineData", info_string, &status);
		return status;
	}

/* Strip all the leading blanks */
	begin=0;
	while (isspace(line[begin])) {
		begin++;
	}

/* Strip all the leading blanks */
	end=loc-1;
	while (isspace(line[end])) {
		end--;
	}

/* copy the name */
	len = end-begin+1;
	strncpy(name, line+begin, len);
	name[len] = '\0';
	
/* Strip all the leading blanks */
	begin=loc+1;
	while (isspace(line[begin])) {
		begin++;
	}

/* Strip all the leading blanks */
	end=strlen(line)-1;
	while (isspace(line[end])) {
		end--;
	}

/* copy the value */
	len = end-begin+1;
	strncpy(value, line+begin, len);
	value[len] = '\0';

/* Convert name to lower case */
	for (j=0;j<strlen(name);j++) name[j] = (char) tolower(name[j]);
	
	return status;
}

/******************************************************************************
 *
 * Routine that gets the first line with text after the current file 
 * position.
 *
 ******************************************************************************/
static int HANDEL_API xiaGetLine(FILE *fp, char *line)
{
	int status = XIA_SUCCESS;
	unsigned int j;

	char *cstatus;

/* Now fine the match to the section entry */
	do {
		cstatus = fgets(line,132,fp); 
/* Check for any alphanumeric character in the line */
		for (j=0;j < strlen(line);j++) { 
			if (isgraph(line[j])) return status;
		}
	} while (cstatus!=NULL);

	return status;
}

/******************************************************************************
 *
 * Routine that searches thru the .ini file and finds the start and end of a 
 * specific section of the .ini file starting at [section] and ending at the 
 * next [].
 *
 ******************************************************************************/
static int HANDEL_API xiaFindEntryLimits(FILE *fp, const char *section, 
										 fpos_t *start, fpos_t *end)
{
	int status = XIA_SUCCESS;
	unsigned int j;

	char line[132];
	char *cstatus;

/* Zero the start and end variables */
	*start=*end=0;

/* First rewind the file to the beginning */
	rewind(fp);

/* Now fine the match to the section entry */
	do {
		do {
			cstatus = fgets(line,132,fp); 
		} while ((line[0]!='[')&&(cstatus!=NULL)); 
		if(cstatus==NULL) {
			status = XIA_NOSECTION;
			sprintf(info_string,"Unable to find section %s",section);
			dxp_md_error("XiaFindEntryStart",info_string,&status);
			return status;
		}
/* Find the terminating ] to this section */
		for (j=1;j < strlen(line);j++) { 
			if (line[j]==']') break;
		}
/* Did we not find the terminating ]? */
		if (j==strlen(line)) {
			status = XIA_FORMAT_ERROR;
			sprintf(info_string,"Syntax error in Init file, no terminating ] found");
			dxp_md_error("XiaFindEntryStart",info_string,&status);
			return status;
		}
/* Check if this is the section we are interested in. */
		if (strncmp(line+1,section,j-1)==0) {
/* Recorcd the starting position) */
			fgetpos(fp, start);
		}
/* Else look for the next section entry */
	} while (*start==0);

	do {
/* Get the current file position before the next read.  If the file
 * ends or we find a '[' then we are done and want to sent the ending 
 * position to the location of the previous read
 */
		fgetpos(fp, end);
		cstatus = fgets(line,132,fp); 
	} while ((line[0]!='[')&&(cstatus!=NULL));

/* No real way to detect error at this point, either the file will end 
 * or a '[' will be found.
 */
	return status;
}

/******************************************************************************
 *
 * Initialize the DEFAULTS data structures to an empty state
 *
 ******************************************************************************/
static int HANDEL_API xiaInitDetectorDS(void)
{
	int status = XIA_SUCCESS;
	Detector *next;
	Detector *current = xiaDetectorsHead;

/* Search thru the Detector LL and clear them out */
	while ((current != NULL) && (status==XIA_SUCCESS)) {
		next = current->next;
/* Clear the memory first */
		status = xiaFreeDetector(current);
		current = next;
	}
/* And clear out the pointer to the head of the list(the memory is freed) */
	xiaDetectorsHead = NULL;
	
/* NULL the current defaults in the static variable, to ensure we dont point 
 * free memory */
	workingDetector = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an Dsp_Defaults Structure
 *
 ******************************************************************************/
static int HANDEL_API xiaFreeDetector(Detector *detector)
/* Detector *detector;					Input: pointer to structure to free	*/
{
	int status=XIA_SUCCESS;

	if (detector==NULL) {
		status = XIA_NOMEM;
		sprintf(info_string,"Detector object unallocated:  can not free");
        dxp_md_error("xiaFreeDetector",info_string,&status);
        return status;
	}

	if (detector->alias!=NULL)
				dxp_md_free(detector->alias);
	if (detector->polarity!=NULL)
				dxp_md_free(detector->polarity);
	if (detector->gain!=NULL) 
				dxp_md_free(detector->gain);

/* Free the Board_Info structure */
	dxp_md_free(detector);
	detector = NULL;

	return status;
}

/******************************************************************************
 *
 * Initialize the DEFAULTS data structures to an empty state
 *
 ******************************************************************************/
static int HANDEL_API xiaInitFirmwareSetDS(void)
{
	int status = XIA_SUCCESS;
	FirmwareSet *next;
	FirmwareSet *current = xiaFirmwareSetHead;

/* Search thru the FirmwareSet LL and clear them out */
	while ((current != NULL) && (status==XIA_SUCCESS)) {
		next = current->next;
/* Clear the memory first */
		status = xiaFreeFirmwareSet(current);
		current = next;
	}
/* And clear out the pointer to the head of the list(the memory is freed) */
	xiaFirmwareSetHead = NULL;
	
/* NULL the current FirmwareSet in the static variable, to ensure we dont point 
 * free memory */
	workingFirmwareSet = NULL;

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an FirmwareSet Structure
 *
 ******************************************************************************/
static int HANDEL_API xiaFreeFirmwareSet(FirmwareSet *firmwareSet)
/* FirmwareSet *firmwareSet;				Input: pointer to structure to free	*/
{
	int status=XIA_SUCCESS;

	Firmware *next, *current;
	
	if (firmwareSet==NULL) {
		status = XIA_NOMEM;
		sprintf(info_string,"FirmwareSet object unallocated:  can not free");
        dxp_md_error("xiaFreeFirmwareSet",info_string,&status);
        return status;
	}

	if (firmwareSet->alias!=NULL)
				dxp_md_free(firmwareSet->alias);

/* Loop over the Firmware information, deallocating memory */
	current = firmwareSet->firmware;
	while (current!=NULL) {
		next = current->next;
		xiaFreeFirmware(current);
		current = next;
	}

/* Free the FirmwareSet structure */
	dxp_md_free(firmwareSet);
	firmwareSet = NULL;

	return status;
}
/******************************************************************************
 *
 * Routine to free the memory allocated to an FirmwareSet Structure
 *
 ******************************************************************************/
static int HANDEL_API xiaFreeFirmware(Firmware *firmware)
/* Firmware *firmware;				Input: pointer to structure to free	*/
{
	int status=XIA_SUCCESS;

	if (firmware==NULL) {
		status = XIA_NOMEM;
		sprintf(info_string,"Firmware object unallocated:  can not free");
        dxp_md_error("xiaFreeFirmware",info_string,&status);
        return status;
	}

	dxp_md_free(firmware);
	firmware = NULL;

	return status;
}

/******************************************************************************
 *
 * Initialize the DaqValues structures 
 *
 ******************************************************************************/
static int HANDEL_API xiaInitDaqValuesDS(void)
{
	int status = XIA_SUCCESS;
	
	Board *next;
	Board *current = system_head;

/* Search thru the Boards LL and clear out the DaqValues structures */
	while ((current != NULL) && (status == XIA_SUCCESS)) {
		next = current->next;
/* Clear the memory first */
		status = xiaFreeDaqValues(current);
		current = next;
	}

	return status;
}

/******************************************************************************
 *
 * Routine to free the memory allocated to an FirmwareSet Structure
 *
 ******************************************************************************/
static int HANDEL_API xiaFreeDaqValues(Board *current)
/* Board *current;				Input: Pointer to Board containing DaqValues */
{
	int status = XIA_SUCCESS;
	unsigned int i;

	if (current==NULL) {
		status = XIA_NOMEM;
		sprintf(info_string,"Board object unallocated:  can not clear DaqValues");
        dxp_md_error("xiaFreeDaqValues",info_string,&status);
        return status;
	}

/* If there are ROIs defined, remove */
	for (i=0;i<current->nchan;i++) {
		if (current->daq[i].nroi!=0) {
			dxp_md_free(current->daq[i].low_roi);
			dxp_md_free(current->daq[i].hi_roi);
		}
	}

/* Free the DaqValues structure */
	dxp_md_free(current->daq);
	current->daq = NULL;

	return status;
}
